/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */

#include <stdlib.h>
#include <memory.h>
#include <dnpap.h>
#include <config.h>
#include <message.h>
#include <mac.h>
#include <hash.h>
#include <sysmib.h>
#include <protocol.h>
#include <maxmem.h>
#include <channel.h>

#include "dmatrixd.h"
#include "dmatrixe.h"
#include "dmatrixc.h"

IMPORT BOOLEAN BeholderStrictRMON;


static LONG MatrixMaxNrSrcDsts = 8000;

static BOOLEAN MatrixCallback(MAC_COLL *collector, PROT_PKT *pkt);
static VOID MatrixHandlePkt(DNPAP_MATRIX_CONTROL *matrixcontrol, int type, PROT_PKT *pkt);

static LRUEntry *MatrixAddLRUList(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * entry);
static BOOLEAN MatrixRemoveLRUList(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * entry);
static BOOLEAN MatrixUpdateLRUList(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * connaddr);

static BOOLEAN MatrixDelConns(DNPAP_MATRIX_CONTROL * matrixcontrol);

static BOOLEAN MatrixDelConn(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * entry);
static VOID MatrixDelHost(DNPAP_MATRIX_CONTROL * matrixcontrol, Host * host, BOOLEAN source);
static VOID ListDelConn(Host * host, ConnInfo * conn);

static BOOLEAN MatrixAddHost(DNPAP_MATRIX_CONTROL * matrixcontrol, Host * host, BOOLEAN source);

static BOOLEAN ListAddTop(Host * host, ConnInfo * conn);
static BOOLEAN ListUpdateTop(Host * host, ConnInfo * conn);
static BOOLEAN ListAddLex(Host * host, ConnInfo * conn, WORD addrlen);

static ConnInfo *ListConnSearch(Host * host, BYTE * addr, WORD addrlen);


#ifdef DEBUG
static VOID ListCheckConns(Host * host);
#endif



BOOLEAN DnpapMatrixConfigInit(VOID)
{
	ConfigGetLong("beholder.dmatrix.maxnrsrcdsts", &MatrixMaxNrSrcDsts);

	if (MatrixMaxNrSrcDsts < 2)
	{
		DnpapMessage(DMC_WARNING, DMATRIX_MAX, "dmatrix: beholder.dmatrix.maxnrsrcdsts < 2, setting it to 2");
		MatrixMaxNrSrcDsts = 2;
	}
	return TRUE;
}

BOOLEAN DnpapMatrixCInit(DNPAP_MATRIX_CONTROL * matrixcontrol)
{
	LONG source[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};

	memcpy(matrixcontrol->Source, source, sizeof(source));
	matrixcontrol->SourceLen = sizeof(source) / sizeof(source[0]);
	matrixcontrol->Channel = FALSE;
    matrixcontrol->ChannelCallback = NULL;
    matrixcontrol->AddressLen = 0;
	matrixcontrol->TableSize = 0;
	matrixcontrol->LastDeleteTime = 0;
	matrixcontrol->Owner[0] = '\0';
	matrixcontrol->OwnerLen = 0;
	matrixcontrol->Status = SNMP_INVALID;

	if ((matrixcontrol->Iface =
		 MacIfaceGet((WORD) matrixcontrol->Source[matrixcontrol->SourceLen - 1])) == NULL)
	{
		DnpapMessage(DMC_ERROR, DMATRIX_NETINIT, "dmatrix: network initialisation failed");
		return (FALSE);
	}

	matrixcontrol->SrcTable = NULL;
	matrixcontrol->DstTable = NULL;

	matrixcontrol->LRUList = NULL;
	matrixcontrol->LRULast = NULL;

	matrixcontrol->SrcList = NULL;
	matrixcontrol->DstList = NULL;

	BooleanSetAllTrue(matrixcontrol->ObjectSet);
	BooleanSetFalse(matrixcontrol->ObjectSet, MATRIX_BOOLEAN_DATASOURCE);
	BooleanSetFalse(matrixcontrol->ObjectSet, MATRIX_BOOLEAN_OWNER);

	return TRUE;
}

BOOLEAN DnpapMatrixCStart(DNPAP_MATRIX_CONTROL * matrixcontrol)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(matrixcontrol->ObjectSet))
		return FALSE;

	if ((matrixcontrol->SrcTable = NewHash(MatrixMaxNrSrcDsts/2+11, NULL)) == NULL)
	{
		DnpapMessage(DMC_ERROR, DMATRIX_HASHERR, "dmatrix: can not create source hashtable");
		return FALSE;
	}

	if ((matrixcontrol->DstTable = NewHash(MatrixMaxNrSrcDsts/2+11, NULL)) == NULL)
	{
		DnpapMessage(DMC_ERROR, DMATRIX_HASHERR, "dmatrix: can not create destination hashtable");
		return FALSE;
	}

	if (matrixcontrol->Channel == TRUE)
    {
        if ((matrixcontrol->ChannelCallback = ChannelRegisterSink(matrixcontrol->Source[matrixcontrol->SourceLen - 1],
                                                                  DnpapMatrixChannelPkt, DnpapMatrixDeRegister, matrixcontrol)) == NULL)
		{
			DnpapMessage(DMC_ERROR, DMATRIX_CHANNELERR, "dmatrix: failed to register to channel");
			return FALSE;
		}
    }
    else
	{
		matrixcontrol->Coll.Rcve = MatrixCallback;
		matrixcontrol->Coll.specific = matrixcontrol;
		if (!MacCollRegister(&(matrixcontrol->Coll)))
		{
			DnpapMessage(DMC_ERROR, DMATRIX_NETERR, "dmatrix: network initialisation failed");
			return FALSE;
		}
	}

	return TRUE;
}

BOOLEAN DnpapMatrixCStop(DNPAP_MATRIX_CONTROL * matrixcontrol)
{
    if (matrixcontrol->Channel == TRUE)
	{
        ChannelDeRegisterSink(matrixcontrol->Source[matrixcontrol->SourceLen - 1],
                              matrixcontrol->ChannelCallback);
		matrixcontrol->ChannelCallback = NULL;
	}
    else
        MacCollRemove(&(matrixcontrol->Coll));

	MatrixDelConns(matrixcontrol);

	DelHash(matrixcontrol->SrcTable);
	matrixcontrol->SrcTable = NULL;
	DelHash(matrixcontrol->DstTable);
	matrixcontrol->DstTable = NULL;

	matrixcontrol->SrcTable = NULL;
	matrixcontrol->DstTable = NULL;

	matrixcontrol->LRUList = NULL;
	matrixcontrol->LRULast = NULL;

	matrixcontrol->SrcList = NULL;
	matrixcontrol->DstList = NULL;

	matrixcontrol->TableSize = 0;
	
    matrixcontrol->AddressLen = 0;
	matrixcontrol->LastDeleteTime = 0;

	return TRUE;
}

BOOLEAN MatrixCallback(MAC_COLL *collector, PROT_PKT *pkt)
{
DNPAP_MATRIX_CONTROL *matrixcontrol = collector->specific;
PROT_OBJ Interface = {0, {1,2}};

    if (matrixcontrol->Status == SNMP_VALID &&
        ProtGetField(pkt,&Interface) == TRUE &&
        Interface.Syntax.LngInt == matrixcontrol->Source[matrixcontrol->SourceLen-1])
        MatrixHandlePkt(matrixcontrol, matrixcontrol->Iface->type, pkt);
		
	return TRUE;
}


VOID DnpapMatrixChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink)
{
	DNPAP_MATRIX_CONTROL *matrixcontrol = (DNPAP_MATRIX_CONTROL *) sink;

	if (matrixcontrol->Status == SNMP_VALID && matrixcontrol->Channel == TRUE &&
		matrixcontrol->Source[matrixcontrol->SourceLen - 1] == channelIndex)
		MatrixHandlePkt(matrixcontrol, type, pkt);
}


VOID MatrixHandlePkt(DNPAP_MATRIX_CONTROL *matrixcontrol, int type, PROT_PKT *pkt)
{
	PROT_OBJ EthSize =	{0, {1, 4}};
	PROT_OBJ EthDst =	{1, {2, 1}};
	PROT_OBJ EthSrc =	{1, {2, 2}};
	BYTE *s, *d;
	LWORD size;
	WORD addrlen;
	Host *src;
	Host *dst;
	ConnInfo *conndst;
	ConnInfo *connsrc;
	ConnStats *stats;
	LRUEntry *entry, *oldentry;
	BOOLEAN newconn;

	s = NULL;
	d = NULL;
									  
	switch (type)
	{
	case MAC_TYPE_ETHERNET_CSMACD:
	case MAC_TYPE_88023_CSMACD:
    	if (ProtGetField(pkt,&EthSize) == FALSE)
			return;
	    size = EthSize.Syntax.LngUns + 4L;
        if (ProtGetField(pkt,&EthSrc) == FALSE)
			return;
		s = EthSrc.Syntax.BufChr;
	    if (ProtGetField(pkt,&EthDst) == FALSE)
			return;
		d = EthDst.Syntax.BufChr;
        if (EthSrc.SyntaxLen != ETH_SIZE_ADDR)
			s = NULL;
		if (EthDst.SyntaxLen != ETH_SIZE_ADDR)
			d = NULL;
		addrlen = ETH_SIZE_ADDR;
		if (matrixcontrol->AddressLen == 0)
			matrixcontrol->AddressLen = addrlen;
		else
		if (matrixcontrol->AddressLen != addrlen)
		{
			DnpapMessage(DMC_FATAL, DMATRIX_INCONSISTENT, "dmatrix: inconsistent address length encountered");
			DnpapExit(1);
		}
		break;
	default:
		return;
	}

	newconn = FALSE;

	if ((src = HashSearch(matrixcontrol->SrcTable, s, addrlen)) == NULL)
	{
		if ((src = DnpapMalloc(sizeof(Host))) != NULL)
		{
			memcpy(src->Addr, s, addrlen);
			src->LexList = NULL;
			src->TopList = NULL;
			src->TopLast = NULL;
			src->NumberConnect = 0;
			src->Next = NULL;
			src->Prev = NULL;

			if (HashAdd(matrixcontrol->SrcTable, src->Addr, addrlen, src) == NULL)
			{
				DnpapMessage(DMC_WARNING, DMATRIX_MAX, "dmatrix: source could not be added to the table");
				DnpapFree(src);
				src = NULL;
				return;
			}
			else
			{
				MatrixAddHost(matrixcontrol, src, TRUE);
			}

			newconn = TRUE;
		}
	}

	if ((dst = HashSearch(matrixcontrol->DstTable, d, addrlen)) == NULL)
	{
		if ((dst = DnpapMalloc(sizeof(Host))) != NULL)
		{
			memcpy(dst->Addr, d, addrlen);
			dst->LexList = NULL;
			dst->TopList = NULL;
			dst->TopLast = NULL;
			dst->NumberConnect = 0;
			dst->Next = NULL;
			dst->Prev = NULL;

			if (HashAdd(matrixcontrol->DstTable, dst->Addr, addrlen, dst) == NULL)
			{
				DnpapMessage(DMC_WARNING, DMATRIX_MAX, "dmatrix: destination could not be added to the table");
				DnpapFree(dst);
				dst = NULL;
				return;
			}
			else
			{
				MatrixAddHost(matrixcontrol, dst, FALSE);
			}

			newconn = TRUE;
		}
	}

	conndst = ListConnSearch(src, dst->Addr, matrixcontrol->AddressLen);

	if (conndst == NULL)
	{
		if ((conndst = DnpapMalloc(sizeof(ConnInfo))) != NULL)
		{
			conndst->ConnHost = dst;
			conndst->Stats = NULL;
			conndst->LexNext = NULL;
			conndst->LexPrev = NULL;
			conndst->TopNext = NULL;
			conndst->TopPrev = NULL;
			conndst->Entry = NULL;

			ListAddLex(src, conndst, matrixcontrol->AddressLen);
			ListAddTop(src, conndst);

   			src->NumberConnect++;
               
			newconn = TRUE;
		}
		else
		{
			DnpapMessage(DMC_WARNING, DMATRIX_MAX, "dmatrix: not enough memory");
			DnpapFree(conndst);
			return;
		}
	}

	connsrc = ListConnSearch(dst, src->Addr, matrixcontrol->AddressLen);

	if (connsrc == NULL)
	{
		if ((connsrc = DnpapMalloc(sizeof(ConnInfo))) != NULL)
		{
			connsrc->ConnHost = src;
			connsrc->Stats = NULL;
			connsrc->LexNext = NULL;
			connsrc->LexPrev = NULL;
			connsrc->TopNext = NULL;
			connsrc->TopPrev = NULL;
			connsrc->Entry = NULL;

			ListAddLex(dst, connsrc, matrixcontrol->AddressLen);
			ListAddTop(dst, connsrc);

    		dst->NumberConnect++;
           
			newconn = TRUE;
		}
		else
		{
			DnpapMessage(DMC_WARNING, DMATRIX_MAX, "dmatrix: not enough memory");
               ListDelConn(src, conndst);
			DnpapFree(connsrc);
			return;
		}
	}

	if (newconn == TRUE)
	{
		matrixcontrol->TableSize++;

		if ((entry = DnpapMalloc(sizeof(LRUEntry))) == NULL)
		{
			DnpapMessage(DMC_WARNING, DMATRIX_MAX, "dmatrix: LRU entry could not be added to the table");
		}
		else
		{
			entry->ConnSrc = connsrc;
			entry->ConnDst = conndst;
           
               connsrc->Entry = entry;
               conndst->Entry = entry;
               
			oldentry = MatrixAddLRUList(matrixcontrol, entry);

			if (matrixcontrol->TableSize > MatrixMaxNrSrcDsts)
			{
				MatrixRemoveLRUList(matrixcontrol, oldentry);
				MatrixDelConn(matrixcontrol, oldentry);

				DnpapFree(oldentry);

				matrixcontrol->LastDeleteTime = SysTime();
				matrixcontrol->TableSize--;
			}
		}

		if ((stats = DnpapMalloc(sizeof(ConnStats))) == NULL)
		{
			DnpapMessage(DMC_ERROR, DMATRIX_STATSERR, "dmatrix: could not allocate statistics entry");

			MatrixRemoveLRUList(matrixcontrol, entry);
			MatrixDelConn(matrixcontrol, entry);

			DnpapFree(entry);

			matrixcontrol->TableSize--;

			return;
		}
		else
		{
			stats->Pkts = 0;
			stats->Octets = 0;
			stats->Errors = 0;

			if (connsrc->Stats != NULL || conndst->Stats != NULL)
			{
				DnpapMessage(DMC_FATAL, DMATRIX_STATSERR, "dmatrix: error in internal tables");
				DnpapExit(DMATRIX_STATSERR);
			}

			connsrc->Stats = stats;
			conndst->Stats = stats;
		}
	}
	else
	{
		entry = connsrc->Entry;
		if (entry != conndst->Entry)
		{
			DnpapMessage(DMC_FATAL, DMATRIX_LRUERR, "dmatrix: error in internal tables");
			DnpapExit(DMATRIX_LRUERR);
		}

		stats = connsrc->Stats;
		if (stats != conndst->Stats)
		{
			DnpapMessage(DMC_FATAL, DMATRIX_STATSERR, "dmatrix: error in internal tables");
			DnpapExit(DMATRIX_STATSERR);
		}
	}

	if (stats != NULL)
	{
		stats->Pkts++;
		stats->Octets += size;

		ListUpdateTop(dst, connsrc);
		ListUpdateTop(src, conndst);

		MatrixUpdateLRUList(matrixcontrol, entry);
	}

#ifdef DEBUG
    ListCheckConns(src);
    ListCheckConns(dst);
#endif
    
	return;
}

VOID DnpapMatrixDeRegister(ULONG channelIndex, VOID *sink)
{
DNPAP_MATRIX_CONTROL *data = (DNPAP_MATRIX_CONTROL *) sink;

    if (data->Channel == TRUE && data->Source[data->SourceLen - 1] == channelIndex)
    {
        ChannelDeRegisterSink(data->Source[data->SourceLen - 1],
                              data->ChannelCallback);
		data->ChannelCallback = NULL;
    }
}

LRUEntry *MatrixAddLRUList(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * entry)
{
	if (matrixcontrol->LRUList == NULL)
	{
		matrixcontrol->LRUList = entry;
		matrixcontrol->LRULast = entry;
		entry->LRUPrev = NULL;
		entry->LRUNext = NULL;

		return NULL;
	}

	entry->LRUNext = matrixcontrol->LRUList;
	entry->LRUPrev = NULL;
	matrixcontrol->LRUList->LRUPrev = entry;
	matrixcontrol->LRUList = entry;

	return matrixcontrol->LRULast;
}

BOOLEAN MatrixRemoveLRUList(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * entry)
{
	if (entry->LRUPrev != NULL)
		entry->LRUPrev->LRUNext = entry->LRUNext;
	else
		matrixcontrol->LRUList = entry->LRUNext;
	if (entry->LRUNext != NULL)
		entry->LRUNext->LRUPrev = entry->LRUPrev;
	else
		matrixcontrol->LRULast = entry->LRUPrev;

	return TRUE;
}

BOOLEAN MatrixUpdateLRUList(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * entry)
{
	if (matrixcontrol->TableSize > 1 && entry != NULL)
	{
		MatrixRemoveLRUList(matrixcontrol, entry);
		MatrixAddLRUList(matrixcontrol, entry);
	}
	return TRUE;
}

BOOLEAN MatrixDelConns(DNPAP_MATRIX_CONTROL * matrixcontrol)
{
	LRUEntry *t;

    /*  take the first entry each time  */
	for (t = matrixcontrol->LRUList; t != NULL; t = matrixcontrol->LRUList)
	{
		MatrixRemoveLRUList(matrixcontrol, t);
		MatrixDelConn(matrixcontrol, t);
		DnpapFree(t);
	}

	return TRUE;
}

BOOLEAN MatrixDelConn(DNPAP_MATRIX_CONTROL * matrixcontrol, LRUEntry * entry)
{
	Host *srchost, *dsthost;
	ConnInfo *connsrc, *conndst;

	connsrc = entry->ConnSrc;
	conndst = entry->ConnDst;

	srchost = connsrc->ConnHost;
	dsthost = conndst->ConnHost;

	if (conndst->Stats != connsrc->Stats)
	{
		DnpapMessage(DMC_FATAL, DMATRIX_STATSERR, "dmatrix: error in internal tables");
		DnpapExit(DMATRIX_STATSERR);
	}

	DnpapFree(connsrc->Stats);

	ListDelConn(srchost, conndst);
	ListDelConn(dsthost, connsrc);

	if (srchost->LexList == 0)
		MatrixDelHost(matrixcontrol, srchost, TRUE);
	if (dsthost->LexList == 0)
		MatrixDelHost(matrixcontrol, dsthost, FALSE);

	return TRUE;
}

VOID ListDelConn(Host * host, ConnInfo * conn)
{
	if (conn->LexPrev == NULL)
		host->LexList = conn->LexNext;
	else
		conn->LexPrev->LexNext = conn->LexNext;
	if (conn->LexNext != NULL)
		conn->LexNext->LexPrev = conn->LexPrev;

	if (conn->TopPrev == NULL)
		host->TopList = conn->TopNext;
	else
		conn->TopPrev->TopNext = conn->TopNext;
	if (conn->TopNext != NULL)
		conn->TopNext->TopPrev = conn->TopPrev;
	else
		host->TopLast = conn->TopPrev;

	DnpapFree(conn);

	host->NumberConnect--;
    
    if (host->NumberConnect == 0 && 
        (host->LexList != NULL || host->TopList != NULL))
    {
		DnpapMessage(DMC_FATAL, DMATRIX_CONNERR, "dmatrix: error in internal tables");
		DnpapExit(DMATRIX_CONNERR);
    }
}

VOID MatrixDelHost(DNPAP_MATRIX_CONTROL * matrixcontrol, Host * host, BOOLEAN source)
{
	Host **list;
	HashTable *table;

	if (source == TRUE)
	{
		list = &matrixcontrol->SrcList;
		table = matrixcontrol->SrcTable;
	}
	else
	{
		list = &matrixcontrol->DstList;
		table = matrixcontrol->DstTable;
	}

	if (host->Prev == NULL)
		*list = host->Next;
	else
		host->Prev->Next = host->Next;
	if (host->Next != NULL)
		host->Next->Prev = host->Prev;

	HashRemove(table, host->Addr, matrixcontrol->AddressLen);

	DnpapFree(host);
}

BOOLEAN MatrixAddHost(DNPAP_MATRIX_CONTROL * matrixcontrol, Host * host, BOOLEAN source)
{
	Host *h, *p, **list;

	if (source == TRUE)
		list = &matrixcontrol->SrcList;
	else
		list = &matrixcontrol->DstList;

	if (*list == NULL)
	{
		*list = host;
		host->Prev = NULL;
		host->Next = NULL;
		return TRUE;
	}

	for (h = *list, p = NULL; h != NULL; p = h, h = h->Next)
		if (memcmp(h->Addr, host->Addr, matrixcontrol->AddressLen) >= 0)
			break;

	host->Prev = p;
	host->Next = h;

	if (h != NULL)
		h->Prev = host;
	if (p != NULL)
		p->Next = host;
	else
		*list = host;

	return TRUE;
}

BOOLEAN ListAddLex(Host * host, ConnInfo * conn, WORD addrlen)
{
	ConnInfo *s, *t;

	if (host->LexList == NULL)
	{
		host->LexList = conn;
		conn->LexNext = NULL;
		conn->LexPrev = NULL;
		return TRUE;
	}

	for (s = host->LexList, t = NULL; s != NULL; t = s, s = s->LexNext)
		if (memcmp(s->ConnHost->Addr, conn->ConnHost->Addr, addrlen) >= 0)
			break;

	conn->LexPrev = t;
	conn->LexNext = s;

	if (s != NULL)
		s->LexPrev = conn;
	if (t != NULL)
		t->LexNext = conn;
	else
		host->LexList = conn;

	return TRUE;
}

BOOLEAN ListAddTop(Host * host, ConnInfo * conn)
{
	if (host->TopLast == NULL)
	{
		host->TopList = conn;
		host->TopLast = conn;
		conn->TopNext = NULL;
		conn->TopPrev = NULL;
		return TRUE;
	}

	conn->TopPrev = host->TopLast;
	conn->TopNext = NULL;
	host->TopLast->TopNext = conn;
	host->TopLast = conn;

	return TRUE;
}

BOOLEAN ListUpdateTop(Host * host, ConnInfo * conn)
{
	ConnInfo *a;

	if (host->TopList == conn)
		return TRUE;

	for (a = conn->TopPrev; a != NULL && a->Stats->Pkts <= conn->Stats->Pkts; a = a->TopPrev)
		;

	if (a == conn->TopPrev)
		return TRUE;

	conn->TopPrev->TopNext = conn->TopNext;
	if (conn->TopNext != NULL)
		conn->TopNext->TopPrev = conn->TopPrev;
	else
		host->TopLast = conn->TopPrev;

	conn->TopPrev = a;
	if (a == NULL)
	{
		conn->TopNext = host->TopList;
		host->TopList->TopPrev = conn;
		host->TopList = conn;
	}
	else
	{
		conn->TopNext = a->TopNext;
		a->TopNext->TopPrev = conn;
		a->TopNext = conn;
	}

	return TRUE;
}

Host *DnpapMatrixSearchHost(DNPAP_MATRIX_CONTROL * matrixcontrol, BYTE * addr, WORD len, BOOLEAN source)
{
	Host *a;

	if (source == TRUE)
		a = matrixcontrol->SrcList;
	else
		a = matrixcontrol->DstList;

	for (; a != NULL; a = a->Next)
	{
		if (memcmp(a->Addr, addr, len) >= 0)
			break;
	}

	return a;
}

ConnInfo *DnpapMatrixSearchHostObj(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN source, Host ** host, INT * rank)
{
	HashTable *table;
	BYTE addr[HOST_SIZE_ADDR], other[HOST_SIZE_ADDR];
	INT i, len;
	ConnInfo *conn;
	WORD addrlen = matrixcontrol->AddressLen;

	len = obj->IdLen - idlen - 3;

	if (len < 0)
		len = 0;
                                 
    if (len != 2 * addrlen)
        return NULL;

	if (obj->Id[idlen+1] != addrlen || obj->Id[idlen+1+addrlen+1] != addrlen)
        return NULL;
        
	for (i = 0; i < addrlen; i++)
	{
		addr[i] = obj->Id[idlen + 2 + i];
		other[i] = obj->Id[idlen + 3 + addrlen + i];
	}

	if (source == TRUE)
		table = matrixcontrol->SrcTable;
	else
		table = matrixcontrol->DstTable;

	if ((*host = HashSearch(table, addr, addrlen)) == NULL)
        return NULL;

	*rank = DMatrixListSearchRank(*host, other, &conn, matrixcontrol->AddressLen);
	/* conn == NULL als niet gevonden */

	return conn;
}

ConnInfo *DnpapMatrixSearchTopObj(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN source, Host ** host, INT * rank)
{
	HashTable *table;
	BYTE addr[HOST_SIZE_ADDR];
	INT i, len;
	WORD addrlen = matrixcontrol->AddressLen;

	len = obj->IdLen - idlen - 2;

	if (len < 0)
		len = 0;
                                 
    if (len != addrlen + 1)
        return NULL;
        
	if (obj->Id[idlen+1] != addrlen)
        return NULL;
        
	for (i = 0; i < addrlen; i++)
		addr[i] = obj->Id[idlen + 2 + i];
	*rank = obj->Id[idlen + 2 + addrlen];

	if (source == TRUE)
		table = matrixcontrol->SrcTable;
	else
		table = matrixcontrol->DstTable;

	if ((*host = HashSearch(table, addr, addrlen)) == NULL)
        return NULL;
        
	return DMatrixListSearchTop(*host, *rank);
}

ConnInfo *DMatrixListSearchAddr(Host * host, BYTE * addr, WORD len)
{
	ConnInfo *b;

	for (b = host->LexList; b != NULL; b = b->LexNext)
	{
		if (memcmp(b->ConnHost->Addr, addr, len) >= 0)
			break;
	}

	return b;
}

ConnInfo *DMatrixListSearchTop(Host * host, INT rank)
{
	ConnInfo *a;
	INT i = 1;

	for (a = host->TopList; a != NULL; a = a->TopNext, i++)
		if (i == rank)
			return a;

	return NULL;
}

ConnInfo *ListConnSearch(Host * host, BYTE * addr, WORD addrlen)
{
	ConnInfo *a;

	for (a = host->LexList; a != NULL; a = a->LexNext)
		if (memcmp(a->ConnHost->Addr, addr, addrlen) == 0)
			break;

	return a;
}

INT DMatrixListSearchRank(Host * host, BYTE * addr, ConnInfo ** conn, WORD addrlen)
{
	INT rank = 1;

	for (*conn = host->TopList; *conn != NULL; *conn = (*conn)->TopNext)
	{
		if (memcmp((*conn)->ConnHost->Addr, addr, addrlen) == 0)
			return rank;
		rank++;
	}

	return 0;
}
 

#ifdef DEBUG

static VOID ListCheckConns(Host * host)
{
INT n;
ConnInfo *c;

    for (c = host->LexList, n = 0; c != NULL; c = c->LexNext)
        n++;                     
    if (n != host->NumberConnect)
        DnpapMessage(DMC_ERROR, DMATRIX_CONSIST, "dmatrix: inconsistent list");

    for (c = host->TopList, n = 0; c != NULL; c = c->TopNext)
        n++;                     
    if (n != host->NumberConnect)
        DnpapMessage(DMC_ERROR, DMATRIX_CONSIST, "dmatrix: inconsistent list");
}

#endif
