/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <dnpap.h>
#include <memory.h>
#include <maxmem.h>
#include <dmath.h>
#include <mibsup.h>
#include <stdlib.h>
#include <sysmib.h>
#include <timer.h>
#include <dhost.h>

#include "dtopn_d.h"
#include "dtopn_e.h"
#include "dtopn_c.h"

IMPORT BOOLEAN BeholderStrictRMON;

static LONG TopNMaxNrEntries = 100;

static BOOLEAN DnpapTopNEntryReset(DNPAP_TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, DNPAP_HOST * host, WORD addrlen);
static BOOLEAN DnpapTopNEntryUpdate(DNPAP_TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, DNPAP_HOST * host, WORD addrlen, int flags, LWORD size);
static BOOLEAN DnpapTopNHostStop(DNPAP_TOPN_CONTROL *topncontrol, ULONG hostcontrolindex);

static VOID TopNTimerCallback(TIMER_DESCR * timer, ULONG now, VOID * param);

static VOID DelEntries(DNPAP_TOPN_CONTROL * topncontrol);


#define DTOPN_UPDATE_ENTRY(value) \
                    if (entry->PrevValue > host->value) \
                        entry->Rate += host->value + (ULONG_MAX - entry->PrevValue); \
                    else \
                        entry->Rate += host->value - entry->PrevValue; \
                    entry->PrevValue = host->value


BOOLEAN DnpapTopNCInit(DNPAP_TOPN_CONTROL * topncontrol)
{
	topncontrol->HostIndex = 1;
	topncontrol->HostCallback = NULL;
	topncontrol->RateBase = 1;
	topncontrol->TimeRemaining = 0;
	topncontrol->Duration = 0;
	topncontrol->RequestedSize = 10;
	topncontrol->GrantedSize = DnpapTopNGranted(topncontrol->RequestedSize);
	topncontrol->StartTime = 0;
	topncontrol->Owner[0] = '\0';
	topncontrol->OwnerLen = 0;
	topncontrol->Status = SNMP_INVALID;

	topncontrol->TableSize = 0;
	topncontrol->Table = NULL;

	topncontrol->TopNEntries = NULL;
	topncontrol->TopNLast = NULL;
	topncontrol->TopNArray = NULL;

	topncontrol->Timer = NULL;

	BooleanSetAllTrue(topncontrol->ObjectSet);
	BooleanSetFalse(topncontrol->ObjectSet, TOPN_BOOLEAN_DATASOURCE);
	BooleanSetFalse(topncontrol->ObjectSet, TOPN_BOOLEAN_OWNER);

	return TRUE;
}

BOOLEAN DnpapTopNCStart(DNPAP_TOPN_CONTROL * topncontrol)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(topncontrol->ObjectSet))
		return FALSE;

	if (topncontrol->TimeRemaining > 0)
	{
		topncontrol->HostCallback = DnpapHostRegisterTopN(topncontrol->HostIndex, DnpapHostTopNNotify, topncontrol);
		
		if ((topncontrol->Table = NewHash(TopNMaxNrEntries / 2 + 11, NULL)) == NULL)
		{
			DnpapMessage(DMC_ERROR, DTOPN_TIMER, "dtopN: can not create hashtable");
        	DnpapHostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
        	topncontrol->HostCallback = NULL;
			return FALSE;
		}

		if ((topncontrol->Timer = TimerRegister(TopNTimerCallback, topncontrol, 1000, TIMER_FOREVER, TIMER_TYPE_RECOVER)) == NULL)
		{
			DnpapMessage(DMC_ERROR, DTOPN_TIMER, "dtopN: can not register a timer");
        	DnpapHostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
        	topncontrol->HostCallback = NULL;
			return FALSE;
		}

		topncontrol->GrantedSize = DnpapTopNGranted(topncontrol->RequestedSize);
		
		topncontrol->AddressLen = 0;
        
		topncontrol->StartTime = SysTime();
	}

	return TRUE;
}

BOOLEAN DnpapTopNCStop(DNPAP_TOPN_CONTROL * topncontrol)
{
	DnpapHostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
	topncontrol->HostCallback = NULL;
	
	TimerRemove(topncontrol->Timer);
	topncontrol->Timer = NULL;
	
	DelEntries(topncontrol);

	DelHash(topncontrol->Table);
	topncontrol->TableSize = 0;
	topncontrol->Table = NULL;

	DnpapFree(topncontrol->TopNArray);
	topncontrol->TopNEntries = NULL;
	topncontrol->TopNLast = NULL;
	topncontrol->TopNArray = NULL;

	topncontrol->TimeRemaining = 0;
	topncontrol->Duration = 0;
	topncontrol->StartTime = 0;

	topncontrol->GrantedSize = DnpapTopNGranted(topncontrol->RequestedSize);
	
	return TRUE;
}

LONG DnpapTopNGranted(LONG requested)
{
FLOAT factor;

	factor = (FLOAT)LongMin(MemoryGetMaxChunk(), MemoryGetAvail()) / 
				(100*requested*sizeof(DNPAP_TOPN_ENTRY));
				
	if (factor >= 1.0)
		return requested;
	return factor*requested;
}

VOID TopNTimerCallback(TIMER_DESCR * timer, ULONG now, VOID * param)
{
	DNPAP_TOPN_CONTROL *topncontrol = param;

	topncontrol->TimeRemaining--;
	if (topncontrol->TimeRemaining <= 0)
	{
		if (DnpapFillEntryTable(topncontrol, topncontrol->GrantedSize) != TRUE)
			DnpapMessage(DMC_WARNING, DTOPN_FILLFAIL, "dtopN: creation of entry array failed");
    	DnpapHostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
	    topncontrol->HostCallback = NULL;
		TimerRemove(topncontrol->Timer);
		topncontrol->Timer = NULL;
	}
}

VOID DnpapHostTopNNotify(ULONG hostindex, int event, DNPAP_HOST *host, WORD addrlen, int flags, LWORD size, VOID *topn)
{
	switch (event)
	{
	case HOST_EVENT_RESET:
		DnpapTopNEntryReset(topn, hostindex, host, addrlen);
		break;
	case HOST_EVENT_UPDATE:
		DnpapTopNEntryUpdate(topn, hostindex, host, addrlen, flags, size);
		break;
	case HOST_EVENT_STOP:
		DnpapTopNHostStop(topn, hostindex);
		break;
	}
}

BOOLEAN DnpapTopNEntryReset(DNPAP_TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, DNPAP_HOST * host, WORD addrlen)
{
	DNPAP_TOPN_ENTRY *entry;

	if (topncontrol->Status == SNMP_VALID && topncontrol->HostIndex == hostcontrolindex &&
		topncontrol->TimeRemaining > 0)
	{
		if (topncontrol->AddressLen == 0)
			topncontrol->AddressLen = addrlen;
		else if (topncontrol->AddressLen != addrlen)
		{
			DnpapMessage(DMC_FATAL, DTOPN_INCONSISTENT, "dtopncontrol: inconsistent address length encountered");
			DnpapExit(1);
		}

		if ((entry = HashSearch(topncontrol->Table, host->Address, addrlen)) != NULL)
			entry->PrevValue = 0;
	}
	return TRUE;
}

BOOLEAN DnpapTopNEntryUpdate(DNPAP_TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, DNPAP_HOST * host, WORD addrlen, int flags, LWORD size)
{
	DNPAP_TOPN_ENTRY *entry;

	if (topncontrol->Status == SNMP_VALID && topncontrol->HostIndex == hostcontrolindex &&
		topncontrol->TimeRemaining > 0)
	{
		if (topncontrol->AddressLen == 0)
			topncontrol->AddressLen = addrlen;
		else if (topncontrol->AddressLen != addrlen)
		{
			DnpapMessage(DMC_FATAL, DTOPN_INCONSISTENT, "dtopncontrol: inconsistent address length encountered");
			DnpapExit(1);
		}

		switch (topncontrol->RateBase)
		{
		case DTOPN_RATE_INPKTS:
		case DTOPN_RATE_INOCTETS:
			if ((flags & HOST_FLAG_IN) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_OUTPKTS:
		case DTOPN_RATE_OUTOCTETS:
		case DTOPN_RATE_OUTERRORS:
		case DTOPN_RATE_OUTBROADCASTPKTS:
		case DTOPN_RATE_OUTMULTICASTPKTS:
			if ((flags & HOST_FLAG_OUT) == 0)
				return TRUE;
			break;
#ifdef NOT_HOST_GENERIC
		case DTOPN_RATE_INTCPPKTS:
		case DTOPN_RATE_INOTHERPKTS:
			if ((flags & HOST_FLAG_IN) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_OUTTCPPKTS:
		case DTOPN_RATE_OUTOTHERPKTS:
			if ((flags & HOST_FLAG_OUT) == 0)
				return TRUE;
			break;
#endif
		case DTOPN_RATE_INPKTS64:
		case DTOPN_RATE_INPKTS65TO127:
		case DTOPN_RATE_INPKTS128TO255:
		case DTOPN_RATE_INPKTS256TO511:
		case DTOPN_RATE_INPKTS512TO1023:
		case DTOPN_RATE_INPKTS1024TO1518:
			if ((flags & HOST_FLAG_IN) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_OUTPKTS64:
		case DTOPN_RATE_OUTPKTS65TO127:
		case DTOPN_RATE_OUTPKTS128TO255:
		case DTOPN_RATE_OUTPKTS256TO511:
		case DTOPN_RATE_OUTPKTS512TO1023:
		case DTOPN_RATE_OUTPKTS1024TO1518:
			if ((flags & HOST_FLAG_OUT) == 0)
				return TRUE;
			break;
		}
				
		switch (topncontrol->RateBase)
		{
		case DTOPN_RATE_OUTERRORS:
			if ((flags & HOST_FLAG_ERROR) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_OUTBROADCASTPKTS:
			if ((flags & HOST_FLAG_BROADCAST) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_OUTMULTICASTPKTS:
			if ((flags & HOST_FLAG_MULTICAST) == 0)
				return TRUE;
			break;
#ifdef NOT_HOST_GENERIC
		case DTOPN_RATE_INTCPPKTS:
		case DTOPN_RATE_OUTTCPPKTS:
			if ((flags & HOST_FLAG_TCP) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_INOTHERPKTS:
		case DTOPN_RATE_OUTOTHERPKTS:
			if ((flags & HOST_FLAG_OTHER) == 0)
				return TRUE;
			break;
#endif
		case DTOPN_RATE_INPKTS64:
		case DTOPN_RATE_OUTPKTS64:
			if ((flags & HOST_FLAG_64) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_INPKTS65TO127:
		case DTOPN_RATE_OUTPKTS65TO127:
			if ((flags & HOST_FLAG_65) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_INPKTS128TO255:
		case DTOPN_RATE_OUTPKTS128TO255:
			if ((flags & HOST_FLAG_128) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_INPKTS256TO511:
		case DTOPN_RATE_OUTPKTS256TO511:
			if ((flags & HOST_FLAG_256) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_INPKTS512TO1023:
		case DTOPN_RATE_OUTPKTS512TO1023:
			if ((flags & HOST_FLAG_512) == 0)
				return TRUE;
			break;
		case DTOPN_RATE_INPKTS1024TO1518:
		case DTOPN_RATE_OUTPKTS1024TO1518:
			if ((flags & HOST_FLAG_1024) == 0)
				return TRUE;
			break;
		}
				
		if ((entry = HashSearch(topncontrol->Table, host->Address, addrlen)) == NULL)
		{
			if ((entry = DnpapMalloc(sizeof(DNPAP_TOPN_ENTRY))) == NULL)
				return FALSE;
			else
			{
				memset(entry, 0, sizeof(DNPAP_TOPN_ENTRY));
				memcpy(entry->Address, host->Address, addrlen);

				entry->PrevValue = 0;
				switch (topncontrol->RateBase)
				{
				case DTOPN_RATE_INPKTS:
				case DTOPN_RATE_OUTPKTS:
				case DTOPN_RATE_OUTERRORS:
				case DTOPN_RATE_OUTBROADCASTPKTS:
				case DTOPN_RATE_OUTMULTICASTPKTS:
					entry->Rate = 1;
					break;
				case DTOPN_RATE_INOCTETS:
					entry->Rate = size;
					entry->PrevValue = host->InOctets;
					break;
				case DTOPN_RATE_OUTOCTETS:
					entry->Rate = size;
					entry->PrevValue = host->OutOctets;
					break;
#ifdef NOT_HOST_GENERIC
				case DTOPN_RATE_INTCPPKTS:
				case DTOPN_RATE_OUTTCPPKTS:
				case DTOPN_RATE_INOTHERPKTS:
				case DTOPN_RATE_OUTOTHERPKTS:
					entry->Rate = 1;
					break;
#endif
				case DTOPN_RATE_INPKTS64:
				case DTOPN_RATE_INPKTS65TO127:
				case DTOPN_RATE_INPKTS128TO255:
				case DTOPN_RATE_INPKTS256TO511:
				case DTOPN_RATE_INPKTS512TO1023:
				case DTOPN_RATE_INPKTS1024TO1518:
				case DTOPN_RATE_OUTPKTS64:
				case DTOPN_RATE_OUTPKTS65TO127:
				case DTOPN_RATE_OUTPKTS128TO255:
				case DTOPN_RATE_OUTPKTS256TO511:
				case DTOPN_RATE_OUTPKTS512TO1023:
				case DTOPN_RATE_OUTPKTS1024TO1518:
					entry->Rate = 1;
					break;
				}

				if (HashAdd(topncontrol->Table, entry->Address, addrlen, entry) == NULL)
				{
					DnpapFree(entry);
					return FALSE;
				}
				else
				{
					DnpapEntryAddList(topncontrol, entry);
					topncontrol->TableSize++;
				}
			}
		}
		else
		{
			switch (topncontrol->RateBase)
			{
			case DTOPN_RATE_INPKTS:
			case DTOPN_RATE_OUTPKTS:
			case DTOPN_RATE_OUTERRORS:
			case DTOPN_RATE_OUTBROADCASTPKTS:
			case DTOPN_RATE_OUTMULTICASTPKTS: 
				entry->Rate++;
				break;
			case DTOPN_RATE_INOCTETS:
				DTOPN_UPDATE_ENTRY(InOctets);
				break;
			case DTOPN_RATE_OUTOCTETS:
				DTOPN_UPDATE_ENTRY(OutOctets);
				break;
#ifdef NOT_HOST_GENERIC
			case DTOPN_RATE_INTCPPKTS:
			case DTOPN_RATE_OUTTCPPKTS:
			case DTOPN_RATE_INOTHERPKTS:
			case DTOPN_RATE_OUTOTHERPKTS:
				entry->Rate++;
				break;
#endif
			case DTOPN_RATE_INPKTS64:
			case DTOPN_RATE_INPKTS65TO127:
			case DTOPN_RATE_INPKTS128TO255:
			case DTOPN_RATE_INPKTS256TO511:
			case DTOPN_RATE_INPKTS512TO1023:
			case DTOPN_RATE_INPKTS1024TO1518:
			case DTOPN_RATE_OUTPKTS64:
			case DTOPN_RATE_OUTPKTS65TO127:
			case DTOPN_RATE_OUTPKTS128TO255:
			case DTOPN_RATE_OUTPKTS256TO511:
			case DTOPN_RATE_OUTPKTS512TO1023:
			case DTOPN_RATE_OUTPKTS1024TO1518:
				entry->Rate++;
				break;
			}

			DnpapEntryUpdateList(topncontrol, entry);
		}
	}
	return TRUE;
}

BOOLEAN DnpapTopNHostStop(DNPAP_TOPN_CONTROL *topncontrol, ULONG hostcontrolindex)
{
	if (topncontrol->Status == SNMP_VALID && topncontrol->HostIndex == hostcontrolindex)
	{
		if (DnpapTopNCStop(topncontrol) == FALSE)
			return FALSE;
	}
	return TRUE;
}

DNPAP_TOPN_ENTRY *DnpapTopNEntrySearch(DNPAP_TOPN_CONTROL * topncontrol, SNMP_OBJECT * obj, WORD idlen)
{
	if (topncontrol->Status == SNMP_VALID && topncontrol->TimeRemaining == 0 &&
		topncontrol->TopNArray != NULL)
		return topncontrol->TopNArray[obj->Id[idlen + 1] - 1];
	return NULL;
}

BOOLEAN DnpapFillEntryTable(DNPAP_TOPN_CONTROL * topncontrol, LONG reqsize)
{
	DNPAP_TOPN_ENTRY *entry;
	DNPAP_TOPN_ENTRY **array;
	LONG count;
	LONG arraysize = min(reqsize, topncontrol->TableSize);

	if (arraysize == 0 && topncontrol->Status == SNMP_VALID)
	{														 
		/*
		DnpapMessage(DMC_MESSAGE, DTOPN_ARRAY, "dtopN: table is empty");
		 */
		return TRUE;
	}

	if ((array = (DNPAP_TOPN_ENTRY **) DnpapMalloc(arraysize * sizeof(DNPAP_TOPN_ENTRY *))) == NULL)
	{
		DnpapMessage(DMC_ERROR, DTOPN_ARRAY, "dtopN: can not create entry array of size %ld", arraysize);
		return FALSE;
	}
	DnpapFree(topncontrol->TopNArray);
	memset(topncontrol->TopNArray = array, 0, (size_t) arraysize * sizeof(DNPAP_TOPN_ENTRY *));

	for (entry = topncontrol->TopNEntries, count = 0;
		 entry != NULL && count < arraysize;
		 entry = entry->Next, count++)
		topncontrol->TopNArray[count] = entry;

	return TRUE;
}

VOID DelEntries(DNPAP_TOPN_CONTROL * topncontrol)
{
	DNPAP_TOPN_ENTRY *entry1, *entry2;

	for (entry2 = topncontrol->TopNEntries; entry2 != NULL; entry2 = entry1)
	{
		entry1 = entry2->Next;
		DnpapFree(entry2);
		topncontrol->TableSize--;
	}
	topncontrol->TopNEntries = NULL;
	topncontrol->TopNLast = NULL;
}

BOOLEAN DnpapEntryAddList(DNPAP_TOPN_CONTROL * topncontrol, DNPAP_TOPN_ENTRY * entry)
{
	if (topncontrol->TopNEntries == NULL)
	{
		topncontrol->TopNEntries = entry;
		topncontrol->TopNLast = entry;
		entry->Prev = NULL;
		entry->Next = NULL;
		return TRUE;
	}

	entry->Prev = topncontrol->TopNLast;
	entry->Next = NULL;
	topncontrol->TopNLast->Next = entry;
	topncontrol->TopNLast = entry;

	return DnpapEntryUpdateList(topncontrol, entry);
}

BOOLEAN DnpapEntryUpdateList(DNPAP_TOPN_CONTROL * topncontrol, DNPAP_TOPN_ENTRY * entry)
{
	DNPAP_TOPN_ENTRY *p;
	BOOLEAN update = FALSE;

	if (topncontrol->TopNEntries == entry)
		return TRUE;

	for (p = entry->Prev; p != NULL; p = p->Prev)
	{
		if (p->Rate >= entry->Rate)
			break;
		update = TRUE;
	}

	if (update == FALSE)
		return TRUE;

	entry->Prev->Next = entry->Next;
	if (entry->Next != NULL)
		entry->Next->Prev = entry->Prev;
	else
		topncontrol->TopNLast = entry->Prev;

	if (p != NULL)
	{
		entry->Prev = p;
		entry->Next = p->Next;
		p->Next->Prev = entry;
		p->Next = entry;
	}
	else
	{
		entry->Prev = p;
		entry->Next = topncontrol->TopNEntries;
		topncontrol->TopNEntries->Prev = entry;
		topncontrol->TopNEntries = entry;
	}

	return TRUE;
}
