/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <dmath.h>
#include <agent.h>
#include <config.h>
#include <log.h>
#include <snmp.h>
#include <message.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysmib.h>
#include <trap.h>

#include "event_d.h"
#include "event_e.h"
#include "event_x.h"
#include "event_c.h"
#include "event.h"

#define CONTROLINDEXSIZE	1
#define EVENTINDEXSIZE		2

IMPORT BOOLEAN BeholderStrictRMON;

static CHAR MODULE[] = "event";

static MIB_LOCAL *EventInstance = NULL;

static LONG LogMaxNrLogs = 1000;

static BOOLEAN EventMibNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);
static BOOLEAN EventFirstLastLog(Event * event, LONG * firstlog, LONG * lastlog);
static BOOLEAN EventLogInRange(LONG firstlog, LONG lastlog, ULONG logindex);
static Log *EventLogEntry(SNMP_OBJECT * obj, WORD idlen, MIB_LOCAL * local);

static int PrintVarbind(SNMP_OBJECT * obj, CHAR * varbind);

BOOLEAN EventMInit(VOID)
{
	MessageConfig(EVENT_ERROR, "Event");
	ConfigGetLong("beholder.event.maxnrlogs", &LogMaxNrLogs);
	if (LogMaxNrLogs < 2)
	{
		DnpapMessage(DMC_WARNING, EVENT_MAX, "event: beholder.event.maxnrlogs < 2, setting it to 2");
		LogMaxNrLogs = 2;
	}

	return TRUE;
}

#define CONTROL_DATA    ((Event*)(*local)->Data)

BOOLEAN EventMibNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	BOOLEAN first = FALSE;
	LONG firstlog, lastlog;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case EVENTINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if (EventFirstLastLog(CONTROL_DATA, &firstlog, &lastlog) == TRUE &&
				EventLogInRange(firstlog, lastlog, obj->Id[idlen + 1]) == TRUE &&
				obj->Id[idlen + 1] != lastlog)
			{
				obj->Id[idlen + 1]++;
				if (obj->Id[idlen + 1] == 0)
					obj->Id[idlen + 1] = 1;
				obj->IdLen = idlen + EVENTINDEXSIZE;
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL && EventFirstLastLog(CONTROL_DATA, &firstlog, &lastlog) == FALSE)
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = firstlog;
		obj->IdLen = idlen + EVENTINDEXSIZE;
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}

	return TRUE;
}

BOOLEAN EventFirstLastLog(Event * event, LONG * firstlog, LONG * lastlog)
{
	if (event->NrLogs == 0)
	{
		*firstlog = *lastlog = 0;
		return FALSE;
	}
	if (event->LogBounced == TRUE && event->LogPos < event->NrLogs)
		*firstlog = event->Logs[event->LogPos]->Index;
	else
		*firstlog = event->Logs[0]->Index;
	*lastlog = event->Logs[(event->LogPos > 0) ? (event->LogPos - 1) : (event->NrLogs - 1)]->Index;
	return TRUE;
}

BOOLEAN EventLogInRange(LONG firstlog, LONG lastlog, ULONG logindex)
{
	if (firstlog <= lastlog)
		return (logindex >= firstlog && logindex <= lastlog);
	else
		return (logindex >= firstlog || logindex <= lastlog);
}

WORD EventIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, EventInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventDescription(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Event *data = NULL;

	if ((local = MibRmon(obj, EventInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Event *) local->Data;
		memcpy(obj->Syntax.BufChr, data->Description, obj->SyntaxLen = data->DescrLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Event *) local->Data;
		if (obj->SyntaxLen > EVENT_SIZE_DESCR)
			return SNMP_BADVALUE;
		memcpy(data->Description, obj->Syntax.BufChr, data->DescrLen = obj->SyntaxLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventType(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Event *data = NULL;

	if ((local = MibRmon(obj, EventInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Event *) local->Data;
		obj->Syntax.LngInt = data->Type;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Event *) local->Data;
		if (obj->Syntax.LngInt < EVENTTYPE_MIN || obj->Syntax.LngInt > EVENTTYPE_MAX)
			return SNMP_BADVALUE;
		data->Type = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventCommunity(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Event *data = NULL;

	if ((local = MibRmon(obj, EventInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Event *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, EVENT_BOOLEAN_COMMUNITY))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Community, obj->SyntaxLen = data->CommLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Event *) local->Data;
		if (obj->SyntaxLen > EVENT_SIZE_COMM)
			return SNMP_BADVALUE;
		memcpy(data->Community, obj->Syntax.BufChr, data->CommLen = obj->SyntaxLen);
		BooleanSetTrue(data->ObjectSet, EVENT_BOOLEAN_COMMUNITY);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventLastTimeSent(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Event *data = NULL;

	if ((local = MibRmon(obj, EventInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Event *) local->Data;
		obj->Syntax.LngUns = data->LastTimeSent;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Event *data = NULL;

	if ((local = MibRmon(obj, EventInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Event *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, EVENT_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, obj->SyntaxLen = data->OwnerLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Event *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, data->OwnerLen = obj->SyntaxLen);
		BooleanSetTrue(data->ObjectSet, EVENT_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Event *data;

	local = MibRmon(obj, EventInstance, idlen, 1);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Event *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 1) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Event *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &EventInstance, idlen, 1)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = NewEvent()) == NULL)
				{
					MibRemove(obj, &EventInstance, idlen, 1);
					return SNMP_GENERROR;
				}
				data = (Event *) local->Data;
				if (EventCInit(data) == TRUE)
				{
					data->Index = local->Index;
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, EVENT_CREATE, "event: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DelEvent(local->Data);
				MibRemove(obj, &EventInstance, idlen, 1);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (Event *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (EventCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, EVENT_ACTIVE, "event: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				else
					return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (EventCStop(data) == TRUE)
				{
					DelEvent(local->Data);
					DnpapMessage(DMC_MESSAGE, EVENT_DESTROY, "event: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &EventInstance, idlen, 1) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (EventCStop(data) == TRUE)
				{
					DelEvent(local->Data);
					DnpapMessage(DMC_MESSAGE, EVENT_DESTROY, "event: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &EventInstance, idlen, 1) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (EventCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventLogEventIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Log *log;

	if ((local = MibRmon(obj, EventInstance, idlen, 2)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 2) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if ((log = EventLogEntry(obj, idlen, local)) == NULL)
			return SNMP_NOSUCHNAME;
		return LogEventIndex(obj, idlen, log);
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventLogIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Log *log;

	if ((local = MibRmon(obj, EventInstance, idlen, 2)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 2) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if ((log = EventLogEntry(obj, idlen, local)) == NULL)
			return SNMP_NOSUCHNAME;
		return LogIndex(obj, idlen, log);
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventLogTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Log *log;

	if ((local = MibRmon(obj, EventInstance, idlen, 2)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 2) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if ((log = EventLogEntry(obj, idlen, local)) == NULL)
			return SNMP_NOSUCHNAME;
		return LogTime(obj, idlen, log);
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EventLogDescription(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Log *log;

	if ((local = MibRmon(obj, EventInstance, idlen, 2)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (EventMibNext(obj, &local, idlen, 2) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if ((log = EventLogEntry(obj, idlen, local)) == NULL)
			return SNMP_NOSUCHNAME;
		return LogDescription(obj, idlen, log);
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

Log *EventLogEntry(SNMP_OBJECT * obj, WORD idlen, MIB_LOCAL * local)
{
	LONG firstlog, lastlog, logindex, realindex;
	Event *event = (Event *) local->Data;

	logindex = obj->Id[idlen + 1];

	if (EventFirstLastLog(event, &firstlog, &lastlog) == FALSE || EventLogInRange(firstlog, lastlog, logindex) == FALSE)
		return NULL;

	if (logindex >= firstlog)
		realindex = event->LogPos + logindex - firstlog;
	else
		realindex = event->LogPos - lastlog + logindex - 1;

	if (realindex >= event->NrLogs)
		realindex -= event->NrLogs;
	else if (realindex < 0)
		realindex += event->NrLogs;

	return event->Logs[realindex];
}

PartyCallback *EventRegisterParty(ULONG eventIndex, EventPartyCallback notify, EventPartyDeRegisterCallback deregister, VOID * party)
{
	MIB_LOCAL *local;
	Event *data;
	PartyCallback *callback;

	for (local = EventInstance; local != NULL; local = local->Next)
	{
		data = (Event *) local->Data;
		if (data->Index <= eventIndex)
			continue;
		if (data->Index != eventIndex)
			return NULL;
		if ((callback = EventAddParty(data, notify, deregister, party)) != NULL)
			return callback;
		break;
	}

	return NULL;
}

BOOLEAN EventDeRegisterParty(ULONG eventIndex, PartyCallback * callback)
{
	MIB_LOCAL *local;
	Event *data;

	for (local = EventInstance; local != NULL; local = local->Next)
	{
		data = (Event *) local->Data;
		if (data->Index <= eventIndex)
			continue;
		if (data->Index != eventIndex)
			return FALSE;
		if (EventRemoveParty(data, callback) == TRUE)
			return TRUE;
		break;
	}

	return FALSE;
}

EventHook *EventRegisterGenerator(ULONG eventIndex, EventGeneratorDeRegisterCallback deregister, VOID * generator)
{
	EventHook *hook;
	MIB_LOCAL *local;
	Event *data;

	if (eventIndex == 0)
		return NULL;

	for (local = EventInstance; local != NULL; local = local->Next)
	{
		data = (Event *) local->Data;
		if (data->Index < eventIndex)
			continue;
		if (data->Index != eventIndex)
			return NULL;
		if ((hook = EventAddGenerator(data, deregister, generator)) != NULL)
			return hook;
		break;
	}

	return NULL;
}

BOOLEAN EventDeRegisterGenerator(ULONG eventIndex, EventHook * hook)
{
	MIB_LOCAL *local;
	Event *data;

	if (eventIndex == 0 || hook == NULL)
		return TRUE;

	for (local = EventInstance; local != NULL; local = local->Next)
	{
		data = (Event *) local->Data;
		if (data->Index < eventIndex)
			continue;
		if (data->Index != eventIndex)
			return FALSE;
		if (EventRemoveGenerator(data, hook) == TRUE)
			return TRUE;
		break;
	}

	return FALSE;
}

BOOLEAN EventGenerate(EventHook * hook, CHAR * generator, WORD code, WORD specific, SNMP_OBJECT * list, WORD listlen)
{
	Event *event;
	Log **ptr;
	ULONG systime;
	CHAR description[LOG_SIZE_DESCR];
	CHAR varbinds[LOG_SIZE_DESCR], varbind[LOG_SIZE_DESCR], *p;
	int i, l, len;

	if (hook == NULL)
		return FALSE;

	event = hook->Event;

	if (event->Status == SNMP_VALID)
	{
		systime = SysTime();

		EventNotify(event);

		if (event->Type == EVENT_LOG || event->Type == EVENT_LOG_TRAP)
		{
			if (event->LogBounced == TRUE || event->NrLogs == LogMaxNrLogs ||
				(ptr = ReNewLog(event->Logs, event->NrLogs + 1, event->NrLogs, LogMaxNrLogs)) == NULL)
			{
				event->LogBounced = TRUE;
				if (event->NrLogs == 0)
				{
					DnpapMessage(DMC_ERROR, EVENT_NOLOGS, "event: could not create any log entries");
					return FALSE;
				}
				if (event->LogPos == event->NrLogs)
					event->LogPos = 0;
			}
			else
			{
				event->Logs = ptr;
				event->NrLogs++;
			}
			event->Logs[event->LogPos]->Index = event->LogIndex++;
			if (event->Logs[event->LogPos]->Index == 0)
				event->Logs[event->LogPos]->Index = event->LogIndex++;
			event->Logs[event->LogPos]->Time = systime;
			sprintf(description, "%s %d %d %d%n",
					generator, code, specific, listlen, &len);
			varbinds[0] = '\0';
			for (i = 0, p = varbinds; i < listlen; i++)
			{
				l = PrintVarbind(list + i, varbind);
				if (i + 1 < listlen)
				{
					strcat(varbind, " ");
					l++;
				}
				if (len + l + 4 > LOG_SIZE_DESCR)	/* reserve 4 characters to print 'i' */
					break;
				sprintf(p, "%s", varbind);
				len += l;
				p += l;
			}
			sprintf(event->Logs[event->LogPos]->Description, "%s %d %s",
					description, i, varbinds);
			event->Logs[event->LogPos]->DescrLen = strlen(event->Logs[event->LogPos]->Description);
			event->LogPos++;
		}
		if (event->Type == EVENT_SNMP_TRAP || event->Type == EVENT_LOG_TRAP)
			if (AgentSendTrap(event->Community, event->CommLen, systime, code, specific, list, listlen) == FALSE)
				DnpapMessage(DMC_WARNING, EVENT_TRAPFAIL, "event: trap send failed");
		event->LastTimeSent = systime;

		return TRUE;
	}

	return FALSE;
}

int PrintVarbind(SNMP_OBJECT * obj, CHAR * varbind)
{
	int i, len;
	CHAR *p = varbind;

	for (i = 0; i < obj->IdLen; i++)
	{
		if (i + 1 < obj->IdLen)
			sprintf(p, "%lu.%n", obj->Id[i], &len);
		else
			sprintf(p, "%lu%n", obj->Id[i], &len);
		p += len;
	}

	sprintf(p, " %d %n", obj->Type, &len);
	p += len;

	switch (obj->Type)
	{
	case SNMP_INTEGER:
		sprintf(p, "%ld", obj->Syntax.LngInt);
		break;
	case SNMP_OCTETSTR:
		/* case SNMP_DISPLAYSTR: */
	case SNMP_OPAQUE:
		for (i = 0; i < obj->SyntaxLen; i++)
		{
			if (i + 1 < obj->SyntaxLen)
				sprintf(p, "%2.2x:%n", obj->Syntax.BufChr[i], &len);
			else
				sprintf(p, "%2.2x%n", obj->Syntax.BufChr[i], &len);
			p += len;
		}
		break;
	case SNMP_OBJECTID:
		for (i = 0; i < obj->SyntaxLen; i++)
		{
			if (i + 1 < obj->SyntaxLen)
				sprintf(p, "%lu.%n", obj->Syntax.BufInt[i], &len);
			else
				sprintf(p, "%lu%n", obj->Syntax.BufInt[i], &len);
			p += len;
		}
		break;
	case SNMP_IPADDR:
		sprintf(p, "%hu.%hu.%hu.%hu", (BYTE) ((obj->Syntax.LngUns & 0xff000000) >> 24),
				(BYTE) ((obj->Syntax.LngUns & 0x00ff0000) >> 16),
				(BYTE) ((obj->Syntax.LngUns & 0x0000ff00) >> 8),
				(BYTE) ((obj->Syntax.LngUns & 0x000000ff)));
		break;
	case SNMP_COUNTER:
	case SNMP_GAUGE:
	case SNMP_TIMETICKS:
		sprintf(p, "%lu", obj->Syntax.LngUns);
		break;
	}

	p += strlen(p);

	return p - varbind;
}
