/*
 * Copyright (c) 1992 by Sun Microsystems, Inc.
 */

#ifndef	_SYS_KSTAT_H
#define	_SYS_KSTAT_H

#pragma	ident	"@(#)kstat.h	1.4	92/09/04 SMI"

/*
 * Definition of general kernel statistics structures and /dev/kstat ioctls
 */

#include <sys/types.h>
#include <sys/t_lock.h>

#ifdef	__cplusplus
extern "C" {
#endif

typedef	longlong_t	kstime_t;	/* kstat timestamp */
typedef long		kid_t;		/* unique kstat id */

/*
 * Kernel statistics driver (/dev/kstat) ioctls
 */

#define	KSTAT_IOC_BASE		('K' << 8)

#define	KSTAT_IOC_CHAIN_ID	KSTAT_IOC_BASE | 0x01
#define	KSTAT_IOC_READ		KSTAT_IOC_BASE | 0x02
#define	KSTAT_IOC_WRITE		KSTAT_IOC_BASE | 0x03
#define	KSTAT_IOC_TIME		KSTAT_IOC_BASE | 0x04

/*
 * /dev/kstat ioctl usage (kd denotes /dev/kstat descriptor):
 *
 *	kcid = ioctl(kd, KSTAT_IOC_CHAIN_ID, NULL);
 *	kcid = ioctl(kd, KSTAT_IOC_READ, kstat_t *);
 *	kcid = ioctl(kd, KSTAT_IOC_WRITE, kstat_t *);
 *	kcid = ioctl(kd, KSTAT_IOC_TIME, kstime_t *);
 */

#define	KSTAT_STRLEN	31	/* 30 chars + NULL; must be 16 * n - 1 */

/*
 * The generic kstat header
 */

typedef struct kstat {
	/*
	 * Fields relevant to both kernel and user
	 */
	kstime_t	ks_crtime;	/* creation time (from kstat_time()) */
	struct kstat	*ks_next;	/* kstat chain linkage */
	kid_t		ks_kid;		/* unique kstat ID */
	char		ks_module[KSTAT_STRLEN]; /* creator module name */
	uchar_t		ks_resv;	/* reserved, currently just padding */
	int		ks_instance;	/* creator module's instance */
	char		ks_name[KSTAT_STRLEN]; /* kstat name */
	uchar_t		ks_type;	/* kstat data type */
	char		ks_class[KSTAT_STRLEN]; /* kstat class */
	uchar_t		ks_flags;	/* kstat flags */
	void		*ks_data;	/* kstat type-specific data */
	ulong_t		ks_ndata;	/* # of type-specific data records */
	ulong_t		ks_data_size;	/* total size of kstat data section */
	ulong_t		ks_resv2;	/* reserved for future use */
	/*
	 * Fields relevant to kernel only
	 */
	kmutex_t	*ks_lock;	/* protects this kstat's data */
	int		(*ks_update)(struct kstat *, int); /* dynamic update */
	void		*ks_private;	/* arbitrary client-private data */
	int		(*ks_copy)(struct kstat *, struct kstat *, int);
	ulong_t		ks_resv3;	/* reserved for future use */
} kstat_t;

/*
 * kstat structure
 *
 * Each kstat consists of a header section (a kstat_t) and a data section.
 * The data section may be any of the kstat types defined below.
 */

/*
 * kstat locking
 *
 * The system maintains a linked list of kstat structs, protected by
 * kstat_chain_lock.  kstat_chain_lock protects all additions to/deletions
 * from this list, as well as all changes to kstat headers.  kstat data
 * sections are *optionally* protected by the per-kstat ks_lock.  If
 * ks_lock is non-NULL, /dev/kstat will acquire this lock for all of its
 * operations on that kstat.  It is up to the driver to decide whether
 * guaranteeing consistent data to /dev/kstat is sufficiently important.
 * Note, however, that most statistic updates already occur under some
 * driver mutex, so if the driver sets ks_lock to point to that mutex,
 * then kstat locking is free.
 */

#if	defined(_KERNEL)

#define	KSTAT_ENTER(x)	if (x != NULL) mutex_enter(x)
#define	KSTAT_EXIT(x)	if (x != NULL) mutex_exit(x)

#endif	/* defined(_KERNEL) */

/*
 * kstat time
 *
 * All times associated with kstats (e.g. creation time, kstat_timer_t
 * and kstat_io_t timestamps, etc.) are 64-bit nanosecond values, as
 * returned by kstat_time().  The accuracy of kstat timestamps is machine
 * dependent, but the precision (units) is the same across all platforms.
 */

/*
 * kstat identity (KID)
 *
 * Each kstat is assigned a unique KID (kstat ID) when it is added to the
 * global kstat chain.  The KID is used as a cookie by /dev/kstat to
 * request information about the corresponding kstat.  There is also
 * an identity associated with the entire kstat chain, kstat_chain_id,
 * which is bumped each time a kstat is added or deleted.  /dev/kstat
 * clients can use the kstat chain ID to detect changes in the kstat chain
 * between ioctls.
 */

/*
 * kstat module, kstat instance
 *
 * ks_module and ks_instance contain the name and instance of the the
 * module that created the kstat.  In cases where there can only be one
 * instance, ks_instance is 0.  The "core" kernel (/kernel/unix) uses
 * "unix" as the module name.
 */

/*
 * kstat name
 *
 * ks_name gives a meaningful name to a kstat.  The full kstat namespace
 * is module.instance.name, so the name only need be unique within a
 * module.  kstat_create() will fail if you try to create a kstat with
 * an already-used (ks_module, ks_instance, ks_name) triplet.  Spaces are
 * allowed in kstat names, but strongly discouraged, since they hinder
 * awk-style processing at user level.
 */

/*
 * kstat type
 *
 * The kstat mechanism provides several flavors of kstat data, defined
 * below.  The "raw" kstat type is just treated as an array of bytes; you
 * can use this to export any kind of data you want.
 *
 * Some kstat types allow multiple data structures per kstat, e.g.
 * KSTAT_TYPE_NAMED; others do not.  This is a required part of the spec
 * for all kstat types.
 *
 * User-level tools should *not* rely on the #define KSTAT_NUM_TYPES.  To
 * get this information, read out the standard system kstat "kstat_types".
 */

#define	KSTAT_TYPE_RAW		0	/* can be anything */
					/* ks_ndata > 1 OK */
#define	KSTAT_TYPE_NAMED	1	/* name/value pair */
					/* ks_ndata > 1 OK */
#define	KSTAT_TYPE_INTR		2	/* interrupt statistics */
					/* ks_ndata == 1 */
#define	KSTAT_TYPE_IO		3	/* I/O statistics */
					/* ks_ndata == 1 */
#define	KSTAT_TYPE_TIMER	4	/* event timer */
					/* ks_ndata > 1 OK */

#define	KSTAT_NUM_TYPES		5

/*
 * kstat class
 *
 * Each kstat can be characterized as belonging to some broad class
 * of statistics, e.g. disk, tape, net, vm, streams, etc.  This field
 * can be used as a filter to extract related kstats.  The following
 * values are currently in use: disk, tape, net, controller, vm, kvm,
 * hat, streams, kstat, and misc.  (The kstat class encompasses things
 * like kstat_types.)
 */

/*
 * kstat flags
 *
 * Any combination of the following flags may be passed to kstat_create():
 *
 *	KSTAT_FLAG_VIRTUAL:
 *
 *		Tells kstat_create() not to allocate memory for the
 *		kstat data section; instead, you will set the ks_data
 *		field to point to the data you wish to export.  This
 *		provides a convenient way to export existing data
 *		structures.
 *
 *	KSTAT_FLAG_VAR_SIZE:
 *
 *		The size of the kstat you are creating will vary over time.
 *		For example, you may want to use the kstat mechanism to
 *		export a linked list.  N.B.: The kstat framework does not
 *		manage the data section, so all VAR_SIZE kstats must be VIRTUAL
 *		kstats.  See the section on "kstat copy" for further details.
 *
 *	KSTAT_FLAG_WRITABLE:
 *
 *		Makes the kstat's data section writable by root.
 *		The ks_copy routine (see below) does not need to check for
 *		this; permission checking is handled in the kstat driver.
 */

#define	KSTAT_FLAG_VIRTUAL		0x01
#define	KSTAT_FLAG_VAR_SIZE		0x02
#define	KSTAT_FLAG_WRITABLE		0x04

/*
 * Dynamic update support
 *
 * The kstat mechanism allows for an optional ks_update function to update
 * kstat data.  This is useful for drivers where the underlying device
 * keeps cheap hardware stats, but extraction is expensive.  Instead of
 * constantly keeping the kstat data section up to date, you can supply a
 * ks_update function which updates the kstat's data section on demand.
 * To take advantage of this feature, simply set the ks_update field before
 * calling kstat_install().
 *
 * The ks_update function, if supplied, must have the following structure:
 *
 *	int
 *	foo_kstat_update(kstat_t *ksp, int rw)
 *	{
 *		if (rw == KSTAT_WRITE) {
 *			... update the underlying stats from ksp->ks_data;
 *				return EACCES if you don't support this
 *		} else {
 *			... update ksp->ks_data from the underlying stats
 *		}
 *	}
 *
 * The ks_update return codes are: 0 for success, EACCES if you don't allow
 * KSTAT_WRITE, and EIO for any other type of error.
 *
 * In general, the ks_update function may need to refer to client-private
 * data; for example, it may need a pointer to the client's raw statistics.
 * The ks_private field is available for this purpose.  Its use is entirely
 * at the client's discretion.
 *
 * No kstat locking should be done inside the ks_update routine.  The caller
 * will already be holding the kstat's ks_lock (to ensure consistent data)
 * *and* the global kstat_chain_lock (to prevent the kstat from being
 * deleted by another thread).
 */

#define	KSTAT_READ	0
#define	KSTAT_WRITE	1

/*
 * Kstat copy
 *
 * Normally, /dev/kstat returns kstat data to the user via the routine
 * default_kstat_copy().  (The superuser can also write kstat data --
 * more on this below).  However, this routine only works for fixed-size
 * kstats.  If you define a variable-size kstat, or if you have other
 * special needs, you must set ks_copy to point to a routine to copy
 * data to/from userland.  The routine you supply must have the following
 * prototype:
 *
 *	int foo_kstat_copy(kstat_t *ksp, kstat_t *user_ksp, int rw);
 *
 * foo_kstat_copy() must have the following basic structure:
 *
 *	limit = user_ksp->ks_data_size;
 *	user_ksp->ks_data_size = foo_size;
 *	if (rw == KSTAT_WRITE) {
 *		if (limit != foo_size)
 *			return (ENOMEM);
 *		if (kstat_copyin(user_ksp->ks_data, ksp->ks_data, limit))
 *			return (EFAULT);
 *		return (0);
 *	}
 *	if (foo_size > limit)
 *		return (ENOMEM);
 *	if (kstat_copyout(ksp->ks_data, user_ksp->ks_data, foo_size))
 *		return (EFAULT);
 *	return (0);
 *
 * You should also set user_ksp->ks_ndata if that is meaningful, as in
 * the case of copying out a linked list of structures.  For example:
 *
 *	ulong_t	limit = user_ksp->ks_data_size;
 *	char *buf = (char *) user_ksp->ks_data;
 *	ulong_t	foo_ndata = 0;
 *	ulong_t	foo_size = 0;
 *
 *	if (rw == KSTAT_WRITE)
 *		return (EACCES);		... See below ...
 *
 *	for (foo = first_foo; foo; foo = foo->next) {
 *		foo_ndata++;
 *		foo_size += sizeof (struct foo);
 *	}
 *	user_ksp->ks_ndata = foo_ndata;
 *	user_ksp->ks_data_size = foo_size;
 *	if (foo_size > limit)
 *		return (ENOMEM);
 *	for (foo = first_foo; foo; foo = foo->next) {
 *		if (kstat_copyout((char *) foo, buf, sizeof (struct foo)))
 *			return (EFAULT);
 *		buf += sizeof (struct foo);
 *	}
 *	return (0);
 *
 * In the example above, we have decided that we don't want to allow
 * KSTAT_WRITE access, so we return EACCES if this is attempted.
 *
 * The key points here are:
 *
 *	(1) Data gets copied to/from the user-specified buffer,
 *		user_ksp->ks_data, depending whether rw is KSTAT_READ
 *		or KSTAT_WRITE, respectively.
 *	(2) On entry, user_ksp->ks_data_size is the size of the user's
 *		buffer; on exit, it must be set to current size of the kstat's
 *		data section.
 *	(3) *Always* update user_ksp->ks_data_size before returning from
 *		a KSTAT_READ, even if it's an error return.
 *	(4) *Never* attempt to copy in/out more bytes than the user's
 *		buffer size.
 *	(5) ks_copy return values are: 0 for success, ENOMEM if user's buffer
 *		is too small (for copyout) or if the buffer sizes don't
 *		match (for copyin), EFAULT for copyin/copyout errors,
 *		EACCES if you don't allow KSTAT_WRITE, and EIO for any other
 *		type of error.
 *
 * NOTE: you should always use the wrapper routines kstat_copyin() and
 * kstat_copyout() wherever you would normally use copyin() and copyout(),
 * respectively.
 */

/*
 * Named statistics.
 *
 * List of arbitrary name=value statistics.
 */

typedef struct kstat_named {
	char	name[KSTAT_STRLEN];	/* name of counter */
	uchar_t	data_type;		/* data type */
	union {
		char		c[16];	/* enough for 128-bit ints */
		long		l;
		ulong_t		ul;
		longlong_t	ll;
		u_longlong_t	ull;
		float		f;
		double		d;
	} value;			/* value of counter */
} kstat_named_t;

#define	KSTAT_DATA_CHAR		0
#define	KSTAT_DATA_LONG		1
#define	KSTAT_DATA_ULONG	2
#define	KSTAT_DATA_LONGLONG	3
#define	KSTAT_DATA_ULONGLONG	4
#define	KSTAT_DATA_FLOAT	5
#define	KSTAT_DATA_DOUBLE	6

#define	KSTAT_NAMED_PTR(kptr)	((kstat_named_t *) (kptr)->ks_data)

/*
 * Interrupt statistics.
 *
 * An interrupt is a hard interrupt (sourced from the hardware device
 * itself), a soft interrupt (induced by the system via the use of
 * some system interrupt source), a watchdog interrupt (induced by
 * a periodic timer call), spurious (an interrupt entry point was
 * entered but there was no interrupt condition to service),
 * or multiple service (an interrupt condition was detected and
 * serviced just prior to returning from any of the other types).
 *
 * Measurement of the spurious class of interrupts is useful for
 * autovectored devices in order to pinpoint any interrupt latency
 * problems in a particular system configuration.
 *
 * Devices that have more than one interrupt of the same
 * type should use multiple structures.
 */

#define	KSTAT_INTR_HARD			0
#define	KSTAT_INTR_SOFT			1
#define	KSTAT_INTR_WATCHDOG		2
#define	KSTAT_INTR_SPURIOUS		3
#define	KSTAT_INTR_MULTSVC		4

#define	KSTAT_NUM_INTRS			5

typedef struct kstat_intr {
	ulong_t		intrs[KSTAT_NUM_INTRS];	/* interrupt counters */
} kstat_intr_t;

#define	KSTAT_INTR_PTR(kptr)	((kstat_intr_t *) (kptr)->ks_data)

/*
 * I/O statistics.
 */

typedef struct kstat_io {

	/*
	 * Basic counters.
	 *
	 * The counters should be updated at the end of service
	 * (e.g., just prior to calling biodone()).
	 */

	u_longlong_t	nread;		/* number of bytes read */
	u_longlong_t	nwritten;	/* number of bytes written */
	ulong_t		reads;		/* number of read operations */
	ulong_t		writes;		/* number of write operations */

	/*
	 * Accumulated time and queue length statistics.
	 *
	 * Accumulated time statistics are kept as a running sum
	 * of "active" time.  Queue length statistics are kept as a
	 * running sum of the product of queue length and elasped time
	 * at that length -- i.e., a Riemann sum for queue length
	 * integrated against time.  (You can also think of active time
	 * as a Riemann sum, for the boolean function (queue_length > 0)
	 * integrated against time, or you can think of it as the
	 * Lebesgue measure of the set on which queue_length > 0.)
	 *
	 *		^
	 *		|			_________
	 *		8			| i4	|
	 *		|			|	|
	 *	Queue	6			|	|
	 *	Length	|	_________	|	|
	 *		4	| i2	|_______|	|
	 *		|	|	    i3		|
	 *		2_______|			|
	 *		|    i1				|
	 *		|_______________________________|
	 *		Time->	t1	t2	t3	t4
	 *
	 * At each change of state (entry or exit from the queue),
	 * we add the elapsed time (since the previous state change)
	 * to the active time if the queue length was non-zero during
	 * that interval; and we add the product of the elapsed time
	 * times the queue length to the running length*time sum.
	 *
	 * This method is generalizable to measuring residency
	 * in any defined system: instead of queue lengths, think
	 * of "outstanding RPC calls to server X".
	 *
	 * A large number of I/O subsystems have at least two basic
	 * "lists" of transactions they manage: one for transactions
	 * that have been accepted for processing but for which processing
	 * has yet to begin, and one for transactions which are actively
	 * being processed (but not done). For this reason, two cumulative
	 * time statistics are defined here: pre-service (wait) time,
	 * and service (run) time.
	 *
	 * All times are 64-bit nanoseconds (kstime_t), as returned by
	 * kstat_time().
	 *
	 * The units of cumulative busy time are accumulated nanoseconds.
	 * The units of cumulative length*time products are elapsed time
	 * times queue length.
	 *
	 * Updates to the fields below are performed implicitly by calls to
	 * these five functions:
	 *
	 *	kstat_waitq_enter()
	 *	kstat_waitq_exit()
	 *	kstat_runq_enter()
	 *	kstat_runq_exit()
	 *
	 *	kstat_waitq_to_runq()		(see below)
	 *	kstat_runq_back_to_waitq()	(see below)
	 *
	 * Since kstat_waitq_exit() is typically followed immediately
	 * by kstat_runq_enter(), there is a single kstat_waitq_to_runq()
	 * function which performs both operations.  This is a performance
	 * win since only one timestamp is required.
	 *
	 * In some instances, it may be necessary to move a request from
	 * the run queue back to the wait queue, e.g. for write throttling.
	 * For these situations, call kstat_runq_back_to_waitq().
	 *
	 * These fields should never be updated by any other means.
	 */

	kstime_t wtime;		/* cumulative wait (pre-service) time */
	kstime_t wlentime;	/* cumulative wait length*time product */
	kstime_t wlastupdate;	/* last time wait queue changed */
	kstime_t rtime;		/* cumulative run (service) time */
	kstime_t rlentime;	/* cumulative run length*time product */
	kstime_t rlastupdate;	/* last time run queue changed */

	ulong_t	wcnt;		/* count of elements in wait state */
	ulong_t	rcnt;		/* count of elements in run state */

} kstat_io_t;

#define	KSTAT_IO_PTR(kptr)	((kstat_io_t *) (kptr)->ks_data)

/*
 * Event timer statistics - cumulative elapsed time and number of events.
 *
 * Updates to these fields are performed implicitly by calls to
 * kstat_timer_start() and kstat_timer_stop().
 */

typedef struct kstat_timer {
	char		name[KSTAT_STRLEN];	/* event name */
	uchar_t		resv;			/* reserved */
	u_longlong_t	num_events;		/* number of events */
	kstime_t	elapsed_time;		/* cumulative elapsed time */
	kstime_t	min_time;		/* shortest event duration */
	kstime_t	max_time;		/* longest event duration */
	kstime_t	start_time;		/* previous event start time */
	kstime_t	stop_time;		/* previous event stop time */
} kstat_timer_t;

#define	KSTAT_TIMER_PTR(kptr)	((kstat_timer_t *) (kptr)->ks_data)

#if	defined(_KERNEL)

extern kstat_t	*kstat_chain;		/* kstat chain head */
extern kmutex_t	kstat_chain_lock;	/* protects the kstat chain */
extern kid_t	kstat_chain_id;		/* bumped at each state change */
extern int	kstat_clients;		/* non-zero if active kstat clients */

/*
 * Adding and deleting kstats.
 *
 * The typical sequence to add a kstat is:
 *
 *	ksp = kstat_create(module, instance, name, class, type, ndata, flags);
 *	if (ksp) {
 *		... client initialization, if necessary
 *		kstat_install(ksp);
 *	}
 *
 * In general, adding a kstat to the system is a three-step process:
 * system initialization, client initialization, and installation.
 *
 * Step 1: System Initialization
 *
 * kstat_create() performs system initialization.  kstat_create()
 * allocates memory for the entire kstat (header plus data), initializes
 * all header fields, initializes the data section to all zeroes, and puts
 * the kstat onto the system's kstat chain with a KID of -1.  The kstat
 * will not be visible to /dev/kstat until the KID is set to a non-negative
 * value by kstat_install().
 *
 * Step 2: Client Initialization
 *
 * The client performs any necessary initialization of the data section,
 * e.g. setting the name fields in a KSTAT_TYPE_NAMED.  The client also
 * has the option of assigning new values to any of the following header
 * fields: ks_lock, ks_update, ks_private, ks_copy.  Virual kstats also
 * set the ks_data field at this time.  If the client has a dynamic update
 * routine, it sets the ks_update field to the address of the update
 * function; similarly, the client may install an alternate ks_copy routine.
 * Optionally, the client can also set the ks_private field however
 * it sees fit.
 *
 * Step 3: Installation
 *
 * Once the kstat is completely initialized, kstat_install() assigns it
 * a unique non-negative KID, making it visible to the outside world.
 *
 * Removing a kstat from the system
 *
 * kstat_delete(ksp) removes ksp from the kstat chain and frees all
 * associated system resources.
 */

/*
 * kstat_t *
 * kstat_create(char *ks_module, int ks_instance, char *ks_name,
 *	char *ks_class, uchar_t ks_type, ulong_t ks_ndata, uchar_t ks_flags)
 *
 *	Purpose:
 *
 *		Create and initialize a new kstat
 *
 *	Arguments:
 *
 *		ks_module: the name of the caller's module, e.g. sd, esp, etc.
 *		The "core" kernel (/kernel/unix) uses the name "unix".
 *
 *		ks_instance: the caller's instance, as from ddi_get_instance().
 *		Modules which don't have a meaningful instance should use 0.
 *
 *		ks_name: pointer to a string that will uniquely identify this
 *		structure. Only KSTAT_STRLEN - 1 characters are significant.
 *
 *		ks_class: the general class that this kstat belongs to
 *
 *		ks_type: the type of kstat to allocate (raw, named,
 *		interrupt, i/o, timer)
 *
 *		ks_ndata: the number of type-specific data records to allocate
 *
 *		ks_flags: various flags for this kstat
 *
 *	Failure Return Value:
 *
 *		A null pointer due to the inability to create a new
 *		kstat structure or a (ks_module, ks_instance, ks_name)
 *		namespace collision.
 *
 *	Success Return Value:
 *
 *		A pointer to an initialized kstat structure, whose
 *		data field points to zero-initialized type-specific data.
 */

extern kstat_t *kstat_create(char *, int, char *, char *, uchar_t,
	ulong_t, uchar_t);

/*
 * void
 * kstat_install(kstat_t *ksp)
 *
 *	Puspose:
 *
 *		Add an initialized kstat to the system's kstat chain
 *
 *	Arguments:
 *
 *		ksp: pointer to an initialized kstat structure
 */

extern void kstat_install(kstat_t *);

/*
 * void
 * kstat_delete(kstat_t *ksp)
 *
 *	Puspose:
 *
 *		Delete a kstat from the system's kstat chain, and
 *		free all associated system resources
 *
 *	Arguments:
 *
 *		ksp: pointer to a currently installed kstat
 */

extern void kstat_delete(kstat_t *);

/*
 * kstat_t *
 * kstat_lookup_byname(char *ks_module, int ks_instance, char *ks_name)
 *
 *	Puspose:
 *
 *		Find the kstat with the specified (module, instance, name)
 */

extern kstat_t *kstat_lookup_byname(char *, int, char *);

/*
 * kstat_t *
 * kstat_lookup_bykid(kid_t ks_kid)
 *
 *	Puspose:
 *
 *		Find the kstat with the specified KID
 */

extern kstat_t *kstat_lookup_bykid(kid_t);

/*
 * kstime_t
 * kstat_time(void)
 *
 *	Puspose:
 *
 *		Returns the current kstat time.  This is a 64-bit
 *		nanosecond value.
 */

extern kstime_t kstat_time();

/*
 * void
 * kstat_set_string(char *dst, char *src)
 *
 *	Purpose:
 *
 *		Sets any string field of a kstat.  This is a convenience
 *		routine: it bzero()s the name field (dst), then copies in
 *		up to KSTAT_STRLEN - 1 characters of src.  This ensures
 *		that kstat strings are always null-terminated.
 */

extern void kstat_set_string(char *, char *);

/*
 * void
 * kstat_named_init(kstat_named_t *knp, char *name, uchar_t data_type)
 *
 * void
 * kstat_timer_init(kstat_timer_t *ktp, char *name)
 *
 *	Purpose:
 *
 *		Initialize type-specific data fields.
 */

extern void kstat_named_init(kstat_named_t *, char *, uchar_t);
extern void kstat_timer_init(kstat_timer_t *, char *);

/*
 * int
 * kstat_copyin(char *src, char *dst, size_t size)
 *
 * int
 * kstat_copyout(char *src, char *dst, size_t size)
 *
 *	Puspose:
 *
 *		Copy in/out a chunk of kstat data from/to userland.
 *		Currently, these are just wrappers around copyin() and
 *		copyout(), but this may change in the future.
 */

extern int kstat_copyin(char *, char *, size_t);
extern int kstat_copyout(char *, char *, size_t);

/*
 * int
 * default_kstat_copy(kstat_t *ksp, kstat_t *user_ksp, int rw)
 *
 *	Puspose:
 *
 *		The default routine to copy a kstat's data to/from userland.
 *		Performs certain massaging first, e.g. it corrects i/o
 *		statistics to compensate for incomplete transactions.
 */

extern int default_kstat_copy(kstat_t *, kstat_t *, int);

/*
 * Call kstat_waitq_enter() when a request arrives
 * and is placed into a pre-service state
 * (e.g., just prior to calling disksort).
 */

extern void kstat_waitq_enter(kstat_io_t *);

/*
 * Call kstat_waitq_exit() when a request is
 * removed from its pre-service state.
 * (e.g., just prior to calling XXstart).
 */

extern void kstat_waitq_exit(kstat_io_t *);

/*
 * Call kstat_runq_enter() when a request
 * is placed in its service state
 * (e.g., just prior to calling XXstart,
 * but after kstat_waitq_exit()).
 */

extern void kstat_runq_enter(kstat_io_t *);

/*
 * Call kstat_runq_exit() when a request is
 * removed from its service state
 * (e.g., just prior to calling biodone).
 */

extern void kstat_runq_exit(kstat_io_t *);

/*
 * kstat_waitq_to_runq() is equivalent to
 * kstat_waitq_exit() immediately followed by
 * kstat_runq_enter().  This is cheaper since
 * only one timestamp is required.
 */

extern void kstat_waitq_to_runq(kstat_io_t *);

/*
 * Call kstat_runq_back_to_waitq() if a request
 * has to be moved from the run queue back to the
 * wait queue, e.g. due to write throttling.
 */

extern void kstat_runq_back_to_waitq(kstat_io_t *);

/*
 * Call kstat_timer_start() to begin timing an event.
 * Call kstat_timer_stop() when that event completes.
 */

extern void kstat_timer_start(kstat_timer_t *);
extern void kstat_timer_stop(kstat_timer_t *);

#endif	/* defined(_KERNEL) */

#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_KSTAT_H */
