/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/************************************************************************
 ** MODULE INFORMATION*
 **********************
 **     FILE     NAME:       select.c
 **     SYSTEM   NAME:       block
 **     ORIGINAL AUTHOR(S):  Jan van Oorschot
 **     VERSION  NUMBER:     1.00
 **     CREATION DATE:       3 march 1993
 **
 ** DESCRIPTION: This module contains a blocking library based on
 **              the 'select' call of the socket library/kernel.
 **              It keeps a lists of sockets which will be used
 **              to block on.
 **              
 *************************************************************************
 ** CHANGES INFORMATION **
 *************************
 ** REVISION: $Revision$
 ** AUTHOR:   $Author$
 ** DATE:     $Date$
 ** LOG:      $Log$
 *************************************************************************/
#ifndef BSD_SELECT
#define BSD_SELECT
#endif

#ifdef OS2
#include <types.h>
#include <netlib.h>
#include <sys/select.h>					/* contains FD_*** macros */
#else
#include <sys/types.h>
#define sock_init()	0
#define soclose(x)	close(x)
#endif
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <dnpap.h>
#include <config.h>
#include <message.h>
#include <timer.h>

#include "block.h"

typedef struct _BlockList BLOCKLIST;
struct _BlockList
{
	BLOCK *blk;
	VOID *param;
	int (*cb) (BLOCK * block, int reason, VOID *param);
	BLOCKLIST *next;
};

static BLOCKLIST *BlockSockList;
static int BlockSockHigh;
static fd_set FDReads;
static fd_set FDWrites;
static fd_set FDExcepts;

/************************************************************************
 ** NAME:        BlockInit
 ** SYNOPSIS:    int BlockInit(VOID)
 ** DESCRIPTION: Initialises the blocking environment. It
 **              should alocate the nececary resources .
 ** RETURNS:     0 -->   no error
 **              else    error code
 ** RETURNS:     VOID
 *************************************************************************/
int BlockInit(VOID)
{
	MessageConfig(BLOCK_ERROR, "Block");
	
	BlockSockHigh = 0;					/* highest socket number until know */
	BlockSockList = NULL;				/* empty socket list */
	/* clear the static fd_set's */
	FD_ZERO(&FDReads);
	FD_ZERO(&FDWrites);
	FD_ZERO(&FDExcepts);
	return 0;
}

/************************************************************************
 ** NAME:        BlockClose
 ** SYNOPSIS:    int BlockClose(VOID)
 ** DESCRIPTION: Closes down the blocking environment, deregistering
 **              all the current blocks and freeing any memory.
 ** RETURNS:     0 -->   no error
 **              else    error code
 ** RETURNS:     VOID
 *************************************************************************/
int BlockClose(VOID)
{
	BLOCKLIST *ptr;

	ptr = BlockSockList;
	while (ptr != NULL)
	{
		BlockDeRegister(ptr->blk);
		ptr = ptr->next;
	}
	BlockSockList = NULL;
	return 0;
}

/************************************************************************
 ** NAME:        BlockRegister
 ** SYNOPSIS:    int BlockRegister(blk,param,cb)
 **              BLOCK   *blk        block structure
 **              VOID *param,
 **              int (*cb)(BLOCK *block,int reason,VOID *param)
 ** DESCRIPTION: Registers a new block at the blocking environment.
 **              The callback function <cb> will be called every
 **              time the block is triggered. <cb> is defined
 **              in the BLOCK structure.
 ** RETURNS:     0 -->   no error
 **              else    error code
 ** RETURNS:     VOID
 *************************************************************************/
int BlockRegister(BLOCK * blk,
				  VOID *param,
				  int (*cb) (BLOCK * block, int reason, VOID *param)
)
{
	BLOCKLIST *ptr;

	/* add this block to the block linked list */
	if ((ptr = DnpapMalloc(sizeof(BLOCKLIST))) == NULL)
		return BLOCK_ERROR + 1;
	ptr->next = BlockSockList;
	ptr->param = param;
	ptr->blk = blk;
	ptr->cb = cb;
	BlockSockList = ptr;
	/* check if this socket has the highest port number */
	if (BlockSockHigh <= blk->sckt)
	{
		BlockSockHigh = blk->sckt + 1;
	}
	/* mark this new socket in the static fs_set's */
	FD_SET(blk->sckt, &FDReads);
	FD_SET(blk->sckt, &FDWrites);
	FD_SET(blk->sckt, &FDExcepts);
	return 0;
}

/************************************************************************
 ** NAME:        BlockDeRegister
 ** SYNOPSIS:    int BlockDeRegister(BLOCK *blk)
 **              BLOCK *blk;     block structure
 ** DESCRIPTION: Removes the block <blk> from the blocking
 **              environment. This should always be possible
 **              since we evidently are currently not blocked.
 ** RETURNS:     0 -->   no error
 **              else    error code
 ** RETURNS:     VOID
 *************************************************************************/
int BlockDeRegister(BLOCK * blk)
{
	int tmpmax;
	BLOCKLIST *bptr, *ptr;

	/* find the block in the list and remove it */
	if (BlockSockList == NULL)
		return BLOCK_ERROR + 2;
	if (BlockSockList->blk == blk)
	{
		bptr = BlockSockList;
		BlockSockList = BlockSockList->next;
	}
	else
	{
		ptr = BlockSockList;
		bptr = BlockSockList->next;
		while (bptr != NULL && bptr->blk != blk)
		{
			ptr = bptr;
			bptr = bptr->next;
		}
		if (bptr == NULL)
			return BLOCK_ERROR + 3;
		ptr->next = bptr->next;
	}
	/* check if this was the highest socket, and if so, find the new highest */
	if (blk->sckt == (BlockSockHigh - 1))
	{
		ptr = BlockSockList;
		tmpmax = 0;
		while (ptr != NULL)
		{
			if (ptr->blk->sckt > tmpmax)
				tmpmax = ptr->blk->sckt;
			ptr = ptr->next;
		}
		BlockSockHigh = tmpmax + 1;
	}
	/* remove the the bit from the fd_set's */
	FD_CLR(blk->sckt, &FDReads);
	FD_CLR(blk->sckt, &FDWrites);
	FD_CLR(blk->sckt, &FDExcepts);
	return 0;
}

/************************************************************************
 ** NAME:        Block
 ** SYNOPSIS:    int Block(VOID);
 ** DESCRIPTION: Blocks until one of the block's is released.
 **              The callback function of this block will be called
 **              with a reason code. 
 **              In a normal environment, this function will be put
 **              in a tight loop.
 ** RETURNS:     0 -->   no error
 **              else    error code
 ** RETURNS:     VOID
 *************************************************************************/
int Block(VOID)
{
	fd_set fd_read, fd_write, fd_except;
	LONG t;
	struct timeval timeout;
	int iret;
	BLOCKLIST *ptr;

	/* copy static fd_set's to the local copies */
	memcpy(&fd_read, &FDReads, sizeof(fd_read));
	memcpy(&fd_write, &FDWrites, sizeof(fd_write));
	memcpy(&fd_except, &FDExcepts, sizeof(fd_except));
    /*
    FD_ZERO(&fd_read);
    FD_ZERO(&fd_write);
    FD_ZERO(&fd_except);
    ptr = BlockSockList;
    while(ptr != NULL)
    {
    	FD_SET(ptr->blk->sckt,&fd_read);
	    FD_SET(ptr->blk->sckt,&fd_write);
    	FD_SET(ptr->blk->sckt,&fd_except);
    	ptr = ptr->next;
    }
    */
	/* request the next timeout from the timer library and do select */
	if ((t = TimerRemaining()) == -1L)
	{
		iret = select(BlockSockHigh, &fd_read, NULL, &fd_except, NULL);
	}
	else
	{
		if (t == 0)
		{
			iret = 0;
		}
		else
		{
			timeout.tv_sec = t / 1000;
			timeout.tv_usec = (t % 1000) * 1000;
			iret = select(BlockSockHigh, &fd_read, NULL, &fd_except, &timeout);
		}
	}

	/* select reason: error */
	if (iret < 0)
	{
		DnpapMessage(DMC_WARNING, BLOCK_ERROR + 4, "block: select error");
		return (0 - (BLOCK_ERROR + 4));
	}
	/* select reason: timeout */
	if (iret == 0)
	{
		if (!TimerCheck())
		{
			DnpapMessage(DMC_WARNING, BLOCK_ERROR + 5, "block: timeout without reason");
		}
		return 0;
	}
	/* select reason: socket event */
	/* for all sockets, check if something has happend */
	ptr = BlockSockList;
	while (ptr != NULL)
	{
		if (FD_ISSET(ptr->blk->sckt, &fd_except))
		{
			(ptr->cb)
				(ptr->blk, BLOCKEVENT_EXCEPT, ptr->param);
		}
		if (FD_ISSET(ptr->blk->sckt, &fd_read))
		{
			(ptr->cb)
				(ptr->blk, BLOCKEVENT_READ, ptr->param);
		}
		ptr = ptr->next;
	}
	return 0;
}

/************************************************************************
 ** NAME:        BlockCheck
 ** SYNOPSIS:    int BlockCheck(VOID)
 ** DESCRIPTION: Does the same as Block, but without the actual 
 **              blocking. Should seldom be used, since it
 **              implements busy-waiting like behaviour.
 ** RETURNS:     0 -->   no error
 **              else    error code
 ** RETURNS:     VOID
 *************************************************************************/
int BlockCheck(VOID)
{
	return 0;
}
