/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/**********************************************************************
** MODULE INFORMATION*
**********************
**      FILE     NAME:       maxmem.c
**      SYSTEM   NAME:       maxmem
**      ORIGINAL AUTHOR(S):  Alfred Kayser
**      VERSION  NUMBER:     1.0
**      CREATION DATE:       91/09/16
**
** DESCRIPTION: Memory functions.
***********************************************************************
** CHANGES INFORMATION **
*************************
** REVISION:   $Revision: $  
** AUTHOR:     $Author: $ 
** DATE:       $Date: $
** LOG:        $Log: $
**********************************************************************/

#define LIBRARY							/* for #ifdef LIBRARY in dnpap.h  */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <dnpap.h>
#include <dmemory.h>
#include <config.h>

#include "maxmem.h"

/* Default configuration */
                            
/* limits are imposed by the OS and 
 * the signedness of maxmemChunk and maxmemMax */
#define MEMMINMAX	USHRT_MAX
#define MEMMINCHUNK	UCHAR_MAX

#define MEMDEFMAX   LONG_MAX
#define MEMMAXMAX   LONG_MAX
#if INT_MAX < LONG_MAX
#define MEMDEFCHUNK INT_MAX
#define MEMMAXCHUNK INT_MAX
#else
#define MEMDEFCHUNK LONG_MAX
#define MEMMAXCHUNK LONG_MAX
#endif

#define MAGIC1      	0x15786179UL
#define MAGIC2      	0xABBADEDEUL

#define MEMALIGN		(2*sizeof(ULONG))
#define MEMOVERHEAD		(SIZESIZE+MAGIC1SIZE+MAGIC2SIZE)

/* 's' stands for the requested size and 
 * 'p' for the pointer as the user receives it */
#define ALIGNSIZE(s)	(((ULONG)s + MAGIC2SIZE - 1) & ~(MAGIC2SIZE - 1))
#define TOTALSIZE(s)	(ALIGNSIZE(s) + MEMOVERHEAD)

#define SIZEOFFSET		(-MEMALIGN)
#define MAGIC1OFFSET	(-MAGIC1SIZE)
#define MAGIC2OFFSET(s)	ALIGNSIZE(s)

#define SIZESIZE		(sizeof(ULONG))
#define MAGIC1SIZE		(sizeof(ULONG))
#define MAGIC2SIZE		(sizeof(ULONG))

#define SIZEADDR(p)			((ULONG *)(VOID *)((BYTE *)p + SIZEOFFSET))
#define MAGIC1ADDR(p)		((ULONG *)(VOID *)((BYTE *)p + MAGIC1OFFSET))
#define MAGIC2ADDR(p,s)		((ULONG *)(VOID *)((BYTE *)p + MAGIC2OFFSET(s)))

#define SIZEVAL(p)			(*SIZEADDR(p))
#define MAGIC1VAL(p)		(*MAGIC1ADDR(p))
#define MAGIC2VAL(p,s)		(*MAGIC2ADDR(p,s))

#define USERPOINTER(p)		((BYTE*)p + MEMALIGN)
#define SYSPOINTER(p)		((BYTE*)p - MEMALIGN)

/* Prototypes for private functions */

PRIVAT VOID *MemoryMalloc(ULONG reqSize);
PRIVAT VOID *MemoryRealloc(VOID * memblock, ULONG reqSize);
PRIVAT VOID MemoryFree(VOID * memblock);

/* Local globals */

PRIVAT BOOLEAN initialized = FALSE;
PRIVAT BOOLEAN checking = TRUE;

PRIVAT LONG maxmemChunk = MEMDEFCHUNK;
PRIVAT LONG maxmemMax = MEMDEFMAX;
PRIVAT LONG maxmemUsed = 0;				/* Nothing used yet */

/* Export functions */

/**************************************************************
** NAME:        MemoryInit                                [API]
** SYNOPSIS:    VOID MemoryInit(LONG maxMem, LONG chunkSize)
** DESCRIPTION: Initializes the memory system.
**              This function installs a new DnpapMalloc,
**              DnpapRealloc and DnpapFree. The new functions
**              perform some sanity checks and restrict the
**              maximum amount of allocatable memory (maxMem)
**              and largest chunk of memory (chuckSize).
** RETURNS:     VOID
**************************************************************/
EXPORT VOID
    MemoryInit(LONG maxMem, LONG chunkSize)
{
	if (initialized == TRUE)
	{
		DnpapMessage(DMC_ERROR, MAXMEM_INIT, "maxmem: already initialized");
		return;
	}

	ConfigGetBoolean("maxmem.checking", &checking);

    if (checking == TRUE)
    {
    	ConfigGetLong("maxmem.max", &maxMem);
    	if (maxMem < MEMMINMAX)
    	{
    		DnpapMessage(DMC_ERROR, MAXMEM_LOWMAX, "maxmem: maxmem.max too small, using default value");
    		maxMem = MEMDEFMAX;
    	}
    	else if (maxMem > MEMMAXMAX)
    	{
    		DnpapMessage(DMC_ERROR, MAXMEM_HIGHMAX, "maxmem: maxmem.max too large, using default value");
    		maxMem = MEMDEFMAX;
    	}
    
    	ConfigGetLong("maxmem.chunk", &chunkSize);
    	if (chunkSize < MEMMINCHUNK)
    	{
    		DnpapMessage(DMC_ERROR, MAXMEM_LOWCHUNK, "maxmem: maxmem.chunk too small, using default value");
    		chunkSize = MEMDEFCHUNK;
    	}
    	else if (chunkSize > MEMMAXCHUNK)
    	{
    		DnpapMessage(DMC_ERROR, MAXMEM_HIGHCHUNK, "maxmem: maxmem.chunk too large, using default value");
    		chunkSize = MEMDEFCHUNK;
    	}
    
    	if (chunkSize > maxMem)
    	{
    		DnpapMessage(DMC_ERROR, MAXMEM_HIGHCHUNK, "maxmem: maxmem.chunk larger than maxmem.max, using maximum");
    		chunkSize = maxMem;
    	}
    
    	maxmemMax = maxMem;
    	maxmemChunk = chunkSize;
    }
    
   	DnpapSetMallocFun(MemoryMalloc);
   	DnpapSetReallocFun(MemoryRealloc);
   	DnpapSetFreeFun(MemoryFree);
}

/**************************************************************
** NAME:        MemoryGetAvail                            [API]
** SYNOPSIS:    LONG MemoryGetAvail()
** DESCRIPTION: Returns the max available memory.
** RETURNS:     LONG
**************************************************************/
EXPORT LONG
    MemoryGetAvail()
{                          
LONG used;

    if ((used = TOTALSIZE(1) + maxmemUsed) > maxmemMax)
        return 0;
	return maxmemMax - used;
}

EXPORT LONG
    MemoryGetUsed()
{
	return maxmemUsed;
}

/**************************************************************
** NAME:        MemoryGetMaxChunk                         [API]
** SYNOPSIS:    LONG MemoryGetMaxChunk()
** DESCRIPTION: Returns maximum allocatable size of memory.
** RETURNS:     LONG
**************************************************************/
EXPORT LONG
    MemoryGetMaxChunk()
{                                
LONG overhead;

    if ((overhead = TOTALSIZE(1)) > maxmemChunk)
        return 0;
	return maxmemChunk - overhead;
}

/**************************************************************
** NAME:        MemoryGetMaxMemory                        [API]
** SYNOPSIS:    ULONG MemoryGetMaxMemory()
** DESCRIPTION: Returns maximum allocatable amount of memory.
** RETURNS:     VOID
**************************************************************/
EXPORT LONG
    MemoryGetMaxMemory()
{                             
LONG overhead;

    if ((overhead = TOTALSIZE(1)) > maxmemMax)
        return 0;
	return maxmemMax - overhead;
}

PRIVAT VOID *
    MemoryMalloc(ULONG reqSize)
{
	BYTE *p;
	ULONG size;
    
	if (checking == TRUE)
	{
		size = TOTALSIZE(reqSize);

		if (size > maxmemChunk)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_TOOLARGE,
						 "malloc: request too large (requested=%lu, chunk=%lu)",
						 size, maxmemChunk);
			return NULL;
		}
		if (maxmemUsed + size > maxmemMax)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_OUTMEMORY,
						 "malloc: out of memory (requested=%lu, avail=%lu)",
						 size, MemoryGetAvail());
			return NULL;
		}

		if ((p = malloc(size)) != NULL)
		{
			p = USERPOINTER(p);
			SIZEVAL(p) = reqSize;
			MAGIC1VAL(p) = MAGIC1;
			MAGIC2VAL(p, reqSize) = MAGIC2;

			maxmemUsed += size;
		}
		else
		{
			DnpapMessage(DMC_FATAL, MAXMEM_ALLOC, "malloc: failed to allocate %lu bytes", size);
		}
	}
	else
	{
		p = malloc(reqSize);
	}

	return (VOID *) p;
}

PRIVAT VOID *
    MemoryRealloc(VOID * memblock, ULONG reqSize)
{
	ULONG oldSize;
	ULONG size;
	BYTE *p = memblock;

	if (p == NULL)
		return MemoryMalloc(reqSize);

	if (reqSize == 0)
	{
		MemoryFree(p);
		return NULL;
	}

	if (checking == TRUE)
	{
		size = TOTALSIZE(reqSize);
		oldSize = SIZEVAL(p);

		if (MAGIC1VAL(p) == MAGIC2)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_FREED,
						 "realloc: memory at 0x%lp was already freed", p);
			return NULL;
		}
		if (MAGIC1VAL(p) != MAGIC1)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_MUTILATED,
						 "realloc: magic value not found at start of memory at 0x%lp", p);
			return NULL;
		}
		if (MAGIC2VAL(p, oldSize) != MAGIC2)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_MUTILATED,
						 "realloc: magic value not found at end of memory at 0x%lp", p);
			return NULL;
		}

		if (size > oldSize)
		{
			if (size > maxmemChunk)
			{
				DnpapMessage(DMC_FATAL, MAXMEM_TOOLARGE,
							 "realloc: request too large (requested=%lu, chunk=%lu)",
							 size, maxmemChunk);
				return NULL;
			}
			if (maxmemUsed + size > maxmemMax)
			{
				DnpapMessage(DMC_FATAL, MAXMEM_OUTMEMORY,
							 "realloc: out of memory (requested=%lu, avail=%lu)",
							 size, MemoryGetAvail());
				return NULL;
			}
		}

		if ((p = realloc(SYSPOINTER(p), size)) != NULL)
		{
			p = USERPOINTER(p);
			SIZEVAL(p) = reqSize;
			MAGIC1VAL(p) = MAGIC1;
			MAGIC2VAL(p, reqSize) = MAGIC2;

			maxmemUsed -= TOTALSIZE(oldSize);
			maxmemUsed += size;
		}
		else
		{
			DnpapMessage(DMC_FATAL, MAXMEM_ALLOC, "realloc: failed to allocate %lu bytes", size);
		}
	}
	else
	{
		p = realloc(p, reqSize);
	}

	return (VOID *) p;
}

PRIVAT VOID
    MemoryFree(VOID * memblock)
{
	ULONG oldSize;
	BYTE *p = memblock;

	if (p == NULL)
		return;

	if (checking == TRUE)
	{
		oldSize = SIZEVAL(p);

		if (MAGIC1VAL(p) == MAGIC2)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_FREED,
						 "free: memory at 0x%lp was already freed", p);
			return;
		}
		if (MAGIC1VAL(p) != MAGIC1)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_MUTILATED,
						 "free: magic value not found at start of memory at 0x%lp", p);
			return;
		}
		if (MAGIC2VAL(p, oldSize) != MAGIC2)
		{
			DnpapMessage(DMC_FATAL, MAXMEM_MUTILATED,
						 "free: magic value not found at end of memory at 0x%lp", p);
			return;
		}

		SIZEVAL(p) = (ULONG) - 1;		/* set size to -1 */
		MAGIC1VAL(p) = MAGIC2;			/* switch magic values */
		MAGIC2VAL(p, oldSize) = MAGIC1;

		free(SYSPOINTER(p));

		maxmemUsed -= TOTALSIZE(oldSize);
	}
	else
	{
		free(p);
	}
}
