/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/************************************************************************
** MODULE INFORMATION **
************************
** FILE NAME:          timer.c
** MODULE NAME:        timer
** ORIGINAL AUTHOR(S): Dirk Wisse
** VERSION NUMBER:     1.0
** CREATION DATE:      1992/1/21 (y/m/d)
** DESCRIPTION:        is a module for timing purposes.
*************************************************************************
** CHANGES INFORMATION **
*************************
** REVISION:    $Revision$
** WORKFILE:    $Workfile$
** LOGINFO:     $Log$
*************************************************************************
** COPYRIGHT:   CARDIT (c) 1991
**              Department Of Computer Architecture and Digital
**              Techniques
**              Electrotechnical Faculty
**              Delft University of Technology
**              All rights reserved.
************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#ifdef OS2
#include <process.h>
#endif
#include <limits.h>
#include <dnpap.h>
#include "timer.h"

#ifndef UNIX
#if CLK_TCK==1000
#define TIMER_CLOCK (clock())
#else
#define TIMER_CLOCK ((clock()/CLK_TCK)*1000L)
#endif
#else
#include <sys/time.h>
static LONG msecclock()
{
    struct timeval tp;
    gettimeofday(&tp,NULL);
    return(tp.tv_sec*1000+tp.tv_usec/1000);
}
#define TIMER_CLOCK (msecclock())
#endif


TIMER_DESCR    *timerList=NULL;
ULONG           timerOffset;

static VOID     Insert(TIMER_DESCR *timer);
static BOOLEAN  Delete(TIMER_DESCR *timer);

static VOID TimerList(CHAR *s);

/*****************************************************************
** NAME:        TimerInit
** SYNOPSIS:    BOOLEAN TimerInit(VOID)  
** DESCRIPTION: initializes timer module.
**              Finds the global info segment and starts a thread
**              that periodically generates TIMER_EVENT_EXPIRE
**              events.
**              You can call TimerInit as often as you want, it
**              will only initialize the timer if it is not
**              already done.
**              So you could use TimerInit to make sure that
**              the timer was initialized.
** RETURNS:     BOOLEAN
*******************************************************************/
BOOLEAN TimerInit(VOID)
{
    static BOOLEAN init=FALSE;
    
    if (!init)
    {
        timerOffset=TIMER_CLOCK;    
        init=TRUE;
    }
    return init;
}                                



/*****************************************************************
** NAME:        TimerRegister
** SYNOPSIS:    TIMER_DESCR *TimerRegister(
**                  TIMER_CALLBACK callback,
**                  VOID  *parm;
**                  ULONG msecs,
**                  ULONG count,
**                  int type)
** PARAMETERS:  - TIMER_CALLBACK callback
**              The callback function should have the following
**              synopsis:
**              VOID FAR *callback(TIMER_DESCR *timer,
**                                      ULONG now, VOID *parm).
**              - VOID *parm
**              Optional parmater for callback function.
**              - ULONG msecs
**              Timer period in milliseconds.
**              - ULONG count
**              Number of periods before the timer is removed or
**              TIMER_FOREVER if the timer should run until it is
**              removed with TimerRemove.
**              - int type
**              Type which specifies what to do if we have missed
**              some periods. TIMER_TYPE_SKIP means that these periods
**              are skiped. TIMER_TYPE_RECOVER means that we try to
**              recover from this situation by some shorter
**              periods so the mean length of a period remains
**              always the same. 
** DESCRIPTION: registers a callback which is called periodically.
**              - A one shot timer can be registered with:
**              TimerRegister(callback, 1000L, 1, TIMER_SKIP)
**              - A normal timer is registered with:
**              TimerRegister(callback, 1000L,
**                                  TIMER_FOREVER, TIMER_RECOVER)
** RETURNS:     TIMER_DESCR *timer.
** SEE ALSO:    TimerRemove.
*******************************************************************/
TIMER_DESCR *TimerRegister(TIMER_CALLBACK callback, VOID *parm,
                    ULONG msecs, ULONG count, int type)
{
	TIMER_DESCR *timer;

	timer=(TIMER_DESCR *)DnpapMalloc(sizeof(TIMER_DESCR));
	if (timer==NULL)
    {
        DnpapMessage(DMC_FATAL, TIMER_ERROR,"Malloc failed");
        DnpapExit(1);
    }
	if (msecs==0)
		msecs=1;
    timer->callback=callback;
    timer->parm=parm;
	timer->msecs=msecs;
    timer->type=type;
    timer->count=count;
	timer->alarm=TIMER_CLOCK+msecs-31;
	Insert(timer);
	return timer;
}

/*****************************************************************
** NAME:        TimerRemove
** SYNOPSIS:    ULONG TimerRemove(TIMER_DESCR *timer)
** PARAMETERS:  - TIMER_DESCR *timer
**              Timer handle obtained by TimerRegister.
** DESCRIPTION: removes a timer registered with TimerRegister.
** RETURNS:     ULONG remaining milliseconds before timer expires.
** SEE ALSO:    TimerRegister.
*******************************************************************/
ULONG TimerRemove(TIMER_DESCR *timer)
{
    ULONG msecs;

    if (timer==0)
        return 0;
    if (!Delete(timer))
        return 0; 
        
	msecs=timer->alarm-TIMER_CLOCK;
    if (msecs > LONG_MAX)
        msecs = 0; 
        
	DnpapFree(timer);
    
	return msecs;
}



/*****************************************************************
** NAME:        TimerRemaining
** SYNOPSIS:    ULONG TimerRemaining(VOID)
** DESCRIPTION: Returns time till one off the timers expires.
**              After this period of time you should call
**              TimerCheck to handle the expired timers.
** RETURNS:     ULONG remaining time.
**              -1 -> No timers.
**               0 -> No time remaining. Call TimerCheck as soon
**                    as possible.
**              >0 -> Remaining time of timer to expire first.
** SEE ALSO:    TimerCheck.
*******************************************************************/
ULONG TimerRemaining(VOID)
{
    ULONG msecs;

    if (timerList==NULL)
        return (ULONG)-1;
    msecs = timerList->alarm-TIMER_CLOCK;
    if (msecs > LONG_MAX)
        msecs = 0; 
    return msecs;
}



/*****************************************************************
** NAME:        TimerCheck
** SYNOPSIS:    BOOLEAN TimerCheck(VOID)
** DESCRIPTION: Main function to get the timers going. Call this
**              function regularly.
** RETURNS:     BOOLEAN success.
*******************************************************************/
BOOLEAN TimerCheck(VOID)
{
    TIMER_DESCR *timer,**p;

/*  TimerList("Check begin");  */

	while (timerList != NULL && TIMER_CLOCK-timerList->alarm<LONG_MAX)
    {
        timer=timerList;
        timer->callback(timer,TIMER_CLOCK-timerOffset,timer->parm);
        if (timer==timerList)    
        {
	        timerList=timerList->next;
            do
            {
                if (timer->count!=TIMER_FOREVER)
                    timer->count--;
                timer->alarm+=timer->msecs;
            }
            while(timer->type==TIMER_TYPE_SKIP &&
                  TIMER_CLOCK-timer->alarm<LONG_MAX && 
                  timer->count!=0);
            if (timer->count==0)
            {
                DnpapFree(timer);
            }
            else
            {
	            for (p=&timerList;
		            *p!=NULL && (timer->alarm-(*p)->alarm)<LONG_MAX;
		            p=&(*p)->next);
	            timer->next=*p;
	            *p=timer;
            }
        }
    }

/*  TimerList("Check end");  */

    return TRUE;
}
                              
/*****************************************************************
** NAME:        TimerNow
** SYNOPSIS:    ULONG TimerNow(VOID)
** DESCRIPTION: Returns the time in milliseconds sinds TimerInit.
** RETURNS:     ULONG time in milliseconds sinds TimerInit.
*******************************************************************/
ULONG TimerNow(VOID)
{
    return TIMER_CLOCK-timerOffset;
}



/*****************************************************************
** NAME:        TimerChange
** SYNOPSIS:    BOOLEAN TimerChange(TIMER_DESCR *timer,
**                      ULONG msecs)
** PARAMETERS:  - TIMER_DESCR *timer
**              Timer handle obtained by TimerRegister.
**              - ULONG msecs
**              New timer period. The change takes place after
**              the next expiration.
** DESCRIPTION: changes period of a timer.
** RETURNS:     BOOLEAN success.
** SEE ALSO:    TimerRegister, TimerRemove.
*******************************************************************/
BOOLEAN TimerChange(TIMER_DESCR *timer, ULONG msecs)
{
    if (timer==0)
        return FALSE;
	if (msecs==0)
		msecs=1;
	timer->msecs=msecs;
	return TRUE;
}



static VOID Insert(TIMER_DESCR *timer)
{
	TIMER_DESCR **p;

	for (p=&timerList;
      	  *p!=NULL && (timer->alarm-(*p)->alarm)<LONG_MAX;
		  p=&(*p)->next);
	timer->next=*p;
	*p=timer;

/*  TimerList("Insert");  */
}


static BOOLEAN Delete(TIMER_DESCR *timer)
{
	TIMER_DESCR **p;
	
	for (p=&timerList;
		  *p!=NULL && *p!=timer;
		  p=&(*p)->next);
	
    if (*p==NULL)  
        return FALSE;
	
    *p=(*p)->next; 

/*  TimerList("Delete");  */
    
    return TRUE;
}


VOID TimerList(CHAR *s)
{
TIMER_DESCR *q;

   	printf("%s report\n", s);
	for (q=timerList; q!=NULL; q=q->next)
    {
    	printf("\ttimer->alarm = %lu\n",q->alarm);
    	printf("\ttimer->msecs = %lu\n",q->msecs);
    }

 	printf("TimerRemaining report\n");
   	printf("TIMER_CLOCK      = %lu\n",TIMER_CLOCK);
	printf("timerList->alarm = %lu\n",timerList->alarm);
	printf("LONG_MAX    	 = %lu\n",LONG_MAX);
   	printf("alarm - TIMER_CLOCK = %lu\n",timerList->alarm - TIMER_CLOCK);
}

