/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <errno.h>
#include <memory.h>
#include <string.h>
#include <stropts.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nwbpf.h"


/* nwbpf.c - NwBpf module functions */


typedef struct _NwBpf
{
	INT Socket;
	CHAR *Interface;
    BYTE  Address[NWBPF_SIZE_ADDRESS];
	INT BufSize;
	TIMEVAL Timeout;
	INT Flags;
	BYTE *Buf;
	LONG  Pos;
	LONG  End;

} NwBpf;


typedef struct _NwBpfDg
{
    NW_DG  NwInfo;
    NwBpf  BpfInfo;

    BLOCK  Block;

    struct _NwBpfDg *Prev;
    struct _NwBpfDg *Next;

} NwBpfDg;


static BOOLEAN NwBpfInit(VOID);
static BOOLEAN NwBpfExit(VOID);

static NW_DG  *NwBpfAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwBpfConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwBpfSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwBpfSendTo(NW_DG *dg, char *addr, BYTE *frame, INT len);
static BOOLEAN NwBpfClose(NW_DG *dg);
static BOOLEAN NwBpfIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwBpfStatistics(char *addr);

static int bpfopen(int flags);


NW_DG_IFACE BpfDgIface = { "bpf",
                            NwBpfInit,   NwBpfExit,
                            NwBpfAccept, NwBpfConnect,
                            NwBpfSend,   NwBpfSendTo,
                            NwBpfClose,  NwBpfIfConfig,
                            NwBpfStatistics
                         };


static NwBpfDg *NwBpfDgList = NULL;

static NwStat nwStat;


static INT bpfsocket;

static BOOLEAN NwBpfCloseDg(NwBpfDg *dg);

static BOOLEAN NwBpfDgAddList(NwBpfDg *dg);
static BOOLEAN NwBpfDgRemoveList(NwBpfDg *dg);

static NwBpfDg* NewNwBpfDg(CHAR *interface, INT bufsize);
static VOID DelNwBpfDg(NwBpfDg *dg);

static BOOLEAN NwBpfRead(NwBpfDg *dg);

static BOOLEAN NwBpfSetup(NwBpfDg *dg, CHAR *interface, INT flags, INT bufsize, TIMEVAL *timeout);

static int NwBpfBlockCallback(BLOCK *blk, int reason, VOID *param);



/* Initializes the NwBpf module
 * Opens a socket for general use
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwBpfInit(VOID)
{
static NwBpfDg *laststat = NULL;

    if ((bpfsocket = bpfopen(O_WRONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_OPEN, "nwbpf: open failed");
        return FALSE;
    }

	laststat = NULL;

    return TRUE;
}


/* Exits the NwBpf module
 * Closes all opened sockets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwBpfExit(VOID)
{
BOOLEAN rc = TRUE;
static NwBpfDg *laststat;
NwBpfDg *dg;

    for (dg = NwBpfDgList; dg != NULL; dg = NwBpfDgList)
        rc &= NwBpfClose((NW_DG *)dg);

    if (bpfsocket > 0)
        close(bpfsocket);

	laststat = NULL;

    return rc;
}


NwStat *NwBpfStatistics(char *addr)
{
static NwBpfDg *laststat;
char *p;
NwBpfDg *dg;
BPFSTAT stats;

	if ((p = strchr(addr, ':')) == NULL)
		return NULL;

	if (laststat != NULL && strcmp(laststat->BpfInfo.Interface, p+1) == 0)
	{
		if (ioctl(dg->BpfInfo.Socket, BIOCGSTATS, &stats) == 0)
			nwStat.InDiscards = stats.bs_drop;
		return &nwStat;
	}

    for (dg = NwBpfDgList; dg != NULL; dg = dg->Next)
	{
		if (strcmp(dg->BpfInfo.Interface, p+1) == 0)
		{
			laststat = dg;
			if (ioctl(dg->BpfInfo.Socket, BIOCGSTATS, &stats) == 0)
				nwStat.InDiscards = stats.bs_drop;
    		return &nwStat;
		}
	}
	return NULL;
}


/* Opens a socket for receiving promiscuous packets
 * Opens the specified interface with capacity 'bufsize' which is
 * is automatically allocated.
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwBpfAccept(char *addr           /* accept specification: 'bpf:(char*)interface:(ulong)bufsize' */,
                  NW_DG_RECV_CALL recv  /* receive callback */,
                  VOID *param           /* extra parameter you want to be forwarded to recv() */)
{
NwBpfDg *dg;
CHAR interface[NWBPF_SIZE_INTERFACE];
INT bufsize;
INT flags;
TIMEVAL timeout;
CHAR *p;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWBPF_PARAM, "nwbpf: invalid parameter list");
        return NULL;
    }

    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';

	if (sscanf(p, ":%d", &bufsize) != 1)
	{
        DnpapMessage(DMC_ERROR, NWBPF_PARAM, "nwbpf: invalid parameters");
	    return NULL;
	}

	flags = BIOCPROMISC;

    /*  set buffer timeout at 1/10 of a second  */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwBpfDg(interface, bufsize)) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWBPF_NEWBPF, "nwbpf: memory allocation failed");
		return NULL;
    }

    if ((dg->BpfInfo.Socket = bpfopen(O_RDONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_OPEN, "nwbpf: open failed");
    	NwBpfCloseDg(dg);
        return NULL;
    }

    if (NwBpfSetup(dg, interface, flags, bufsize, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWBPF_SETUP, "nwbpf: setup failed");
    	NwBpfCloseDg(dg);
        return NULL;
    }

    dg->Block.sckt = dg->BpfInfo.Socket;

    if (BlockRegister(&dg->Block, dg, NwBpfBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_BLOCK, "nwbpf: block registration failed");
    	NwBpfCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &BpfDgIface;
    sprintf(dg->NwInfo.addr, "bpf:%s:%d", interface, bufsize);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;

	strcpy(dg->BpfInfo.Interface, interface);
	dg->BpfInfo.BufSize = bufsize;
	dg->BpfInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->BpfInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->BpfInfo.Flags = flags;
	dg->BpfInfo.Pos = 0;
	dg->BpfInfo.End = 0;

    if (NwBpfDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWBPF_ADDLIST, "nwbpf: could not add to list");
    	NwBpfCloseDg(dg);
		return NULL;
	}

    if (ioctl(dg->BpfInfo.Socket, BIOCFLUSH, 0) < 0)
	{
        DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: read flush failed");
    	NwBpfCloseDg(dg);
		return NULL;
	}

    return (NW_DG *)dg;
}


/* Opens a socket for receiving and sending packets from and to a specific host
 * Opens the specified interface with capacity 'bufsize' which
 * is automatically allocated.
 * The Ethernet address is recorded
 * and the receive callback is only called for packets to or from this address
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwBpfConnect(char *addr           /* connect specification: 'bpf:(char*)interface:(ulong)bufsize:(int[6])ethernet' */,
                    NW_DG_RECV_CALL recv  /* receive callback */,
                    VOID *param           /* extra parameter you want to forward to recv() */)
{
NwBpfDg *dg;
CHAR interface[NWBPF_SIZE_INTERFACE];
INT bufsize;
INT flags;
TIMEVAL timeout;
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWBPF_PARAM, "nwbpf: invalid parameter list");
        return NULL;
    }

    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';

	if (sscanf(p, ":%d:%u.%u.%u.%u.%u.%u", &bufsize, &a1, &a2, &a3, &a4, &a5, &a6) != 8)
	{
        DnpapMessage(DMC_ERROR, NWBPF_PARAM, "nwbpf: invalid parameters");
	    return NULL;
	}

	flags = 0;

    /*  set buffer timeout at 1/10 of a second  */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwBpfDg(interface, bufsize)) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWBPF_NEWBPF, "nwbpf: memory allocation failed");
		return NULL;
    }

    if ((dg->BpfInfo.Socket = bpfopen(O_RDWR)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_OPEN, "nwbpf: open failed");
    	NwBpfCloseDg(dg);
        return NULL;
    }

    if (NwBpfSetup(dg, interface, flags, bufsize, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWBPF_SETUP, "nwbpf: setup failed");
    	NwBpfCloseDg(dg);
        return NULL;
    }

    dg->Block.sckt = dg->BpfInfo.Socket;

    if (BlockRegister(&dg->Block, dg, NwBpfBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_BLOCK, "nwbpf: block registration failed");
    	NwBpfCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &BpfDgIface;
    sprintf(dg->NwInfo.addr, "bpf:%s:%d:%u.%u.%u.%u.%u.%u",
                             interface, bufsize, a1, a2, a3, a4, a5, a6);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;

	strcpy(dg->BpfInfo.Interface, interface);
    dg->BpfInfo.Address[0] = a1;
    dg->BpfInfo.Address[1] = a2;
    dg->BpfInfo.Address[2] = a3;
    dg->BpfInfo.Address[3] = a4;
    dg->BpfInfo.Address[4] = a5;
    dg->BpfInfo.Address[5] = a6;
	dg->BpfInfo.BufSize = bufsize;
	dg->BpfInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->BpfInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->BpfInfo.Flags = flags;
	dg->BpfInfo.Pos = 0;
	dg->BpfInfo.End = 0;

    if (NwBpfDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWBPF_ADDLIST, "nwbpf: could not add to list");
    	NwBpfCloseDg(dg);
		return NULL;
	}

    if (ioctl(dg->BpfInfo.Socket, BIOCFLUSH, 0) < 0)
	{
        DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: read flush failed");
    	NwBpfCloseDg(dg);
		return NULL;
	}

    return (NW_DG *)dg;
}


BOOLEAN NwBpfClose(NW_DG *dg)
{
static NwBpfDg *laststat;
NwBpfDg *nitdg = (NwBpfDg *)dg;

    NwBpfDgRemoveList(nitdg);
    NwBpfCloseDg(nitdg);

	laststat = NULL;

    return TRUE;
}


BOOLEAN NwBpfCloseDg(NwBpfDg *dg)
{
    BlockDeRegister(&dg->Block);

    if (dg->BpfInfo.Socket != 0)
        close(dg->BpfInfo.Socket);

    DelNwBpfDg(dg);

    return TRUE;
}


/* Sends a packet over an existing NW datagram
 * The destination address is checked against the recorded Ethernet address
 * to allow only packets that are sent to that address
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwBpfSend(NW_DG *dg    /* valid NW datagram */,
                 BYTE *frame  /* packet */,
                 INT len      /* packet length */)
{
NwBpfDg *nitdg = (NwBpfDg *) dg;
BYTE noaddr[NWBPF_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (len > NWBPF_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWBPF_SEND, "nwbpf: packet too long");
        return FALSE;
    }

    if (memcmp(nitdg->BpfInfo.Address, noaddr, NWBPF_SIZE_ADDRESS) == 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_SEND, "nwbpf: can not send over accept socket");
        return FALSE;
    }

    memcpy(frame, nitdg->BpfInfo.Address, NWBPF_SIZE_ADDRESS);

    if (write(nitdg->BpfInfo.Socket, frame, len) < 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_SEND, "nwbpf: write failed");
        return FALSE;
    }

    return TRUE;
}


/* Sends a packet to a specified address
 * This function can be used to sent one-shot only packets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwBpfSendTo(NW_DG *dg    /* not supported, set to NULL */,
                   char *addr   /* address specification: 'bpf:(char*)interface:(int[6])ethernet' */,
                   BYTE *frame  /* packet */,
                   INT len      /* packet length */)
{
IFREQ ifr;
CHAR interface[NWBPF_SIZE_INTERFACE];
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;
BYTE address[NWBPF_SIZE_ADDRESS];

    if (dg != NULL)
        DnpapMessage(DMC_WARNING, NWBPF_SEND, "nwbpf: sendto over existing connection has no effect");

    if (bpfsocket == 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_INIT, "nwbpf: not initialized");
        return FALSE;
    }

    if (len > NWBPF_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWBPF_SEND, "nwbpf: packet too long");
        return FALSE;
    }

    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWBPF_PARAM, "nwbpf: invalid parameter list");
        return FALSE;
    }

	if ((p = strchr(++addr, ':')) == NULL)
	{
		p = addr+strlen(addr);
	}
	else
	{
    	if (sscanf(p, ":%u.%u.%u.%u.%u.%u", &a1, &a2, &a3, &a4, &a5, &a6) != 6)
    	{
            DnpapMessage(DMC_ERROR, NWBPF_PARAM, "nwbpf: invalid parameters");
    	    return FALSE;
    	}

        address[0] = a1;
        address[1] = a2;
        address[2] = a3;
        address[3] = a4;
        address[4] = a5;
        address[5] = a6;

        memcpy(frame, address, NWBPF_SIZE_ADDRESS);
    }

    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';

    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    if (ioctl(bpfsocket, BIOCSETIF, &ifr) < 0)
	{
        DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: interface bind failed");
		return FALSE;
	}

    if (write(bpfsocket, frame, len) < 0)
    {
        DnpapMessage(DMC_ERROR, NWBPF_SEND, "nwbpf: write failed");
        return FALSE;
    }

    return TRUE;
}


BOOLEAN NwBpfIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{
NwBpfDg *bpfdg = (NwBpfDg *) dg;
UINT type;
IFREQ ifr;

    if (ioctl(bpfdg->BpfInfo.Socket, BIOCGDLT, &type) < 0)
		return FALSE;

    if (ioctl(bpfdg->BpfInfo.Socket, SIOCGIFADDR, &ifr) < 0)
		return FALSE;

	switch (type)
	{
	case DLT_EN10MB:
		ifconfig->type = MAC_TYPE_ETHERNET_CSMACD;
		ifconfig->mtu = MAC_MTU_ETHERNET_CSMACD;
		ifconfig->speed = MAC_SPEED_ETHERNET_CSMACD;
		ifconfig->addrlen = MAC_ADDRLEN_ETHERNET_CSMACD;
		memcpy(ifconfig->address, ifr.ifr_addr.sa_data, ifconfig->addrlen);
		memcpy(ifconfig->broadcast, "\xff\xff\xff\xff\xff\xff", ifconfig->addrlen);
		break;
	default:
		DnpapMessage(DMC_ERROR, NWBPF_CONFIG, "nwbpf: unknown interface type");
		return FALSE;
	}

	return TRUE;
}


int NwBpfBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwBpfDg *dg = (NwBpfDg *) param;

	switch (reason)
	{
	case BLOCKEVENT_READ:
		if (NwBpfRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWBPF_READ, "nwbpf: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWBPF_WEVENT, "nwbpf: write event");
		break;
	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWBPF_EEVENT, "nwbpf: except event");
		break;
	default:
		DnpapMessage(DMC_ERROR, NWBPF_NEVENT, "nwbpf: unknown block event");
		return 1;
	}

	return 0;
}


BOOLEAN NwBpfRead(NwBpfDg *dg)
{
BYTE *cp;
BPFHDR *hp;
USHORT msglen;
BYTE noaddr[NWBPF_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (!(dg->BpfInfo.Pos < dg->BpfInfo.End))
    {
    	if ((dg->BpfInfo.End = read(dg->BpfInfo.Socket, dg->BpfInfo.Buf, dg->BpfInfo.BufSize)) < 0)
    		return FALSE;

    	dg->BpfInfo.Pos = 0;
    }

	while (dg->BpfInfo.Pos < dg->BpfInfo.End)
	{
		cp = dg->BpfInfo.Buf+dg->BpfInfo.Pos;

		hp = (BPFHDR *)cp;
	    msglen = hp->bh_caplen;

		dg->BpfInfo.Pos += BPF_WORDALIGN(hp->bh_hdrlen+msglen);

        if (memcmp(dg->BpfInfo.Address, noaddr, NWBPF_SIZE_ADDRESS) != 0 &&
            memcmp(dg->BpfInfo.Address, cp, NWBPF_SIZE_ADDRESS) != 0 &&
			memcmp(dg->BpfInfo.Address, cp+NWBPF_SIZE_ADDRESS, NWBPF_SIZE_ADDRESS) != 0)
            continue;

		if (dg->NwInfo.recv != NULL)
			dg->NwInfo.recv((NW_DG *)dg, dg->NwInfo.addr, cp, msglen, dg->NwInfo.parm);
	}

	return TRUE;
}


BOOLEAN NwBpfDgAddList(NwBpfDg *dg)
{
    if (NwBpfDgList != NULL)
		NwBpfDgList->Prev = dg;

    dg->Next = NwBpfDgList;
    dg->Prev = NULL;

    NwBpfDgList = dg;

    return TRUE;
}


BOOLEAN NwBpfDgRemoveList(NwBpfDg *dg)
{
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwBpfDgList = dg->Next;

    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;

    return TRUE;
}


NwBpfDg* NewNwBpfDg(CHAR *interface, INT bufsize)
{
NwBpfDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwBpfDg))) == NULL)
		return NULL;

    dg->NwInfo.iface = NULL;
	dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;

    dg->BpfInfo.Interface = NULL;
    memset(dg->BpfInfo.Address, 0, NWBPF_SIZE_ADDRESS);
    dg->BpfInfo.Buf = NULL;
	dg->BpfInfo.Socket = 0;
	dg->BpfInfo.BufSize = 0;
	dg->BpfInfo.Timeout.tv_sec = 0;
	dg->BpfInfo.Timeout.tv_usec = 0;
	dg->BpfInfo.Flags = 0;
	dg->BpfInfo.Pos = 0;
	dg->BpfInfo.End = 0;

    dg->Next = NULL;
    dg->Prev = NULL;

	if ((dg->BpfInfo.Interface = DnpapMalloc(strlen(interface)+1)) == NULL)
		return NULL;

	if ((dg->BpfInfo.Buf = DnpapMalloc(bufsize)) == NULL)
    {
    	DnpapFree(dg->BpfInfo.Interface);
		return NULL;
	}

	return dg;
}


VOID DelNwBpfDg(NwBpfDg *dg)
{
    DnpapFree(dg->BpfInfo.Interface);
    DnpapFree(dg->BpfInfo.Buf);

    DnpapFree(dg);
}


BOOLEAN NwBpfSetup(NwBpfDg *dg, CHAR *interface, INT flags, INT bufsize, TIMEVAL *timeout)
{
IFREQ ifr;
PROGRAM filter;

#ifdef WORKING_BIOCSBLEN
    if (ioctl(dg->BpfInfo.Socket, BIOCSBLEN, &bufsize) < 0)
	{
        DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: bufsize option set failed");
		return FALSE;
	}
#endif

    if (ioctl(dg->BpfInfo.Socket, BIOCSRTIMEOUT, timeout) < 0)
	{
        DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: timeout option set failed");
		return FALSE;
	}

    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    if (ioctl(dg->BpfInfo.Socket, BIOCSETIF, &ifr) < 0)
	{
        DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: interface bind failed");
		return FALSE;
	}

	filter.bf_len = 0;
	filter.bf_insns = NULL;
	if (ioctl(dg->BpfInfo.Socket, BIOCSETF, &filter) < 0)
	{
        DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: filter option set failed");
		return FALSE;
	}

    if (flags & BIOCPROMISC)
    {
        if (ioctl(dg->BpfInfo.Socket, BIOCPROMISC) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWBPF_IOCTL, "nwbpf: promiscuous mode set failed");
    		return FALSE;
    	}
    }

    return TRUE;
}


int bpfopen(int flags)
{
    int i = 0, fd;
    char device[16];

    do
	{
        sprintf(device, "/dev/%s%d", NWBPFDEVICE, i++);
        fd = open(device, flags);
    }
	while (fd < 0 && errno == EBUSY);

    return fd;
}
