/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#include <stropts.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/dlpi.h>
#include <sys/bufmod.h>

#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nwdlpi.h"

 
/* nwdlpi.c - NwDlpi module functions */

/* The DLPI code in this source is based on the information in
   the Internet Engineering document "How to Use DLPI" by Neal Nuckolls */

            
typedef struct _NwDlpi
{
	INT Socket;
	CHAR *Interface;
    BYTE  Address[NWDLPI_SIZE_ADDRESS];
	ULONG BufSize;
	ULONG SnapLen;
	TIMEVAL Timeout;
	BYTE *Buf;
	LONG  Pos;
	LONG  End;
	         
} NwDlpi;
    

typedef struct _NwDlpiDg
{          
    NW_DG   NwInfo;
    NwDlpi  DlpiInfo;
    
    BLOCK            *Block;
    
    struct _NwDlpiDg *Prev;
    struct _NwDlpiDg *Next;
    
} NwDlpiDg;

      
static BOOLEAN NwDlpiInit(VOID);
static BOOLEAN NwDlpiExit(VOID);

static NW_DG  *NwDlpiAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwDlpiConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwDlpiSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwDlpiSendTo(NW_DG *dg, char *addr, BYTE *frame, INT len);
static BOOLEAN NwDlpiClose(NW_DG *dg);
static BOOLEAN NwDlpiIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwDlpiStatistics(char *addr);


NW_DG_IFACE DlpiDgIface = { "dlpi", 
                            NwDlpiInit,   NwDlpiExit, 
                            NwDlpiAccept, NwDlpiConnect, 
                            NwDlpiSend,   NwDlpiSendTo,
                            NwDlpiClose,  NwDlpiIfConfig,
                            NwDlpiStatistics 
                        };


static NwDlpiDg *NwDlpiDgList = NULL;

static NwStat nwStat;


static INT dlpisocket;
static BYTE dlpibuf[NWDLPI_SIZE_DLBUF];

static BOOLEAN NwDlpiCloseDg(NwDlpiDg *dg);

static BOOLEAN NwDlpiDgAddList(NwDlpiDg *dg);
static BOOLEAN NwDlpiDgRemoveList(NwDlpiDg *dg);

static NwDlpiDg* NewNwDlpiDg(CHAR *interface, ULONG bufsize);
static VOID DelNwDlpiDg(NwDlpiDg *dg);

static BOOLEAN NwDlpiRead(NwDlpiDg *dg);

static BOOLEAN NwDlpiSetup(NwDlpiDg *dg, CHAR *interface, ULONG bufsize, ULONG snaplen, TIMEVAL *timeout);

static int NwDlpiBlockCallback(BLOCK *blk, int reason, VOID *param);

static BOOLEAN DlpiAttach(INT socket, CHAR *interface);
static BOOLEAN DlpiDetach(INT socket, CHAR *interface);
static BOOLEAN DlpiBind(INT socket, CHAR *interface);
static BOOLEAN DlpiUnBind(INT socket, CHAR *interface);
static BOOLEAN DlpiInfo(INT socket, BYTE *buf);
static BOOLEAN DlpiPromisc(INT socket, INT level);
static BOOLEAN DlpiSend(INT socket, BYTE *frame, INT len);
static BOOLEAN DlpiAck(INT socket, ULONG prim, BYTE *buf);

static CHAR *DlpiPrim(ULONG prim);

static int strioctl(int socket, int cmd, int timeout, int len, BYTE *dp);
static int strgetmsg(int socket, STRBUF *ctl, STRBUF *data, INT *flags);

static VOID SigAlarm(INT signal);



/* Initializes the NwDlpi module
 * Initializes the NwDlpi messages and opens a socket for general use
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiInit(VOID)
{       
	if (geteuid() != 0)
	{
        DnpapMessage(DMC_WARNING, NWDLPI_ROOT, "nwdlpi: must be root to use dlpi features");
		return TRUE;
	}
	
    if ((dlpisocket = open(NWDLPI_DEVICE, O_WRONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_OPEN, "nwdlpi: open failed");
        return FALSE;
    }

    return TRUE;
}


/* Exits the NwDlpi module
 * Closes all opened sockets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiExit(VOID)
{             
BOOLEAN rc = TRUE;
NwDlpiDg *dg;

    for (dg = NwDlpiDgList; dg != NULL; dg = NwDlpiDgList)
        rc &= NwDlpiClose((NW_DG *)dg);

    if (dlpisocket != 0)
        close(dlpisocket);        
        
    return rc;
}


NwStat *NwDlpiStatistics(char *addr)
{
    return &nwStat;
}

        
/* Opens a socket for receiving promiscuous packets
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwDlpiAccept(char *addr           /* accept specification: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                   NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                   VOID *param           /* extra parameter you want to be forwarded to recv() */)
{   
NwDlpiDg *dg;
CHAR interface[NWDLPI_SIZE_INTERFACE];
ULONG bufsize, snaplen;
TIMEVAL timeout;
CHAR *p;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (strncmp(interface, NWDLPI_INTERFACE, strlen(NWDLPI_INTERFACE)) != 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid interface");
	    return NULL;
	}
	
	if (sscanf(p, ":%lu:%lu", &bufsize, &snaplen) != 2)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameters");
	    return NULL;
	}
	                           
    /* set buffer timeout at 1/10 of a second */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwDlpiDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWDLPI_NEWDLPI, "nwdlpi: memory allocation failed");
		return NULL;
    }

    if ((dg->DlpiInfo.Socket = open(NWDLPI_DEVICE, O_RDWR)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_OPEN, "nwdlpi: open failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }

    if (NwDlpiSetup(dg, interface, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SETUP, "nwdlpi: setup failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->DlpiInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwDlpiBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_BLOCK, "nwdlpi: block registration failed");
    	NwDlpiCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &DlpiDgIface;                                
    sprintf(dg->NwInfo.addr, "dlpi:%s:%lu:%lu", interface, bufsize, snaplen);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
    
	strcpy(dg->DlpiInfo.Interface, interface);
	dg->DlpiInfo.BufSize = bufsize;
	dg->DlpiInfo.SnapLen = snaplen;
	dg->DlpiInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->DlpiInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->DlpiInfo.Pos = 0;
	dg->DlpiInfo.End = 0;
             
    if (NwDlpiDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_ADDLIST, "nwdlpi: could not add to list");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
	if (DlpiPromisc(dg->DlpiInfo.Socket, DL_PROMISC_PHYS) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_SETUP, "nwdlpi: promiscuous mode set failed");
    	NwDlpiCloseDg(dg);
        return NULL;
	}

	if (DlpiPromisc(dg->DlpiInfo.Socket, DL_PROMISC_SAP) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_SETUP, "nwdlpi: promiscuous mode set failed");
    	NwDlpiCloseDg(dg);
        return NULL;
	}

    if (ioctl(dg->DlpiInfo.Socket, I_FLUSH, (VOID *)FLUSHR) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: read flush failed");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* Opens a socket for receiving and sending packets from and to a specific host
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all. The Ethernet address is recorded 
 * and the receive callback is only called for packets to or from this address
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwDlpiConnect(char *addr            /* connect specification: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen:(int[6])ethernet' */, 
                     NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                     VOID *param           /* extra parameter you want to forward to recv() */)
{   
NwDlpiDg *dg;
CHAR interface[NWDLPI_SIZE_INTERFACE];
ULONG bufsize, snaplen;
TIMEVAL timeout;
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (strncmp(interface, NWDLPI_INTERFACE, strlen(NWDLPI_INTERFACE)) != 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid interface");
	    return NULL;
	}
	
	if (sscanf(p, ":%lu:%lu:%u.%u.%u.%u.%u.%u", &bufsize, &snaplen, &a1, &a2, &a3, &a4, &a5, &a6) != 8)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameters");
	    return NULL;
	}
	
    /* set buffer timeout at 1/10 of a second */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwDlpiDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWDLPI_NEWDLPI, "nwdlpi: memory allocation failed");
		return NULL;
    }

    if ((dg->DlpiInfo.Socket = open(NWDLPI_DEVICE, O_RDWR)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_OPEN, "nwdlpi: open failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }

    if (NwDlpiSetup(dg, interface, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SETUP, "nwdlpi: setup failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->DlpiInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwDlpiBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_BLOCK, "nwdlpi: block registration failed");
    	NwDlpiCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &DlpiDgIface;                                
    sprintf(dg->NwInfo.addr, "dlpi:%s:%lu:%lu:%u.%u.%u.%u.%u.%u", 
                             interface, bufsize, snaplen, a1, a2, a3, a4, a5, a6);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
	
	strcpy(dg->DlpiInfo.Interface, interface);
    dg->DlpiInfo.Address[0] = a1;
    dg->DlpiInfo.Address[1] = a2;
    dg->DlpiInfo.Address[2] = a3;
    dg->DlpiInfo.Address[3] = a4;
    dg->DlpiInfo.Address[4] = a5;
    dg->DlpiInfo.Address[5] = a6;
	dg->DlpiInfo.BufSize = bufsize;
	dg->DlpiInfo.SnapLen = snaplen;
	dg->DlpiInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->DlpiInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->DlpiInfo.Pos = 0;
	dg->DlpiInfo.End = 0;
             
    if (NwDlpiDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_ADDLIST, "nwdlpi: could not add to list");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
	if (DlpiPromisc(dg->DlpiInfo.Socket, DL_PROMISC_SAP) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_SETUP, "nwdlpi: promiscuous mode set failed");
    	NwDlpiCloseDg(dg);
        return NULL;
	}

    if (ioctl(dg->DlpiInfo.Socket, I_FLUSH, (VOID *)FLUSHR) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: read flush failed");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* Closes a NwDlpi datagram socket
 * Closes the specified NwDlpi datagram socket
 * Returns TRUE for success and FALSE for failure
 */
BOOLEAN NwDlpiClose(NW_DG *dg   /* an existing NwDlpi datagram socket */)
{                   
NwDlpiDg *dlpidg = (NwDlpiDg *)dg;

    NwDlpiDgRemoveList(dlpidg);
    NwDlpiCloseDg(dlpidg);

    return TRUE;
}


BOOLEAN NwDlpiCloseDg(NwDlpiDg *dg)
{
    if (dg->Block != NULL)    
        BlockDeRegister(dg->Block);
    
    if (dg->DlpiInfo.Socket != 0)
        close(dg->DlpiInfo.Socket);
        
    DelNwDlpiDg(dg);  
    
    return TRUE;
}    
    

/* Sends a packet over an existing NW datagram
 * The destination address is checked against the recorded Ethernet address
 * to allow only packets that are sent to that address
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiSend(NW_DG *dg    /* valid NW datagram */, 
                   BYTE *frame  /* packet */, 
                   INT len      /* packet length */)
{
NwDlpiDg *dlpidg = (NwDlpiDg *) dg;	
BYTE noaddr[NWDLPI_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (len > NWDLPI_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: packet too long");
        return FALSE;
    }

    if (memcmp(dlpidg->DlpiInfo.Address, noaddr, NWDLPI_SIZE_ADDRESS) == 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: can not send over accept socket");
        return FALSE;
    }
            
    if (DlpiSend(dlpidg->DlpiInfo.Socket, frame, len) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: send failed");
        return FALSE;
    }

    if (ioctl(dlpidg->DlpiInfo.Socket, I_FLUSH, (VOID *)FLUSHW) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: write flush failed");
        return FALSE;
    }
        
    return TRUE;
}


/* Sends a packet to a specified address
 * This function can be used to sent one-shot only packets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiSendTo(NW_DG *dg    /* not supported, set to NULL */, 
                     char *addr   /* address specification: 'dlpi:(char*)interface[:(int[6])ethernet]' */, 
                     BYTE *frame  /* packet */, 
                     INT len      /* packet length */)
{
CHAR interface[NWDLPI_SIZE_INTERFACE];
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;
BYTE address[NWDLPI_SIZE_ADDRESS];
         
    if (dg != NULL)
        DnpapMessage(DMC_WARNING, NWDLPI_SEND, "nwdlpi: sendto over existing connection has no effect");
    
    if (dlpisocket == 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_INIT, "nwdlpi: not initialized");
        return FALSE;
    }
    
    if (len > NWDLPI_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: packet too long");
        return FALSE;
    }

    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameter list");
        return FALSE;
    }   
    									  
	if ((p = strchr(++addr, ':')) == NULL)
	{
		p = addr+strlen(addr);
	}
	else
	{
		if (sscanf(p, ":%u.%u.%u.%u.%u.%u", &a1, &a2, &a3, &a4, &a5, &a6) != 6)
		{
	        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameters");
		    return FALSE;
		}
		
	    address[0] = a1;	
	    address[1] = a2;	
	    address[2] = a3;	
	    address[3] = a4;	
	    address[4] = a5;	
	    address[5] = a6;	
		
	    memcpy(frame, address, NWDLPI_SIZE_ADDRESS);
    }

    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (strncmp(interface, NWDLPI_INTERFACE, strlen(NWDLPI_INTERFACE)) != 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid interface");
	    return FALSE;
	}
	
    DlpiDetach(dlpisocket, interface);
    DlpiUnBind(dlpisocket, interface);

    if (DlpiAttach(dlpisocket, interface) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: interface attach failed");
		return FALSE;
	}
    
    if (DlpiBind(dlpisocket, interface) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: interface bind failed");
		return FALSE;
	}

    if (strioctl(dlpisocket, DLIOCRAW, NWDLPI_INFTIME, 0, NULL) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: raw mode set failed");
		return FALSE;
	}

    if (DlpiSend(dlpisocket, frame, len) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: send failed");
        return FALSE;
    }

    if (ioctl(dlpisocket, I_FLUSH, (VOID *)FLUSHW) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: write flush failed");
        return FALSE;
    }
        
    return TRUE;
}


BOOLEAN DlpiSend(INT socket, BYTE *frame, INT len)
{
    if (write(socket, frame, len) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: write failed");
        return FALSE;
    }
    
    return TRUE;
}


BOOLEAN NwDlpiIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{					 
NwDlpiDg *dlpidg = (NwDlpiDg *) dg;
INFO_ACK *info;

    if (DlpiInfo(dlpidg->DlpiInfo.Socket, dlpibuf) == FALSE)
		return FALSE;
	
	info = (INFO_ACK *) dlpibuf;
	
	if (info->dl_addr_length == 0)
		return FALSE;
		
	switch (info->dl_mac_type)
	{
	case DL_CSMACD:
	case DL_ETHER:
		ifconfig->type = MAC_TYPE_ETHERNET_CSMACD;
		ifconfig->speed = MAC_SPEED_ETHERNET_CSMACD;
		break;
	default:
		DnpapMessage(DMC_ERROR, NWDLPI_CONFIG, "nwdlpi: unknown interface type");
		return FALSE;
	}
	
	ifconfig->mtu = info->dl_max_sdu;
	
	if (info->dl_sap_length > 0)
		ifconfig->addrlen = info->dl_addr_length - info->dl_sap_length;
	else
		ifconfig->addrlen = info->dl_addr_length + info->dl_sap_length;
		
	memcpy(ifconfig->address, dlpibuf + info->dl_addr_offset, ifconfig->addrlen);
	memcpy(ifconfig->broadcast, dlpibuf + info->dl_brdcst_addr_offset, ifconfig->addrlen);
	
	return TRUE;
}


int NwDlpiBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwDlpiDg *dg = (NwDlpiDg *) param;	
	
	switch (reason)
	{
	case BLOCKEVENT_READ:
		if (NwDlpiRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWDLPI_READ, "nwdlpi: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWDLPI_WEVENT, "nwdlpi: write event");
		break;
	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWDLPI_EEVENT, "nwdlpi: except event");
		break;
	default:
		DnpapMessage(DMC_ERROR, NWDLPI_NEVENT, "nwdlpi: unknown block event");
		return 1;
	}
	
	return 0;
}


BOOLEAN NwDlpiRead(NwDlpiDg *dg)
{
BYTE *cp;
SB_HDR *hp;
ULONG msglen;
BYTE noaddr[NWDLPI_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (!(dg->DlpiInfo.Pos < dg->DlpiInfo.End))
    {
		if ((dg->DlpiInfo.End = read(dg->DlpiInfo.Socket, dg->DlpiInfo.Buf, dg->DlpiInfo.BufSize)) < 0)
			return FALSE;

    	dg->DlpiInfo.Pos = 0;
    }

	while (dg->DlpiInfo.Pos < dg->DlpiInfo.End)
	{
		cp = dg->DlpiInfo.Buf+dg->DlpiInfo.Pos;
		hp = (SB_HDR *)cp;      
	    msglen = hp->sbh_msglen;
		if (dg->DlpiInfo.SnapLen > 0 && msglen > dg->DlpiInfo.SnapLen)
			msglen = dg->DlpiInfo.SnapLen;

		dg->DlpiInfo.Pos += hp->sbh_totlen;

        if (memcmp(dg->DlpiInfo.Address, noaddr, NWDLPI_SIZE_ADDRESS) != 0 &&
            memcmp(dg->DlpiInfo.Address, cp, NWDLPI_SIZE_ADDRESS) != 0 &&
            memcmp(dg->DlpiInfo.Address, cp+NWDLPI_SIZE_ADDRESS, NWDLPI_SIZE_ADDRESS) != 0)
            continue;
        
		if (dg->NwInfo.recv != NULL)
			dg->NwInfo.recv((NW_DG *)dg, dg->NwInfo.addr, cp, msglen, dg->NwInfo.parm);
	}
	
	return TRUE;
}


BOOLEAN NwDlpiDgAddList(NwDlpiDg *dg)
{
    if (NwDlpiDgList != NULL)
		NwDlpiDgList->Prev = dg;
	
    dg->Next = NwDlpiDgList;
    dg->Prev = NULL;
    
    NwDlpiDgList = dg;
        
    return TRUE;
}


BOOLEAN NwDlpiDgRemoveList(NwDlpiDg *dg)
{   
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwDlpiDgList = dg->Next;
        
    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;
        
    return TRUE;
}


NwDlpiDg* NewNwDlpiDg(CHAR *interface, ULONG bufsize)
{
NwDlpiDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwDlpiDg))) == NULL)
		return NULL;
	
	dg->Block = NULL;
    
    dg->NwInfo.iface = NULL;
	dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;    
    
    dg->DlpiInfo.Interface = NULL;
    memset(dg->DlpiInfo.Address, 0, NWDLPI_SIZE_ADDRESS);
    dg->DlpiInfo.Buf = NULL;
	dg->DlpiInfo.Socket = 0;		
	dg->DlpiInfo.BufSize = 0;
	dg->DlpiInfo.SnapLen = 0;
	dg->DlpiInfo.Timeout.tv_sec = 0;
	dg->DlpiInfo.Timeout.tv_usec = 0;
	dg->DlpiInfo.Pos = 0;
	dg->DlpiInfo.End = 0;
    
    dg->Next = NULL;
    dg->Prev = NULL;
	
    if ((dg->Block = DnpapMalloc(sizeof(BLOCK))) == NULL)
		return NULL;
    
	if ((dg->DlpiInfo.Interface = DnpapMalloc(strlen(interface)+1)) == NULL)
    {
    	DnpapFree(dg->Block);
		return NULL;
	}
	
	if ((dg->DlpiInfo.Buf = DnpapMalloc(bufsize)) == NULL)
    {
    	DnpapFree(dg->Block);
    	DnpapFree(dg->DlpiInfo.Interface);
		return NULL;
	}
	
	return dg;
}


VOID DelNwDlpiDg(NwDlpiDg *dg)
{
	DnpapFree(dg->Block);
	
    DnpapFree(dg->DlpiInfo.Interface);
    DnpapFree(dg->DlpiInfo.Buf);
    
    DnpapFree(dg);
}


BOOLEAN NwDlpiSetup(NwDlpiDg *dg, CHAR *interface, ULONG bufsize, ULONG snaplen, TIMEVAL *timeout)
{
    /* arrange to get discrete messages from the stream */
    if (ioctl(dg->DlpiInfo.Socket, I_SRDOPT, (VOID *)RMSGN) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: discrete message option set failed");
		return FALSE;
	}

    if (DlpiAttach(dg->DlpiInfo.Socket, interface) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: interface attach failed");
		return FALSE;
	}
    
    if (DlpiBind(dg->DlpiInfo.Socket, interface) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: interface bind failed");
		return FALSE;
	}

    if (DlpiInfo(dg->DlpiInfo.Socket, dlpibuf) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: info retrieval failed");
		return FALSE;
	}

    if (strioctl(dg->DlpiInfo.Socket, DLIOCRAW, NWDLPI_INFTIME, 0, NULL) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: raw mode set failed");
		return FALSE;
	}

    /* push and configure the filtering module */
    if (ioctl(dg->DlpiInfo.Socket, I_PUSH, NWDLPI_FILTER) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PUSH, "nwdlpi: 'pfmod' push failed");
		return FALSE;
	}

    /* push and configure the buffering module */
    if (ioctl(dg->DlpiInfo.Socket, I_PUSH, NWDLPI_BUFFER) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PUSH, "nwdlpi: 'bufmod' push failed");
		return FALSE;
	}

    if (strioctl(dg->DlpiInfo.Socket, SBIOCSCHUNK, NWDLPI_INFTIME, sizeof(bufsize), (VOID *)&bufsize) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: chunksize option set failed");
		return FALSE;
	}

	/* timeout settings must come after bufsize settings:
	 * if timeout == 0 then bufsize is reset too
	 */
    if (strioctl(dg->DlpiInfo.Socket, SBIOCSTIME, NWDLPI_INFTIME, sizeof(*timeout), (VOID *)timeout) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: timeout option set failed");
		return FALSE;
	}
    
    if (snaplen > 0)
    {
        if (strioctl(dg->DlpiInfo.Socket, SBIOCSSNAP, NWDLPI_INFTIME, sizeof(snaplen), (VOID *)&snaplen) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: snapshot length option set failed");
    		return FALSE;
    	}
    }

    return TRUE;
}


BOOLEAN DlpiBind(INT socket, CHAR *interface)
{
BIND_REQ req;
STRBUF ctl;

	req.dl_primitive = DL_BIND_REQ;
	req.dl_sap = 0x800;			/* promiscuous mode later on */
	req.dl_max_conind = 0;
	req.dl_service_mode = DL_CLDLS;
	req.dl_conn_mgmt = 0;
	req.dl_xidtest_flg = 0;

	ctl.maxlen = 0;
	ctl.len = sizeof(req);
	ctl.buf = (char *)&req;

	if (putmsg(socket, &ctl, NULL, 0) < 0)
	{
		DnpapMessage(DMC_FATAL, NWDLPI_PUTMSG, "nwdlpi: putmsg failed");
		return FALSE;
	}

	return DlpiAck(socket, DL_BIND_ACK, dlpibuf);
}


BOOLEAN DlpiUnBind(INT socket, CHAR *interface)
{
UNBIND_REQ req;
STRBUF ctl;

	req.dl_primitive = DL_UNBIND_REQ;

	ctl.maxlen = 0;
	ctl.len = sizeof(req);
	ctl.buf = (char *)&req;

	if (putmsg(socket, &ctl, NULL, 0) < 0)
	{
		DnpapMessage(DMC_FATAL, NWDLPI_PUTMSG, "nwdlpi: putmsg failed");
		return FALSE;
	}

	return DlpiAck(socket, DL_OK_ACK, dlpibuf);
}


BOOLEAN DlpiAttach(INT socket, CHAR *interface)
{
ATTACH_REQ req;
STRBUF ctl;

	req.dl_primitive = DL_ATTACH_REQ;
	req.dl_ppa = atoi(interface+strlen(NWDLPI_INTERFACE));

	ctl.maxlen = 0;
	ctl.len = sizeof(req);
	ctl.buf = (char *)&req;

	if (putmsg(socket, &ctl, NULL, 0) < 0)
	{
		DnpapMessage(DMC_FATAL, NWDLPI_PUTMSG, "nwdlpi: putmsg failed");
		return FALSE;
	}

	return DlpiAck(socket, DL_OK_ACK, dlpibuf);
}

    
BOOLEAN DlpiDetach(INT socket, CHAR *interface)
{
DETACH_REQ req;
STRBUF ctl;

	req.dl_primitive = DL_DETACH_REQ;

	ctl.maxlen = 0;
	ctl.len = sizeof(req);
	ctl.buf = (char *)&req;

	if (putmsg(socket, &ctl, NULL, 0) < 0)
	{
		DnpapMessage(DMC_FATAL, NWDLPI_PUTMSG, "nwdlpi: putmsg failed");
		return FALSE;
	}

	return DlpiAck(socket, DL_OK_ACK, dlpibuf);
}

    
BOOLEAN DlpiInfo(INT socket, BYTE *buf)
{
INFO_REQ req;
STRBUF ctl;

	req.dl_primitive = DL_INFO_REQ;

	ctl.maxlen = 0;
	ctl.len = sizeof(req);
	ctl.buf = (char *)&req;

	if (putmsg(socket, &ctl, NULL, RS_HIPRI) < 0)
	{
		DnpapMessage(DMC_FATAL, NWDLPI_PUTMSG, "nwdlpi: putmsg failed");
		return FALSE;
	}

	return DlpiAck(socket, DL_INFO_ACK, buf);
}


BOOLEAN DlpiPromisc(INT socket, INT level)
{
PROMISCON_REQ req;
STRBUF ctl;

	req.dl_primitive = DL_PROMISCON_REQ;
	req.dl_level = level;

	ctl.maxlen = 0;
	ctl.len = sizeof(req);
	ctl.buf = (char *)&req;

	if (putmsg(socket, &ctl, NULL, 0) < 0)
	{
		DnpapMessage(DMC_FATAL, NWDLPI_PUTMSG, "nwdlpi: putmsg failed");
		return FALSE;
	}

	return DlpiAck(socket, DL_OK_ACK, dlpibuf);
}


BOOLEAN DlpiAck(INT socket, ULONG prim, BYTE *buf)
{
DL_PRIMITIVES *dlp;
STRBUF ctl;
INT flags = 0;

	ctl.maxlen = NWDLPI_SIZE_DLBUF;
	ctl.len = 0;
	ctl.buf = buf;

	if (strgetmsg(socket, &ctl, NULL, &flags) < 0)
		return FALSE;

	dlp = (DL_PRIMITIVES *)ctl.buf;
	if (dlp->dl_primitive != prim) 
	{
		DnpapMessage(DMC_ERROR, NWDLPI_PRIM, "nwdlpi: expected primitive %s but got %s", 
			DlpiPrim(prim), DlpiPrim(dlp->dl_primitive));
		return FALSE;
	}

	if (flags != RS_HIPRI)
	{
		DnpapMessage(DMC_ERROR, NWDLPI_ACK, "nwdlpi: wrong type of acknowledge message");
		return FALSE;
	}

	switch (prim)
	{
	case DL_OK_ACK:
		if (ctl.len < sizeof(OK_ACK))
		{
			DnpapMessage(DMC_ERROR, NWDLPI_ACK, "nwdlpi: response length too short");
			return FALSE;
		}
		break;
	case DL_BIND_ACK:
		if (ctl.len < sizeof(BIND_ACK))
		{
			DnpapMessage(DMC_ERROR, NWDLPI_ACK, "nwdlpi: response length too short");
			return FALSE;
		}
		break;
	case DL_INFO_ACK:
		if (ctl.len < sizeof(INFO_ACK))
		{
			DnpapMessage(DMC_ERROR, NWDLPI_ACK, "nwdlpi: response length too short");
			return FALSE;
		}
		break;
	default:
		DnpapMessage(DMC_ERROR, NWDLPI_ACK, "nwdlpi: unsupported acknowledgement");
		return FALSE;
		break;
	}

	return TRUE;
}


CHAR *DlpiPrim(ULONG prim)
{
	switch (prim)
	{
	case DL_INFO_REQ:
		return "DL_INFO_REQ";
		break;
	case DL_ATTACH_REQ:
		return "DL_ATTACH_REQ";
		break;
	case DL_DETACH_REQ:
		return "DL_DETACH_REQ";
		break;
	case DL_BIND_REQ:
		return "DL_BIND_REQ";
		break;
	case DL_UNBIND_REQ:
		return "DL_UNBIND_REQ";
		break;
	case DL_SUBS_BIND_REQ:
		return "DL_SUBS_BIND_REQ";
		break;
	case DL_SUBS_UNBIND_REQ:
		return "DL_SUBS_UNBIND_REQ";
		break;
	case DL_ENABMULTI_REQ:
		return "DL_ENABMULTI_REQ";
		break;
	case DL_DISABMULTI_REQ:
		return "DL_DISABMULTI_REQ";
		break;
	case DL_PROMISCON_REQ:
		return "DL_PROMISCON_REQ";
		break;
	case DL_PROMISCOFF_REQ:
		return "DL_PROMISCOFF_REQ";
		break;
	case DL_UNITDATA_REQ:
		return "DL_UNITDATA_REQ";
		break;
	case DL_UDQOS_REQ:
		return "DL_UDQOS_REQ";
		break;
	case DL_CONNECT_REQ:
		return "DL_CONNECT_REQ";
		break;
	case DL_TOKEN_REQ:
		return "DL_TOKEN_REQ";
		break;
	case DL_DISCONNECT_REQ:
		return "DL_DISCONNECT_REQ";
		break;
	case DL_RESET_REQ:
		return "DL_RESET_REQ";
		break;
	case DL_DATA_ACK_REQ:
		return "DL_DATA_ACK_REQ";
		break;
	case DL_REPLY_REQ:
		return "DL_REPLY_REQ";
		break;
	case DL_REPLY_UPDATE_REQ:
		return "DL_REPLY_UPDATE_REQ";
		break;
	case DL_XID_REQ:
		return "DL_XID_REQ";
		break;
	case DL_TEST_REQ:
		return "DL_TEST_REQ";
		break;
	case DL_PHYS_ADDR_REQ:
		return "DL_PHYS_ADDR_REQ";
		break;
	case DL_SET_PHYS_ADDR_REQ:
		return "DL_SET_PHYS_ADDR_REQ";
		break;
	case DL_GET_STATISTICS_REQ:
		return "DL_GET_STATISTICS_REQ";
		break;
	}

	return "unknown";
}


VOID SigAlarm(INT signal)
{
	DnpapMessage(DMC_WARNING, NWDLPI_ALARM, "nwdlpi: alarm timeout");
}


int strioctl(int socket, int cmd, int timeout, int len, BYTE *dp)
{
STRIOCTL si;
int ret;

	si.ic_cmd = cmd;
	si.ic_timout = timeout;
	si.ic_len = len;
	si.ic_dp = dp;
	if ((ret = ioctl(socket, I_STR, &si)) < 0)
		return ret;
	return si.ic_len;		
}


int strgetmsg(int socket, STRBUF *ctl, STRBUF *data, INT *flags)
{
	int	rc;

	/*
	 * Start timer.
	 */
	signal(SIGALRM, SigAlarm);
	if (alarm(NWDLPI_MAXWAIT) != 0) 
	{
		DnpapMessage(DMC_ERROR, NWDLPI_ALARM, "nwdlpi: could not start alarm");
		return -1;
	}

	/*
	 * Set flags argument and issue getmsg().
	 */
	*flags = 0;
	if ((rc = getmsg(socket, ctl, data, flags)) < 0) 
	{
		DnpapMessage(DMC_ERROR, NWDLPI_GETMSG, "nwdlpi: getmsg failed");
		return -1;
	}

	/*
	 * Stop timer.
	 */
	alarm(0);

	/*
	 * Check for MOREDATA and/or MORECTL.
	 */
	if ((rc & (MORECTL | MOREDATA)) == (MORECTL | MOREDATA))
	{
		DnpapMessage(DMC_ERROR, NWDLPI_MORE, "nwdlpi: more control or more data");
		return -1;
	}
	if (rc & MORECTL)
	{
		DnpapMessage(DMC_ERROR, NWDLPI_MORE, "nwdlpi: more control");
		return -1;
	}
	if (rc & MOREDATA)
	{
		DnpapMessage(DMC_ERROR, NWDLPI_MORE, "nwdlpi: more data");
		return -1;
	}

	/*
	 * Check for at least sizeof (long) control data portion.
	 */
	if (ctl->len < sizeof(long))
	{
		DnpapMessage(DMC_FATAL, NWDLPI_LENGTH, "nwdlpi: control portion length (%d) < sizeof(long)", ctl->len);
		return -1;
	}

	return 0;
}
