/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <linux/if_ether.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netdb.h>
               
#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nweth.h"
 
            
/* nweth.c - NwEth module functions */

            
typedef struct _NwEth
{
	INT   Socket;
	CHAR *Interface;
    BYTE  Address[NWETH_SIZE_ADDRESS];
	INT   Flags;
	INT   OldFlags;
	BYTE  Buf[NWETH_SIZE_PACKET];
	INT   BufSize;
	LONG  Pos;
	LONG  End;
	         
} NwEth;
    

typedef struct _NwEthDg
{          
    NW_DG  NwInfo;
    NwEth  EthInfo;
    
    BLOCK           *Block;
    
    struct _NwEthDg *Prev;
    struct _NwEthDg *Next;
    
} NwEthDg;

      
static BOOLEAN NwEthInit(VOID);
static BOOLEAN NwEthExit(VOID);

static NW_DG  *NwEthAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwEthConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwEthSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwEthSendTo(NW_DG *dg, char *addr, BYTE *frame, INT len);
static BOOLEAN NwEthClose(NW_DG *dg);
static BOOLEAN NwEthIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwEthStatistics(char *addr);


NW_DG_IFACE EthDgIface = { "eth", 
                           NwEthInit,   NwEthExit, 
                           NwEthAccept, NwEthConnect, 
                           NwEthSend,   NwEthSendTo,
                           NwEthClose,  NwEthIfConfig,
                           NwEthStatistics
                        };


static NwEthDg *NwEthDgList = NULL;

static NwStat nwStat;


static INT ethsocket;

static BOOLEAN NwEthCloseDg(NwEthDg *dg);

static BOOLEAN NwEthDgAddList(NwEthDg *dg);
static BOOLEAN NwEthDgRemoveList(NwEthDg *dg);

static NwEthDg* NewNwEthDg(CHAR *interface);
static VOID DelNwEthDg(NwEthDg *dg);

static BOOLEAN NwEthRead(NwEthDg *dg);

static BOOLEAN NwEthSetup(NwEthDg *dg, CHAR *interface, INT flags);
static BOOLEAN NwEthRestore(NwEthDg *dg);

static int NwEthBlockCallback(BLOCK *blk, int reason, VOID *param);



/* Initializes the NwEth module
 * Initializes the NwEth messages and opens a socket for general use
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEthInit(VOID)
{       
    if ((ethsocket = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_OPEN, "nweth: open failed");
        return FALSE;
    }

    return TRUE;
}


/* Exits the NwEth module
 * Closes all opened sockets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEthExit(VOID)
{             
BOOLEAN rc = TRUE;
NwEthDg *dg;

    for (dg = NwEthDgList; dg != NULL; dg = NwEthDgList)
        rc &= NwEthClose((NW_DG *)dg);

    if (ethsocket != 0)
        close(ethsocket);
                
    return rc;
}


NwStat *NwEthStatistics(char *addr)
{            
    return &nwStat;
}

        
/* Opens a socket for receiving promiscuous packets
 * Opens the specified interface in promiscuous mode
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwEthAccept(char *addr            /* accept specification: 'eth:(char*)interface' */,
                  NW_DG_RECV_CALL recv  /* receive callback */,                                                         
                  VOID *param           /* extra parameter you want to be forwarded to recv() */)                       
{   
NwEthDg *dg;
CHAR interface[NWETH_SIZE_INTERFACE];
INT flags;

    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWETH_PARAM, "nweth: invalid parameter list");
        return NULL;
    }   
    
    strcpy(interface, ++addr);
    interface[strlen(interface)] = '\0';
    
	flags = IFF_PROMISC | NWETH_DEFAULT_FLAGS;
	
	if ((dg = NewNwEthDg(interface)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWETH_NEWETH, "nweth: memory allocation failed");
		return NULL;
    }

    if ((dg->EthInfo.Socket = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_OPEN, "nweth: open failed");
    	NwEthCloseDg(dg);
        return NULL;
    }

    if (NwEthSetup(dg, interface, flags) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWETH_SETUP, "nweth: setup failed");
    	NwEthCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->EthInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwEthBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_BLOCK, "nweth: block registration failed");
    	NwEthCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &EthDgIface;                                
    sprintf(dg->NwInfo.addr, "eth:%s", interface);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
    
	strcpy(dg->EthInfo.Interface, interface);
	dg->EthInfo.Flags = flags;
	dg->EthInfo.Pos = 0;
	dg->EthInfo.End = 0;
             
    if (NwEthDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWETH_ADDLIST, "nweth: could not add to list");
    	NwEthCloseDg(dg);
		return NULL;
	}
    
/*	if (ioctl(dg->EthInfo.Socket, SIOCFLUSH, 0) < 0)
	{
        DnpapMessage(DMC_ERROR, NWETH_IOCTL, "nweth: read flush failed");
    	NwEthCloseDg(dg);
		return NULL;
	}
 */
    
    return (NW_DG *)dg;
}


/* Opens a socket for receiving and sending packets from and to a specific host
 * Opens the specified interface in promiscuous mode
 * The Ethernet address is recorded and the receive callback is only 
 * called for packets to or from this address
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwEthConnect(char *addr            /* connect specification: 'eth:(char*)interface:(int[6])ethernet' */, 
                    NW_DG_RECV_CALL recv  /* receive callback */, 
                    VOID *param           /* extra parameter you want to forward to recv() */)
{   
NwEthDg *dg;
CHAR interface[NWETH_SIZE_INTERFACE];
INT flags;
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWETH_PARAM, "nweth: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (sscanf(p, ":%u.%u.%u.%u.%u.%u", &a1, &a2, &a3, &a4, &a5, &a6) != 6)
	{
        DnpapMessage(DMC_ERROR, NWETH_PARAM, "nweth: invalid parameters");
	    return NULL;
	}
	
	flags = NWETH_DEFAULT_FLAGS;
	
	if ((dg = NewNwEthDg(interface)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWETH_NEWETH, "nweth: memory allocation failed");
		return NULL;
    }

    if ((dg->EthInfo.Socket = socket(AF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_OPEN, "nweth: open failed");
    	NwEthCloseDg(dg);
        return NULL;
    }

    if (NwEthSetup(dg, interface, flags) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWETH_SETUP, "nweth: setup failed");
    	NwEthCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->EthInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwEthBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_BLOCK, "nweth: block registration failed");
    	NwEthCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &EthDgIface;                                
    sprintf(dg->NwInfo.addr, "eth:%s:%u.%u.%u.%u.%u.%u", 
                             interface, a1, a2, a3, a4, a5, a6);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
	
	strcpy(dg->EthInfo.Interface, interface);
    dg->EthInfo.Address[0] = a1;
    dg->EthInfo.Address[1] = a2;
    dg->EthInfo.Address[2] = a3;
    dg->EthInfo.Address[3] = a4;
    dg->EthInfo.Address[4] = a5;
    dg->EthInfo.Address[5] = a6;
	dg->EthInfo.Flags = flags;
	dg->EthInfo.Pos = 0;
	dg->EthInfo.End = 0;
             
    if (NwEthDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWETH_ADDLIST, "nweth: could not add to list");
    	NwEthCloseDg(dg);
		return NULL;
	}
    
/*  if (ioctl(dg->EthInfo.Socket, EIOCFLUSH, 0) < 0)
	{
        DnpapMessage(DMC_ERROR, NWETH_IOCTL, "nweth: read flush failed");
    	NwEthCloseDg(dg);
		return NULL;
	}
 */
    
    return (NW_DG *)dg;
}


BOOLEAN NwEthClose(NW_DG *dg)
{                   
NwEthDg *ethdg = (NwEthDg *)dg;

	NwEthRestore(ethdg);
    NwEthDgRemoveList(ethdg);
    NwEthCloseDg(ethdg);

    return TRUE;
}


BOOLEAN NwEthCloseDg(NwEthDg *dg)
{
    if (dg->Block != NULL)    
        BlockDeRegister(dg->Block);
    
    if (dg->EthInfo.Socket != 0)
        close(dg->EthInfo.Socket);
        
    DelNwEthDg(dg);  
    
    return TRUE;
}    
    

/* Sends a packet over an existing NW datagram
 * The destination address is checked against the recorded Ethernet address
 * to allow only packets that are sent to that address
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEthSend(NW_DG *dg    /* valid NW datagram */, 
                 BYTE *frame  /* packet */, 
                 INT len      /* packet length */)
{
NwEthDg *ethdg = (NwEthDg *) dg;	
BYTE noaddr[NWETH_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (len > NWETH_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWETH_SEND, "nweth: packet too long");
        return FALSE;
    }

    if (memcmp(ethdg->EthInfo.Address, noaddr, NWETH_SIZE_ADDRESS) == 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_SEND, "nweth: can not send over accept socket");
        return FALSE;
    }
            
    memcpy(frame, ethdg->EthInfo.Address, NWETH_SIZE_ADDRESS);

    if (write(ethdg->EthInfo.Socket, frame, len) < 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_SEND, "nweth: write failed");
        return FALSE;
    }

    return TRUE;
}


/* Sends a packet to a specified address
 * This function can be used to sent one-shot only packets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEthSendTo(NW_DG *dg    /* not supported, set to NULL */, 
                   char *addr   /* address specification: 'eth:(char*)interface:(int[6])ethernet' */, 
                   BYTE *frame  /* packet */, 
                   INT len      /* packet length */)
{
SOCKADDR sa;
CHAR interface[NWETH_SIZE_INTERFACE];
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;
BYTE address[NWETH_SIZE_ADDRESS];
         
    if (dg != NULL)
        DnpapMessage(DMC_WARNING, NWETH_SEND, "nweth: sendto over existing connection has no effect");
    
    if (ethsocket == 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_INIT, "nweth: not initialized");
        return FALSE;
    }
    
    if (len > NWETH_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWETH_SEND, "nweth: packet too long");
        return FALSE;
    }

    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWETH_PARAM, "nweth: invalid parameter list");
        return FALSE;
    }   
    
	if ((p = strchr(++addr, ':')) == NULL)
	{
		p = addr+strlen(addr);
	}
	else
	{
    	if (sscanf(p, ":%u.%u.%u.%u.%u.%u", &a1, &a2, &a3, &a4, &a5, &a6) != 6)
    	{
            DnpapMessage(DMC_ERROR, NWETH_PARAM, "nweth: invalid parameters");
    	    return FALSE;
    	}
    
        address[0] = a1;	
        address[1] = a2;	
        address[2] = a3;	
        address[3] = a4;	
        address[4] = a5;	
        address[5] = a6;	
    
        memcpy(frame, address, NWETH_SIZE_ADDRESS);
    }
    
	sa.sa_family = AF_INET;
    strncpy(sa.sa_data, addr, p-addr);
    sa.sa_data[p-addr] = '\0';

    if (sendto(ethsocket, frame, len, 0, &sa, sizeof(sa)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWETH_SEND, "nweth: sendto failed");
        return FALSE;
    }

    return TRUE;
}


BOOLEAN NwEthIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{
NwEthDg *ethdg = (NwEthDg *) dg;	
IFREQ ifr;

    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, ethdg->EthInfo.Interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';

    if (ioctl(ethdg->EthInfo.Socket, SIOCGIFHWADDR, (VOID *)&ifr) < 0)
		return FALSE;
	
    if (ioctl(ethdg->EthInfo.Socket, SIOCGIFMTU, (VOID *)&ifr) < 0)
		return FALSE;
	
	if (strncmp(ethdg->EthInfo.Interface, "eth", strlen("eth")) == 0)
	{	
		ifconfig->type = MAC_TYPE_ETHERNET_CSMACD;
		ifconfig->mtu = ifr.ifr_mtu;
		ifconfig->speed = MAC_SPEED_ETHERNET_CSMACD;
		ifconfig->addrlen = MAC_ADDRLEN_ETHERNET_CSMACD;
		memcpy(ifconfig->address, ifr.ifr_addr.sa_data, ifconfig->addrlen);
		memcpy(ifconfig->broadcast, "\xff\xff\xff\xff\xff\xff", ifconfig->addrlen);
	}
	else
	{
		DnpapMessage(DMC_ERROR, NWETH_CONFIG, "nweth: unknown interface type");
		return FALSE;
	}
		
	return TRUE;
}


int NwEthBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwEthDg *dg = (NwEthDg *) param;	
	
	switch (reason)
	{
	case BLOCKEVENT_READ:
		if (NwEthRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWETH_READ, "nweth: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWETH_WEVENT, "nweth: write event");
		break;
	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWETH_EEVENT, "nweth: except event");
		break;
	default:
		DnpapMessage(DMC_ERROR, NWETH_NEVENT, "nweth: unknown block event");
		return 1;
	}
	
	return 0;
}


BOOLEAN NwEthRead(NwEthDg *dg)
{
BYTE *cp;
int msglen;
SOCKADDR from;
int fromlen = sizeof(from);
BYTE noaddr[NWETH_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

	if ((msglen = recvfrom(dg->EthInfo.Socket, dg->EthInfo.Buf+sizeof(TIMEVAL), dg->EthInfo.BufSize, 0, &from, &fromlen)) < 0)
		return FALSE;
	/* old Linux kernels do not implement SIOCGSTAMP for SOCK_PACKET sockets */
	if (ioctl(dg->EthInfo.Socket, SIOCGSTAMP, dg->EthInfo.Buf) < 0)
	{
		gettimeofday((TIMEVAL *) dg->EthInfo.Buf, NULL);
		return FALSE;
	}
	
	cp = dg->EthInfo.Buf;
	
	if (strncmp(dg->EthInfo.Interface, from.sa_data, strlen(dg->EthInfo.Interface)) != 0 ||
		(memcmp(dg->EthInfo.Address, noaddr, NWETH_SIZE_ADDRESS) != 0 &&
		memcmp(dg->EthInfo.Address, cp, NWETH_SIZE_ADDRESS) != 0 &&
		memcmp(dg->EthInfo.Address, cp+NWETH_SIZE_ADDRESS, NWETH_SIZE_ADDRESS) != 0))
		return TRUE;
	
	if (dg->NwInfo.recv != NULL)
		dg->NwInfo.recv((NW_DG *)dg, dg->NwInfo.addr, cp, msglen, dg->NwInfo.parm);
	
	return TRUE;
}


BOOLEAN NwEthDgAddList(NwEthDg *dg)
{
    if (NwEthDgList != NULL)
		NwEthDgList->Prev = dg;
        
    dg->Next = NwEthDgList;
    dg->Prev = NULL;
    
    NwEthDgList = dg;
        
    return TRUE;
}


BOOLEAN NwEthDgRemoveList(NwEthDg *dg)
{   
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwEthDgList = dg->Next;
        
    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;
        
    return TRUE;
}


NwEthDg* NewNwEthDg(CHAR *interface)
{
NwEthDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwEthDg))) == NULL)
		return NULL;
	
	dg->Block = NULL;
    
    dg->NwInfo.iface = NULL;
	dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;    
    
	dg->EthInfo.Socket = 0;		
    dg->EthInfo.Interface = NULL;
    memset(dg->EthInfo.Address, 0, NWETH_SIZE_ADDRESS);
	dg->EthInfo.Flags = 0;
	dg->EthInfo.OldFlags = 0;
	dg->EthInfo.BufSize = NWETH_SIZE_PACKET;
	dg->EthInfo.Pos = 0;
	dg->EthInfo.End = 0;
    
    dg->Next = NULL;
    dg->Prev = NULL;
	
    if ((dg->Block = DnpapMalloc(sizeof(BLOCK))) == NULL)
		return NULL;
    
	if ((dg->EthInfo.Interface = DnpapMalloc(strlen(interface)+1)) == NULL)
    {
    	DnpapFree(dg->Block);
		return NULL;
	}
	
	return dg;
}


VOID DelNwEthDg(NwEthDg *dg)
{
	DnpapFree(dg->Block);
	
    DnpapFree(dg->EthInfo.Interface);
    
    DnpapFree(dg);
}


BOOLEAN NwEthSetup(NwEthDg *dg, CHAR *interface, INT flags)
{
IFREQ ifr;

    flags |= NWETH_DEFAULT_FLAGS;
                   
    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
/*  if (ioctl(dg->EthInfo.Socket, SIOCSIFLINK, &ifr) < 0)
	{
        DnpapMessage(DMC_ERROR, NWETH_IOCTL, "nweth: interface bind failed");
		return FALSE;
	}
 */
    
    if (flags != 0)
    {
        if (ioctl(dg->EthInfo.Socket, SIOCGIFFLAGS, &ifr) < 0)
		{
            DnpapMessage(DMC_ERROR, NWETH_IOCTL, "nweth: flags get failed");
    		return FALSE;
		}
		dg->EthInfo.OldFlags = ifr.ifr_flags;
		ifr.ifr_flags |= flags;
        if (ioctl(dg->EthInfo.Socket, SIOCSIFFLAGS, &ifr) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWETH_IOCTL, "nweth: flags set failed");
    		return FALSE;
    	}
    }
    
    return TRUE;
}


BOOLEAN NwEthRestore(NwEthDg *dg)
{
IFREQ ifr;

	if (dg->EthInfo.OldFlags != 0)
	{
		memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
		strncpy(ifr.ifr_name, dg->EthInfo.Interface, sizeof(ifr.ifr_name));
		ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
		ifr.ifr_flags = dg->EthInfo.OldFlags;
		if (ioctl(dg->EthInfo.Socket, SIOCSIFFLAGS, &ifr) < 0)
		{
			DnpapMessage(DMC_ERROR, NWETH_IOCTL, "nweth: flags set failed");
			return FALSE;
		}
	}

	return TRUE;
}
