/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#if defined(SUNOS) || defined(SOLARIS)
#include <stropts.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
               
#include <dinet.h>
#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nwicmp.h"
 

/* nwicmp.c - NwIcmp module functions */

           
typedef struct _NwIcmp
{
	INT Socket;
    CHAR  *Host;
    ULONG  Addr;
    USHORT Ident;
    
} NwIcmp;
    

typedef struct _NwIcmpDg
{          
    NW_DG  NwInfo;
    NwIcmp IcmpInfo;
    
    BLOCK          *Block;
    
    struct _NwIcmpDg *Prev;
    struct _NwIcmpDg *Next;
    
} NwIcmpDg;

      

static BOOLEAN NwIcmpInit(VOID);
static BOOLEAN NwIcmpExit(VOID);

static NW_DG  *NwIcmpAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwIcmpConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwIcmpSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwIcmpSendTo(NW_DG *dg, CHAR *addr, BYTE *frame, INT len);
static BOOLEAN NwIcmpClose(NW_DG *dg);
static BOOLEAN NwIcmpIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwIcmpStatistics(char *addr);


NW_DG_IFACE IcmpDgIface = { "icmp", 
                            NwIcmpInit,   NwIcmpExit, 
                            NwIcmpAccept, NwIcmpConnect, 
                            NwIcmpSend,   NwIcmpSendTo,
                            NwIcmpClose,  NwIcmpIfConfig,
                            NwIcmpStatistics
                        };


static NwIcmpDg *NwIcmpDgList = NULL;

static PROTOENT *protoentptr;

static NwStat nwStat;


static INT icmpsocket;

static BOOLEAN NwIcmpCloseDg(NwIcmpDg *dg);

static BOOLEAN NwIcmpDgAddList(NwIcmpDg *dg);
static BOOLEAN NwIcmpDgRemoveList(NwIcmpDg *dg);

static NwIcmpDg* NewNwIcmpDg(CHAR *host);
static VOID DelNwIcmpDg(NwIcmpDg *dg);

static BOOLEAN NwIcmpRead(NwIcmpDg *dg);

static int NwIcmpBlockCallback(BLOCK *blk, int reason, VOID *param);



/* initializes the NwNit module
 * initializes the NwNit messages and opens a socket for general use
 * returns TRUE or FALSE for success or failure
 */
BOOLEAN NwIcmpInit(VOID)
{       
	if (geteuid() != 0)
	{
        DnpapMessage(DMC_WARNING, NWICMP_ROOT, "nwicmp: must be root to use icmp features");
		return TRUE;
	}
	
    if ((protoentptr = getprotobyname("icmp")) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWICMP_PROTO, "nwicmp: icmp protocol not supported");
        return FALSE;
    }
        
    if ((icmpsocket = socket(AF_INET, SOCK_RAW, protoentptr->p_proto)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_OPEN, "nwicmp: open failed");
        return FALSE;
    }
                                
    return TRUE;
}


/* exits the NwNit module
 * closes all opened sockets
 * returns TRUE or FALSE for success or failure
 */
BOOLEAN NwIcmpExit(VOID)
{             
BOOLEAN rc = TRUE;
NwIcmpDg *dg;

    for (dg = NwIcmpDgList; dg != NULL; dg = NwIcmpDgList)
        rc &= NwIcmpClose((NW_DG *)dg);
               
    if (icmpsocket != 0)
        close(icmpsocket);
        
    return rc;
}


NwStat *NwIcmpStatistics(char *addr)
{            
    return &nwStat;
}

        
/* opens a socket for receiving ICMP packets
 * all ICMP packets are received, no parameters are taken into account
 * returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwIcmpAccept(char *addr           /* accept specification: 'icmp' */, 
                    NW_DG_RECV_CALL recv /* receive callback; parameter addr is formatted as follows: 'icmp:(char*)host:(ushort)ident' */, 
                    VOID *param          /* extra parameter you want to be forwarded to recv() */)
{   
NwIcmpDg *dg;

    if (protoentptr == NULL)
    {
        DnpapMessage(DMC_ERROR, NWICMP_INIT, "nwicmp: not initialized");
        return NULL;
    }
    
    if (strcmp(addr, "icmp") != 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_PARAM, "nwicmp: invalid parameter list");
        return NULL;
    }   
    
	if ((dg = NewNwIcmpDg("")) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWICMP_NEWICMP, "nwicmp: memory allocation failed");
		return NULL;
    }

    if ((dg->IcmpInfo.Socket = socket(AF_INET, SOCK_RAW, protoentptr->p_proto)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_OPEN, "nwicmp: open failed");
    	NwIcmpCloseDg(dg);
        return NULL;
    }
                                
    dg->Block->sckt = dg->IcmpInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwIcmpBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_BLOCK, "nwicmp: block registration failed");
    	NwIcmpCloseDg(dg);
		return NULL;
    }
                        
    dg->NwInfo.iface = &IcmpDgIface;
    sprintf(dg->NwInfo.addr, "icmp");
    dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;
    
	strcpy(dg->IcmpInfo.Host, "");
    dg->IcmpInfo.Addr = 0;
	dg->IcmpInfo.Ident = 0;
             
    if (NwIcmpDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWICMP_ADDLIST, "nwicmp: could not add to list");
    	NwIcmpCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* opens a socket for receiving and sending ICMP packets from and to a specific host
 * opens a socket and records the host address and the specified identifying integer
 * for matching
 * returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwIcmpConnect(char *addr            /* connect specification: 'icmp:(char*)host:(ushort)ident' */, 
                     NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'icmp:(char*)host:(ushort)ident' */, 
                     VOID *param           /* extra parameter you want to be forwarded to recv() */)
{   
NwIcmpDg *dg;
CHAR *p;
USHORT ident;
CHAR host[NWICMP_SIZE_HOST];
HOSTENT *hostent;
ULONG address;

    if (protoentptr == NULL)
    {
        DnpapMessage(DMC_ERROR, NWICMP_INIT, "nwicmp: not initialized");
        return NULL;
    }
    
    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWICMP_PARAM, "nwicmp: invalid parameter list");
        return NULL;
    }   
    
    strncpy(host, addr, p-addr);
    host[p-addr] = '\0';
    
	if (sscanf(p, ":%hu", &ident) != 1 || ident == 0)
	{
        DnpapMessage(DMC_ERROR, NWICMP_PARAM, "nwicmp: invalid parameters");
	    return NULL;
	}
    
	if ((dg = NewNwIcmpDg(host)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWICMP_NEWICMP, "nwicmp: memory allocation failed");
		return NULL;
    }

    if ((dg->IcmpInfo.Socket = socket(AF_INET, SOCK_RAW, protoentptr->p_proto)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_OPEN, "nwicmp: open failed");
    	NwIcmpCloseDg(dg);
        return NULL;
    }
                                
    if ((address = inet_addr(host)) == INADDR_NONE)
    {
        if ((hostent=gethostbyname(host)) != NULL)
        {	    
        	address = ((IN_ADDR *)(VOID *)hostent->h_addr_list[0])->s_addr;
        }
        else
        {
            DnpapMessage(DMC_ERROR, NWICMP_ADDR, "nwicmp: hostname can't be resolved");
           	NwIcmpCloseDg(dg);
    	    return NULL;
        }
    }

    dg->Block->sckt = dg->IcmpInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwIcmpBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_BLOCK, "nwicmp: block registration failed");
    	NwIcmpCloseDg(dg);
		return NULL;
    }
                        
    dg->NwInfo.iface = &IcmpDgIface;
    sprintf(dg->NwInfo.addr, "icmp:%s:%hu", host, ident);
    dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;

    dg->IcmpInfo.Addr = address;
	strcpy(dg->IcmpInfo.Host, host);
    
	dg->IcmpInfo.Ident = ident;
             
    if (NwIcmpDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWICMP_ADDLIST, "nwicmp: could not add to list");
    	NwIcmpCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


BOOLEAN NwIcmpClose(NW_DG *c)
{                   
NwIcmpDg *dg = (NwIcmpDg *)c;

    NwIcmpDgRemoveList(dg);
    NwIcmpCloseDg(dg);

    return TRUE;
}


BOOLEAN NwIcmpCloseDg(NwIcmpDg *dg)
{
    if (dg->Block != NULL)    
        BlockDeRegister(dg->Block);
    
    if (dg->IcmpInfo.Socket != 0)
        close(dg->IcmpInfo.Socket);
        
    DelNwIcmpDg(dg);  
    
    return TRUE;
}    
    

/* sends a packet over an existing NW datagram             
 * the recorded address is taken as destination
 * returns TRUE or FALSE for success or failure
 */
BOOLEAN NwIcmpSend(NW_DG *dg   /* a valid datagram link */, 
                   BYTE *frame /* packet to be sent*/, 
                   INT len     /* packet length */)
{
NwIcmpDg *icmpdg = (NwIcmpDg *) dg;	
SOCKADDR_IN addr;
INT i;
    
    if (icmpdg->IcmpInfo.Addr == 0 || icmpdg->IcmpInfo.Ident == 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_SEND, "nwicmp: can not send on accept icmp socket");
        return FALSE;
    }
    
    addr.sin_family = AF_INET;
    memcpy(&addr.sin_addr, &icmpdg->IcmpInfo.Addr, sizeof(icmpdg->IcmpInfo.Addr));
    memset(&addr.sin_zero, 0, 8);

    i = sendto(icmpdg->IcmpInfo.Socket, frame, len, 0, (SOCKADDR *)&addr, sizeof(addr));
        
    if (i < 0 || i != len)
    {
        DnpapMessage(DMC_ERROR, NWICMP_SEND, "nwicmp: sendto failed");
        return FALSE;
    }

    return TRUE;
}


/* sends a packet to a specified address
 * this function can be used to sent one-shot only packets
 * returns TRUE or FALSE for success or failure
 */
BOOLEAN NwIcmpSendTo(NW_DG *dg   /* an existing datagram link will be used if specified */, 
                     CHAR *addr  /* otherwise another destination can/must be specified: 'icmp:(char*)host' */, 
                     BYTE *frame /* the ICMP packet that is to be sent  */,
                     INT len     /* packet length */)
{
NwIcmpDg *icmpdg = (NwIcmpDg *) dg;	
CHAR *p;
CHAR host[NWICMP_SIZE_HOST];
HOSTENT *hostent;
ULONG address;
SOCKADDR_IN sa;
INT i;

    if (protoentptr == NULL)
    {
        DnpapMessage(DMC_ERROR, NWICMP_INIT, "nwicmp: not initialized");
        return FALSE;
    }

	if (icmpdg != NULL && addr != NULL)
	{
        DnpapMessage(DMC_ERROR, NWICMP_SEND, "nwicmp: only one of datagram or address may be specified");
        return FALSE;
	}
	    
    if (icmpdg == NULL)
	{
	    if ((addr = strchr(addr, ':')) == NULL)
	    {
	        DnpapMessage(DMC_ERROR, NWICMP_PARAM, "nwicmp: invalid parameter list");
	        return FALSE;
	    }   
	    					
		addr++;				  
		p = addr+strlen(addr);
	    strncpy(host, addr, p-addr);
	    host[p-addr] = '\0';
	    
	    if ((hostent=gethostbyname(host)) != NULL)
	    {	    
	    	address = ((IN_ADDR *)(VOID *)hostent->h_addr_list[0])->s_addr;
	    }
	    else
	    {
	    	if ((address = inet_addr(host)) == INADDR_NONE)
	    	{
		        DnpapMessage(DMC_ERROR, NWICMP_ADDR, "hostname can't be resolved");
		    	return FALSE;
	    	}
	    }
	
	    sa.sin_family = AF_INET;
	    memcpy(&sa.sin_addr, &address, sizeof(address));
	    memset(&sa.sin_zero, 0, 8);

        i = sendto(icmpsocket, frame, len, 0, (SOCKADDR *)&sa, sizeof(sa));
	}
    else
	{
	    if (icmpdg->IcmpInfo.Addr == 0)
	    {
	        DnpapMessage(DMC_ERROR, NWICMP_SEND, "nwicmp: can not send on accept icmp socket");
	        return FALSE;
	    }
	    
	    sa.sin_family = AF_INET;
	    memcpy(&sa.sin_addr, &icmpdg->IcmpInfo.Addr, sizeof(icmpdg->IcmpInfo.Addr));
	    memset(&sa.sin_zero, 0, 8);

        i = sendto(icmpdg->IcmpInfo.Socket, frame, len, 0, (SOCKADDR *)&sa, sizeof(sa));
	}
        
    if (i < 0 || i != len)
    {
        DnpapMessage(DMC_ERROR, NWICMP_SEND, "nwicmp: sendto failed");
        return FALSE;
    }

    return TRUE;
}


BOOLEAN NwIcmpIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{
	return FALSE;
}


int NwIcmpBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwIcmpDg *dg = (NwIcmpDg *) param;	
	
	switch (reason)
	{
	case BLOCKEVENT_READ:
		if (NwIcmpRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWICMP_READ, "nwicmp: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWICMP_WEVENT, "nwicmp: write event");
		break;
	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWICMP_EEVENT, "nwicmp: except event");
		break;
	default:
		DnpapMessage(DMC_ERROR, NWICMP_NEVENT, "nwicmp: unknown block event");
		return 1;
	}
	
	return 0;
}


BOOLEAN NwIcmpRead(NwIcmpDg *dg)
{
SOCKADDR_IN address;
int addrlen = sizeof(address), icmplen;
IP *ip_ptr;
ICMP *icmp_ptr;
BYTE frame[NWICMP_SIZE_PACKET];
CHAR addr[NWICMP_SIZE_ENTRY];

    if ((icmplen = recvfrom(dg->IcmpInfo.Socket, frame, sizeof(frame), 0, (SOCKADDR *)&address, &addrlen)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWICMP_RECV, "nwicmp: recvfrom failed");
        return 1;
    }
                                             
    ip_ptr = (IP *)(VOID *) frame;
    icmp_ptr = (ICMP *)(VOID *)(frame + (ip_ptr->ip_hl << 2));
    icmplen -= (BYTE *)icmp_ptr - frame;

    if (dg->IcmpInfo.Addr == 0 || dg->IcmpInfo.Ident == 0 ||
        (dg->IcmpInfo.Addr == address.sin_addr.s_addr &&
         dg->IcmpInfo.Ident == icmp_ptr->icmp_id))
    {
        sprintf(addr, "icmp:%s:%hu", Inet_NtoA(address.sin_addr.s_addr), icmp_ptr->icmp_id);
            
    	if (dg->NwInfo.recv != NULL)
    		dg->NwInfo.recv((NW_DG *)dg, addr, (BYTE *) icmp_ptr, icmplen, dg->NwInfo.parm);
    }
	
	return TRUE;
}


BOOLEAN NwIcmpDgAddList(NwIcmpDg *dg)
{
    if (NwIcmpDgList == NULL)
        dg->Next = NULL;
    else
        dg->Next = NwIcmpDgList;
        
    dg->Prev = NULL;
    
    NwIcmpDgList = dg;
        
    return TRUE;
}


BOOLEAN NwIcmpDgRemoveList(NwIcmpDg *dg)
{
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwIcmpDgList = dg->Next;
        
    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;
        
    return TRUE;
}


NwIcmpDg* NewNwIcmpDg(CHAR *host)
{
NwIcmpDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwIcmpDg))) == NULL)
		return NULL;
	
	dg->Block = NULL;
    
    dg->NwInfo.iface = NULL;
    dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;
    
    dg->IcmpInfo.Host = NULL;
	dg->IcmpInfo.Socket = 0;		
	dg->IcmpInfo.Addr = 0;
	dg->IcmpInfo.Ident = 0;
    
    dg->Next = NULL;
    dg->Prev = NULL;
	
    if ((dg->Block = DnpapMalloc(sizeof(BLOCK))) == NULL)
		return NULL;
    
	if ((dg->IcmpInfo.Host = DnpapMalloc(strlen(host)+1)) == NULL)
    {
    	DnpapFree(dg->Block);
		return NULL;
	}
	
	return dg;
}


VOID DelNwIcmpDg(NwIcmpDg *dg)
{
	DnpapFree(dg->Block);
	
    DnpapFree(dg->IcmpInfo.Host);
    
    DnpapFree(dg);
}


/*
 * Checksum routine for Internet Protocol family headers (C Version).
 *
 * Refer to "Computing the Internet Checksum" by R. Braden, D. Borman and
 * C. Partridge, Computer Communication Review, Vol. 19, No. 2, April 1989,
 * pp. 86-101, for additional details on computing this checksum.
 */

/* return checksum in low-order 16 bits */
INT IpCheckSum(USHORT *ptr, INT nbytes)
{
	register long	 sum;           /* assumes long == 32 bits */
	u_short			 oddbyte;
	register u_short answer;        /* assumes u_short == 16 bits */

	/*
	 * Our algorithm is simple, using a 32-bit accumulator (sum),
	 * we add sequential 16-bit words to it, and at the end, fold back
	 * all the carry bits from the top 16 bits into the lower 16 bits.
	 */

	sum = 0;
	while (nbytes > 1)  {
		sum += *ptr++;
		nbytes -= 2;
	}

				/* mop up an odd byte, if necessary */
	if (nbytes == 1) {
		oddbyte = 0;        /* make sure top half is zero */
		*((u_char *) &oddbyte) = *(u_char *)ptr;   /* one byte only */
		sum += oddbyte;
	}

	/*
	 * Add back carry outs from top 16 bits to low 16 bits.
	 */

	sum  = (sum >> 16) + (sum & 0xffff);    /* add high-16 to low-16 */
	sum += (sum >> 16);                     /* add carry */
	answer = ~sum;          /* ones-complement, then truncate to 16 bits */
	return answer;
}
