/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#include <stropts.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/nit_buf.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
               
#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nwnit.h"

 
/* nwnit.c - NwNit module functions */

            
typedef struct _NwNit
{
	INT Socket;
	CHAR *Interface;
    BYTE  Address[NWNIT_SIZE_ADDRESS];
	ULONG BufSize;
	ULONG SnapLen;
	TIMEVAL Timeout;
	ULONG Flags;
	BYTE *Buf;
	LONG  Pos;
	LONG  End;
	         
} NwNit;
    

typedef struct _NwNitDg
{          
    NW_DG  NwInfo;
    NwNit  NitInfo;
    
    BLOCK          *Block;
    
    struct _NwNitDg *Prev;
    struct _NwNitDg *Next;
    
} NwNitDg;

      

static BOOLEAN NwNitInit(VOID);
static BOOLEAN NwNitExit(VOID);

static NW_DG  *NwNitAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwNitConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwNitSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwNitSendTo(NW_DG *dg, char *addr, BYTE *frame, INT len);
static BOOLEAN NwNitClose(NW_DG *dg);
static BOOLEAN NwNitIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwNitStatistics(char *addr);


NW_DG_IFACE NitDgIface = { "nit", 
                           NwNitInit,   NwNitExit, 
                           NwNitAccept, NwNitConnect, 
                           NwNitSend,   NwNitSendTo,
                           NwNitClose,  NwNitIfConfig,
                           NwNitStatistics
                        };


static NwNitDg *NwNitDgList = NULL;

static NwStat nwStat;


static INT nitsocket;

static BOOLEAN NwNitCloseDg(NwNitDg *dg);

static BOOLEAN NwNitDgAddList(NwNitDg *dg);
static BOOLEAN NwNitDgRemoveList(NwNitDg *dg);

static NwNitDg* NewNwNitDg(CHAR *interface, ULONG bufsize);
static VOID DelNwNitDg(NwNitDg *dg);

static BOOLEAN NwNitRead(NwNitDg *dg);

static BOOLEAN NwNitSetup(NwNitDg *dg, CHAR *interface, ULONG flags, ULONG bufsize, ULONG snaplen, TIMEVAL *timeout);

static int NwNitBlockCallback(BLOCK *blk, int reason, VOID *param);



/* Initializes the NwNit module
 * Initializes the NwNit messages and opens a socket for general use
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwNitInit(VOID)
{       
	if (geteuid() != 0)
	{
        DnpapMessage(DMC_WARNING, NWNIT_ROOT, "nwnit: must be root to use nit features");
		return TRUE;
	}
	
    if ((nitsocket = open(NWNITDEVICE, O_WRONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_OPEN, "nwnit: open failed");
        return FALSE;
    }

    return TRUE;
}


/* Exits the NwNit module
 * Closes all opened sockets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwNitExit(VOID)
{             
BOOLEAN rc = TRUE;
NwNitDg *dg;

    for (dg = NwNitDgList; dg != NULL; dg = NwNitDgList)
        rc &= NwNitClose((NW_DG *)dg);

    if (nitsocket != 0)
        close(nitsocket);        
        
    return rc;
}

        
NwStat *NwNitStatistics(char *addr)
{            
    return &nwStat;
}

        
/* Opens a socket for receiving promiscuous packets
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwNitAccept(char *addr            /* accept specification: 'nit:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                   NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'nit:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                   VOID *param           /* extra parameter you want to be forwarded to recv() */)
{   
NwNitDg *dg;
CHAR interface[NWNIT_SIZE_INTERFACE];
ULONG bufsize, snaplen;
ULONG flags;
TIMEVAL timeout;
CHAR *p;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWNIT_PARAM, "nwnit: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (sscanf(p, ":%lu:%lu", &bufsize, &snaplen) != 2)
	{
        DnpapMessage(DMC_ERROR, NWNIT_PARAM, "nwnit: invalid parameters");
	    return NULL;
	}
	                           
	flags = NI_PROMISC | NWNIT_DEFAULT_FLAGS;
	
    /* set buffer timeout at 1/10 of a second */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwNitDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWNIT_NEWNIT, "nwnit: memory allocation failed");
		return NULL;
    }

    if ((dg->NitInfo.Socket = open(NWNITDEVICE, O_RDONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_OPEN, "nwnit: open failed");
    	NwNitCloseDg(dg);
        return NULL;
    }

    if (NwNitSetup(dg, interface, flags, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWNIT_SETUP, "nwnit: setup failed");
    	NwNitCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->NitInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwNitBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_BLOCK, "nwnit: block registration failed");
    	NwNitCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &NitDgIface;                                
    sprintf(dg->NwInfo.addr, "nit:%s:%lu:%lu", interface, bufsize, snaplen);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
    
	strcpy(dg->NitInfo.Interface, interface);
	dg->NitInfo.BufSize = bufsize;
	dg->NitInfo.SnapLen = snaplen;
	dg->NitInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->NitInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->NitInfo.Flags = flags;
	dg->NitInfo.Pos = 0;
	dg->NitInfo.End = 0;
             
    if (NwNitDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWNIT_ADDLIST, "nwnit: could not add to list");
    	NwNitCloseDg(dg);
		return NULL;
	}
    
    if (ioctl(dg->NitInfo.Socket, I_FLUSH, (VOID *)FLUSHR) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: read flush failed");
    	NwNitCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* Opens a socket for receiving and sending packets from and to a specific host
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all. The Ethernet address is recorded 
 * and the receive callback is only called for packets to or from this address
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwNitConnect(char *addr            /* connect specification: 'nit:(char*)interface:(ulong)bufsize:(ulong)snaplen:(int[6])ethernet' */, 
                     NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'nit:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                     VOID *param           /* extra parameter you want to forward to recv() */)
{   
NwNitDg *dg;
CHAR interface[NWNIT_SIZE_INTERFACE];
ULONG bufsize, snaplen;
ULONG flags;
TIMEVAL timeout;
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWNIT_PARAM, "nwnit: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (sscanf(p, ":%lu:%lu:%u.%u.%u.%u.%u.%u", &bufsize, &snaplen, &a1, &a2, &a3, &a4, &a5, &a6) != 8)
	{
        DnpapMessage(DMC_ERROR, NWNIT_PARAM, "nwnit: invalid parameters");
	    return NULL;
	}
	
	flags = NWNIT_DEFAULT_FLAGS;
	
    /* set buffer timeout at 1/10 of a second */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwNitDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWNIT_NEWNIT, "nwnit: memory allocation failed");
		return NULL;
    }

    if ((dg->NitInfo.Socket = open(NWNITDEVICE, O_RDWR)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_OPEN, "nwnit: open failed");
    	NwNitCloseDg(dg);
        return NULL;
    }

    if (NwNitSetup(dg, interface, flags, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWNIT_SETUP, "nwnit: setup failed");
    	NwNitCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->NitInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwNitBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_BLOCK, "nwnit: block registration failed");
    	NwNitCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &NitDgIface;                                
    sprintf(dg->NwInfo.addr, "nit:%s:%lu:%lu:%u.%u.%u.%u.%u.%u", 
                             interface, bufsize, snaplen, a1, a2, a3, a4, a5, a6);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
	
	strcpy(dg->NitInfo.Interface, interface);
    dg->NitInfo.Address[0] = a1;
    dg->NitInfo.Address[1] = a2;
    dg->NitInfo.Address[2] = a3;
    dg->NitInfo.Address[3] = a4;
    dg->NitInfo.Address[4] = a5;
    dg->NitInfo.Address[5] = a6;
	dg->NitInfo.BufSize = bufsize;
	dg->NitInfo.SnapLen = snaplen;
	dg->NitInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->NitInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->NitInfo.Flags = flags;
	dg->NitInfo.Pos = 0;
	dg->NitInfo.End = 0;
             
    if (NwNitDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWNIT_ADDLIST, "nwnit: could not add to list");
    	NwNitCloseDg(dg);
		return NULL;
	}
    
    if (ioctl(dg->NitInfo.Socket, I_FLUSH, (VOID *)FLUSHR) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: read flush failed");
    	NwNitCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* Closes a NwNit datagram socket
 * Closes the specified NwNit datagram socket
 * Returns TRUE for success and FALSE for failure
 */
BOOLEAN NwNitClose(NW_DG *dg   /* an existing NwNit datagram socket */)
{                   
NwNitDg *nitdg = (NwNitDg *)dg;

    NwNitDgRemoveList(nitdg);
    NwNitCloseDg(nitdg);

    return TRUE;
}


BOOLEAN NwNitCloseDg(NwNitDg *dg)
{
    if (dg->Block != NULL)    
        BlockDeRegister(dg->Block);
    
    if (dg->NitInfo.Socket != 0)
        close(dg->NitInfo.Socket);
        
    DelNwNitDg(dg);  
    
    return TRUE;
}    
    

/* Sends a packet over an existing NW datagram
 * The destination address is checked against the recorded Ethernet address
 * to allow only packets that are sent to that address
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwNitSend(NW_DG *dg    /* valid NW datagram */, 
                  BYTE *frame  /* packet */, 
                  INT len      /* packet length */)
{
NwNitDg *nitdg = (NwNitDg *) dg;	
SOCKADDR sa; 
STRBUF ctl, data;
INT llhsize = sizeof(sa.sa_data);
BYTE noaddr[NWNIT_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (len > NWNIT_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWNIT_SEND, "nwnit: packet too long");
        return FALSE;
    }

    if (memcmp(nitdg->NitInfo.Address, noaddr, NWNIT_SIZE_ADDRESS) == 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_SEND, "nwnit: can not send over accept socket");
        return FALSE;
    }
            
    sa.sa_family = AF_UNSPEC;
    memcpy(frame, nitdg->NitInfo.Address, NWNIT_SIZE_ADDRESS);
    memcpy(sa.sa_data, frame, llhsize);

    ctl.maxlen = ctl.len = sizeof(sa);
    ctl.buf = (BYTE *)&sa;
    data.maxlen = data.len = len - llhsize;
    data.buf = frame + llhsize;
    if (putmsg(nitdg->NitInfo.Socket, &ctl, &data, 0) < 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_SEND, "nwnit: putmsg failed");
        return FALSE;
    }

    if (ioctl(nitdg->NitInfo.Socket, I_FLUSH, (VOID *)FLUSHW) < 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: write flush failed");
        return FALSE;
    }
        
    return TRUE;
}


/* Sends a packet to a specified address
 * This function can be used to sent one-shot only packets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwNitSendTo(NW_DG *dg    /* not supported, set to NULL */, 
                    char *addr   /* address specification: 'nit:(char*)interface[:(int[6])ethernet]' */, 
                    BYTE *frame  /* packet */, 
                    INT len      /* packet length */)
{
STRIOCTL si;
IFREQ ifr;
SOCKADDR sa; 
STRBUF ctl, data;
INT llhsize = sizeof(sa.sa_data);
CHAR interface[NWNIT_SIZE_INTERFACE];
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;
BYTE address[NWNIT_SIZE_ADDRESS];
         
    if (dg != NULL)
        DnpapMessage(DMC_WARNING, NWNIT_SEND, "nwnit: sendto over existing connection has no effect");
    
    if (nitsocket == 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_INIT, "nwnit: not initialized");
        return FALSE;
    }
    
    if (len > NWNIT_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWNIT_SEND, "nwnit: packet too long");
        return FALSE;
    }

    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWNIT_PARAM, "nwnit: invalid parameter list");
        return FALSE;
    }   
    									  
	if ((p = strchr(++addr, ':')) == NULL)
	{
		p = addr+strlen(addr);
	}
	else
	{
		if (sscanf(p, ":%u.%u.%u.%u.%u.%u", &a1, &a2, &a3, &a4, &a5, &a6) != 6)
		{
	        DnpapMessage(DMC_ERROR, NWNIT_PARAM, "nwnit: invalid parameters");
		    return FALSE;
		}
		
	    address[0] = a1;	
	    address[1] = a2;	
	    address[2] = a3;	
	    address[3] = a4;	
	    address[4] = a5;	
	    address[5] = a6;	
		
	    memcpy(frame, address, NWNIT_SIZE_ADDRESS);
    }

    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    si.ic_timout = INFTIM;
    si.ic_cmd = NIOCBIND;
    si.ic_len = sizeof(ifr);
    si.ic_dp = (VOID *)&ifr;
    if (ioctl(nitsocket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: interface bind failed");
		return FALSE;
	}
    
    sa.sa_family = AF_UNSPEC;
    memcpy(sa.sa_data, frame, llhsize);

    ctl.maxlen = ctl.len = sizeof(sa);
    ctl.buf = (BYTE *)&sa;
    data.maxlen = data.len = len - llhsize;
    data.buf = frame + llhsize;
    if (putmsg(nitsocket, &ctl, &data, 0) < 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_SEND, "nwnit: putmsg failed");
        return FALSE;
    }

    if (ioctl(nitsocket, I_FLUSH, (VOID *)FLUSHW) < 0)
    {
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: write flush failed");
        return FALSE;
    }
        
    return TRUE;
}


BOOLEAN NwNitIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{			 
NwNitDg *nitdg = (NwNitDg *) dg;	
IFREQ ifr;

    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, nitdg->NitInfo.Interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';

    if (ioctl(nitdg->NitInfo.Socket, SIOCGIFADDR, (VOID *)&ifr) < 0)
		return FALSE;
	
	if (strncmp(nitdg->NitInfo.Interface, "le", strlen("le")) == 0)
	{	
		ifconfig->type = MAC_TYPE_ETHERNET_CSMACD;
		ifconfig->mtu = MAC_MTU_ETHERNET_CSMACD;
		ifconfig->speed = MAC_SPEED_ETHERNET_CSMACD;
		ifconfig->addrlen = MAC_ADDRLEN_ETHERNET_CSMACD;
		memcpy(ifconfig->address, ifr.ifr_addr.sa_data, ifconfig->addrlen);
		memcpy(ifconfig->broadcast, "\xff\xff\xff\xff\xff\xff", ifconfig->addrlen);
	}
	else
	{
		DnpapMessage(DMC_ERROR, NWNIT_CONFIG, "nwnit: unknown interface type");
		return FALSE;
	}
		
	return TRUE;
}


int NwNitBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwNitDg *dg = (NwNitDg *) param;	
	
	switch (reason)
	{
	case BLOCKEVENT_READ:
		if (NwNitRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWNIT_READ, "nwnit: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWNIT_WEVENT, "nwnit: write event");
		break;
	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWNIT_EEVENT, "nwnit: except event");
		break;
	default:
		DnpapMessage(DMC_ERROR, NWNIT_NEVENT, "nwnit: unknown block event");
		return 1;
	}
	
	return 0;
}


BOOLEAN NwNitRead(NwNitDg *dg)
{
BYTE *cp;
NITBUFHDR *hp;
ULONG msglen;
BYTE noaddr[NWNIT_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (!(dg->NitInfo.Pos < dg->NitInfo.End))
    {
    	if ((dg->NitInfo.End = read(dg->NitInfo.Socket, dg->NitInfo.Buf, dg->NitInfo.BufSize)) < 0)
    		return FALSE;
            
    	dg->NitInfo.Pos = 0;
    }
		
	while (dg->NitInfo.Pos < dg->NitInfo.End)
	{
		cp = dg->NitInfo.Buf+dg->NitInfo.Pos;
		
		hp = (NITBUFHDR *)(VOID *)cp;      
		cp += sizeof(*hp);
	    msglen = ((IFLEN *)(VOID *)(cp+sizeof(IFTIME)+sizeof(IFDROPS)))->nh_pktlen;
		if (dg->NitInfo.SnapLen > 0 && msglen > dg->NitInfo.SnapLen)
			msglen = dg->NitInfo.SnapLen;
		
		dg->NitInfo.Pos += hp->nhb_totlen;

        if (memcmp(dg->NitInfo.Address, noaddr, NWNIT_SIZE_ADDRESS) != 0 &&
            memcmp(dg->NitInfo.Address, cp, NWNIT_SIZE_ADDRESS) != 0 &&
            memcmp(dg->NitInfo.Address, cp+NWNIT_SIZE_ADDRESS, NWNIT_SIZE_ADDRESS) != 0)
            continue;
        
		if (dg->NwInfo.recv != NULL)
			dg->NwInfo.recv((NW_DG *)dg, dg->NwInfo.addr, cp, msglen, dg->NwInfo.parm);
	}
	
	return TRUE;
}


BOOLEAN NwNitDgAddList(NwNitDg *dg)
{
    if (NwNitDgList != NULL)
		NwNitDgList->Prev = dg;
	
    dg->Next = NwNitDgList;
    dg->Prev = NULL;
    
    NwNitDgList = dg;
        
    return TRUE;
}


BOOLEAN NwNitDgRemoveList(NwNitDg *dg)
{   
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwNitDgList = dg->Next;
        
    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;
        
    return TRUE;
}


NwNitDg* NewNwNitDg(CHAR *interface, ULONG bufsize)
{
NwNitDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwNitDg))) == NULL)
		return NULL;
	
	dg->Block = NULL;
    
    dg->NwInfo.iface = NULL;
	dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;    
    
    dg->NitInfo.Interface = NULL;
    memset(dg->NitInfo.Address, 0, NWNIT_SIZE_ADDRESS);
    dg->NitInfo.Buf = NULL;
	dg->NitInfo.Socket = 0;		
	dg->NitInfo.BufSize = 0;
	dg->NitInfo.SnapLen = 0;
	dg->NitInfo.Timeout.tv_sec = 0;
	dg->NitInfo.Timeout.tv_usec = 0;
	dg->NitInfo.Flags = 0;
	dg->NitInfo.Pos = 0;
	dg->NitInfo.End = 0;
    
    dg->Next = NULL;
    dg->Prev = NULL;
	
    if ((dg->Block = DnpapMalloc(sizeof(BLOCK))) == NULL)
		return NULL;
    
	if ((dg->NitInfo.Interface = DnpapMalloc(strlen(interface)+1)) == NULL)
    {
    	DnpapFree(dg->Block);
		return NULL;
	}
	
	if ((dg->NitInfo.Buf = DnpapMalloc(bufsize)) == NULL)
    {
    	DnpapFree(dg->Block);
    	DnpapFree(dg->NitInfo.Interface);
		return NULL;
	}
	
	return dg;
}


VOID DelNwNitDg(NwNitDg *dg)
{
	DnpapFree(dg->Block);
	
    DnpapFree(dg->NitInfo.Interface);
    DnpapFree(dg->NitInfo.Buf);
    
    DnpapFree(dg);
}


BOOLEAN NwNitSetup(NwNitDg *dg, CHAR *interface, ULONG flags, ULONG bufsize, ULONG snaplen, TIMEVAL *timeout)
{
STRIOCTL si;
IFREQ ifr;
 
    flags |= NWNIT_DEFAULT_FLAGS;
    
    /* arrange to get discrete messages from the stream */
    if (ioctl(dg->NitInfo.Socket, I_SRDOPT, (VOID *)RMSGN) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: discrete message option set failed");
		return FALSE;
	}
    
    /* setting of options are allowed to take a long time */      
    si.ic_timout = INFTIM;
                   
    /* push and configure the filtering module */
    if (ioctl(dg->NitInfo.Socket, I_PUSH, NWNITFILTER) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_PUSH, "nwnit: 'pf' push failed");
		return FALSE;
	}
    
    /* push and configure the buffering module */
    if (ioctl(dg->NitInfo.Socket, I_PUSH, NWNITBUFFER) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_PUSH, "nwnit: 'nbuf' push failed");
		return FALSE;
	}
    
    si.ic_cmd = NIOCSCHUNK;
    si.ic_len = sizeof(bufsize);
    si.ic_dp = (VOID *)&bufsize;
    if (ioctl(dg->NitInfo.Socket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: chunksize option set failed");
		return FALSE;
	}

	/* timeout settings must come after bufsize settings:
	 * if timeout == 0 then bufsize is reset too
	 */
    si.ic_cmd = NIOCSTIME;
    si.ic_len = sizeof(*timeout);
    si.ic_dp = (VOID *)timeout;
    if (ioctl(dg->NitInfo.Socket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: timeout option set failed");
		return FALSE;
	}
    
    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    si.ic_cmd = NIOCBIND;
    si.ic_len = sizeof(ifr);
    si.ic_dp = (VOID *)&ifr;
    if (ioctl(dg->NitInfo.Socket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: interface bind failed");
		return FALSE;
	}
    
    if (snaplen > 0)
    {
        si.ic_cmd = NIOCSSNAP;
        si.ic_len = sizeof(snaplen);
        si.ic_dp = (VOID *)&snaplen;
        if (ioctl(dg->NitInfo.Socket, I_STR, (VOID *)&si) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: snapshot length option set failed");
    		return FALSE;
    	}
    }
                                   
    if (flags != 0)
    {
        si.ic_cmd = NIOCSFLAGS;
        si.ic_len = sizeof(flags);
        si.ic_dp = (VOID *)&flags;
        if (ioctl(dg->NitInfo.Socket, I_STR, (VOID *)&si) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWNIT_IOCTL, "nwnit: flags set failed");
    		return FALSE;
    	}
    }
    
    return TRUE;
}
