/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#if defined(SUNOS) || defined(SOLARIS)
#include <stropts.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <dnpap.h>
#include <config.h>
#include <message.h>
#include <snmp.h>

#include "nw.h"
#include "nwsnmp.h"
 

/* nwsnmp.c - NwSnmp module functions */

           
typedef struct _NwSnmp
{
    CHAR   Host[NWSNMP_SIZE_HOST];
	ULONG  Address;
	INT    Request;
    BYTE   Community[NWSNMP_SIZE_COMMUNITY];
	INT	   CommLen;
	NW_DG *Udp;
    
} NwSnmp;
    

typedef struct _NwSnmpDg
{          
    NW_DG  NwInfo;
    NwSnmp SnmpInfo;
    
    struct _NwSnmpDg *Prev;
    struct _NwSnmpDg *Next;
    
} NwSnmpDg;


static BOOLEAN NwSnmpInit(VOID);
static BOOLEAN NwSnmpExit(VOID);

static NW_DG  *NwSnmpAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwSnmpConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwSnmpSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwSnmpSendTo(NW_DG *dg, CHAR *addr, BYTE *frame, INT len);
static BOOLEAN NwSnmpClose(NW_DG *dg);
static BOOLEAN NwSnmpIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwSnmpStatistics(char *addr);


NW_DG_IFACE SnmpDgIface = { "snmp", 
                            NwSnmpInit,   NwSnmpExit, 
                            NwSnmpAccept, NwSnmpConnect, 
                            NwSnmpSend,   NwSnmpSendTo,
                            NwSnmpClose,  NwSnmpIfConfig,
                            NwSnmpStatistics
                        };


static init = FALSE;

static NwSnmpDg *NwSnmpDgList = NULL;

static NwStat nwStat;


static VOID NwSnmpCallback(NW_DG *udp, CHAR *addr, BYTE *frame, INT length, VOID *param);
static BOOLEAN NwSnmpCloseDg(NwSnmpDg *dg);
static BOOLEAN NwSnmpDgAddList(NwSnmpDg *dg);
static BOOLEAN NwSnmpDgRemoveList(NwSnmpDg *dg);
static NwSnmpDg* NewNwSnmpDg(VOID);
static VOID DelNwSnmpDg(NwSnmpDg *dg);

static BOOLEAN Frame2Udp(INT request, BYTE *community, INT commlen, 
				  NwSnmpFrame *snmpframe, 
				  BYTE **udpframe, INT *udpframelen);
static BOOLEAN Udp2Frame(INT *request, BYTE *community, INT *commlen, 
				  BYTE *udpframe, INT udpframelen, 
				  NwSnmpFrame *snmpframe);


/* initializes the NwSnmp module
 * initializes the NwSnmp message system
 * returns TRUE or FALSE for success or failure respectively
 */
BOOLEAN NwSnmpInit(VOID)
{
	if (init == FALSE)
		init = TRUE;
    return TRUE;
}


/* exits the NwSnmp module
 * cleans up NwSnmp stuff
 * returns TRUE or FALSE for success or failure respectively
 */
BOOLEAN NwSnmpExit(VOID)
{
BOOLEAN rc = TRUE;
NwSnmpDg *dg;

    for (dg = NwSnmpDgList; dg != NULL; dg = NwSnmpDgList)
        rc &= NwSnmpClose((NW_DG *)dg);
               
    return rc;
}


NwStat *NwSnmpStatistics(char *addr)
{            
    return &nwStat;
}

        
/* opens a SNMP connection on which all requests are received
 * all SNMP requests are received, only UDP port parameter is needed
 * returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwSnmpAccept(char *addr            /* accept specification: 'snmp:(int)port' */, 
					NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'snmp:(char*)host:(int)port:(char*)request:(byte*)community' */,
					VOID *param           /* extra parameter you want to be forwarded to recv() */)
{
NwSnmpDg *dg;
CHAR udpaddr[NWSNMP_SIZE_UDPADDR];
INT port;
NW_DG *udp;

    if (init == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWSNMP_INIT, "nwsnmp: not initialized");
        return NULL;
    }
    
    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWSNMP_PARAM, "nwsnmp: invalid parameter list");
        return NULL;
    }   

	if ((port = atoi(++addr)) <= 0)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_PARAM, "nwsnmp: invalid port value");
        return NULL;
	}

	if ((dg = NewNwSnmpDg()) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWSNMP_NEWSNMP, "nwsnmp: memory allocation failed");
		return NULL;
    }

	sprintf(udpaddr, "udp:*:%d", port);
	if ((udp = NwDgAccept(udpaddr, NwSnmpCallback, dg)) == NULL)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_UDP, "nwsnmp: can't open udp connection");
		return NULL;
	}

    dg->NwInfo.iface = &SnmpDgIface;
    sprintf(dg->NwInfo.addr, "snmp:%d", port);
    dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;

	dg->SnmpInfo.Address = 0;
	dg->SnmpInfo.Request = 0;
	dg->SnmpInfo.Udp = udp;

    if (NwSnmpDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_ADDLIST, "nwsnmp: could not add to list");
    	NwSnmpCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* opens a connection for receiving and sending SNMP requests from and to a specified host
 * opens a SNMP connection and records the host, request and community name
 * returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwSnmpConnect(char *addr            /* connection specification: '(char*)host:(int)port:(char*)request:(byte*)community' */, 
					  NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'snmp:(char*)host:(int)port:(char*)request:(byte*)community' */, 
					  VOID *param           /* extra parameter you want to be forwarded to recv() */)
{
NwSnmpDg *dg;
CHAR host[NWSNMP_SIZE_HOST];
CHAR request[NWSNMP_SIZE_REQUEST];
WORD type;
BYTE community[NWSNMP_SIZE_COMMUNITY];
INT commlen;
CHAR udpaddr[NWSNMP_SIZE_UDPADDR];
INT port;
NW_DG *udp;
HOSTENT *hostent;
ULONG address;

    if (init == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWSNMP_INIT, "nwsnmp: not initialized");
        return NULL;
    }
    
	if (Addr2Snmp(addr, host, &port, request, &type, community, &commlen) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_PARAM, "nwsnmp: invalid parameter list");
        return FALSE;
	}

	if ((dg = NewNwSnmpDg()) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWSNMP_NEWSNMP, "nwsnmp: memory allocation failed");
		return NULL;
    }

    if ((address = inet_addr(host)) == INADDR_NONE)
    {
        if ((hostent=gethostbyname(host)) != NULL)
        {	    
        	address = ((IN_ADDR *)(VOID *)hostent->h_addr_list[0])->s_addr;
        }
        else
        {
            DnpapMessage(DMC_ERROR, NWSNMP_ADDR, "nwsnmp: hostname can't be resolved");
           	NwSnmpCloseDg(dg);
    	    return NULL;
        }
    }

	sprintf(udpaddr, "udp:%s:%d", host, port);
	if ((udp = NwDgConnect(udpaddr, NwSnmpCallback, dg)) == NULL)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_UDP, "nwsnmp: can't open udp connection");
		return NULL;
	}

    dg->NwInfo.iface = &SnmpDgIface;
    sprintf(dg->NwInfo.addr, "snmp:%s:%d:%s:%.*s", host, port, request, commlen, community);
    dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;

	strcpy(dg->SnmpInfo.Host, host);
	dg->SnmpInfo.Address = address;
	dg->SnmpInfo.Request = type;
	memcpy(dg->SnmpInfo.Community, community, commlen);
	dg->SnmpInfo.CommLen = commlen;
	dg->SnmpInfo.Udp = udp;

    if (NwSnmpDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_ADDLIST, "nwsnmp: could not add to list");
    	NwSnmpCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


BOOLEAN NwSnmpClose(NW_DG *c)
{
NwSnmpDg *dg = (NwSnmpDg *)c;

	if (NwDgClose(dg->SnmpInfo.Udp) == TRUE)
	{
		NwSnmpDgRemoveList(dg);
		NwSnmpCloseDg(dg);
		return TRUE;
	}

	return FALSE;
}


BOOLEAN NwSnmpCloseDg(NwSnmpDg *dg)
{
    DelNwSnmpDg(dg);  
    
    return TRUE;
}


/* sends a SNMP request over an existing NW datagram connection
 * the recorded host, request and community are used for the SNMP packet
 * returns TRUE or FALSE for success or failure
 */
BOOLEAN NwSnmpSend(NW_DG *c    /* a valid datagram link */, 
				   BYTE *frame /* SNMP packet to be sent */, 
				   INT len     /* packet length */)
{
NwSnmpDg *dg = (NwSnmpDg *)c;
BYTE udpframe[NWSNMP_SIZE_UDP], *udpframeptr;
INT udpframelen;

	if (dg->SnmpInfo.Address == 0)
	{
		DnpapMessage(DMC_ERROR, NWSNMP_ADDR, "nwsnmp: can't send over an accept connection");
		return FALSE;
	}

	udpframeptr = udpframe;	
	udpframelen = NWSNMP_SIZE_UDP;
	if (Frame2Udp(dg->SnmpInfo.Request, dg->SnmpInfo.Community, dg->SnmpInfo.CommLen, 
				  (NwSnmpFrame *)(VOID *)frame, 
				  &udpframeptr, &udpframelen) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_UDP, "nwsnmp: can't encode the SNMP frame");
		return FALSE;
	}

	if (NwDgSend(dg->SnmpInfo.Udp, udpframeptr, udpframelen) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_UDP, "nwsnmp: can't send");
		return FALSE;
	}

	return TRUE;
}


/* sends a SNMP request to specified address or over existing NW datagram connection
 * the recorded host, request and community are used for the SNMP packet if
 * a 'dg' is specified, otherwise the contents of 'addr' is used. If
 * both are specified, the contents of 'addr' is used but the packet is send
 * over the existing datagram connection
 * returns TRUE or FALSE for success or failure
 */
BOOLEAN NwSnmpSendTo(NW_DG *c    /* an existing datagram link will be used if specified (!= NULL) */, 
					 CHAR *addr  /* otherwise another destination can/must be specified: 'snmp:(char*)host:(int)port:(char*)request:(byte*)community' */, 
					 BYTE *frame /* the SNMP packet that is to be sent */, 
					 INT len     /* packet length */)
{
NwSnmpDg *dg = (NwSnmpDg *)c;
CHAR host[NWSNMP_SIZE_HOST];
INT port;
CHAR request[NWSNMP_SIZE_REQUEST];
WORD type;
BYTE community[NWSNMP_SIZE_COMMUNITY];
INT commlen;
CHAR udpaddr[NWSNMP_SIZE_UDPADDR];
BYTE udpframe[NWSNMP_SIZE_UDP], *udpframeptr;
INT udpframelen;

	if (addr == NULL)
		return NwSnmpSend(c, frame, len);
	else
	{
		if (Addr2Snmp(addr, host, &port, request, &type, community, &commlen) == FALSE)
		{
	        DnpapMessage(DMC_ERROR, NWSNMP_PARAM, "nwsnmp: invalid parameter list");
	        return FALSE;
		}

		udpframeptr = udpframe;	
		udpframelen = NWSNMP_SIZE_UDP;
		if (Frame2Udp(type, community, commlen, 
					  (NwSnmpFrame *)(VOID *)frame, 
					  &udpframeptr, &udpframelen) == FALSE)
		{
	        DnpapMessage(DMC_ERROR, NWSNMP_UDP, "nwsnmp: can't encode the SNMP frame");
			return FALSE;
		}

		sprintf(udpaddr, "udp:%s:%d", host, port);
		if (NwDgSendTo(dg->SnmpInfo.Udp, udpaddr, udpframeptr, udpframelen) == FALSE)
		{
	        DnpapMessage(DMC_ERROR, NWSNMP_UDP, "nwsnmp: can't send");
			return FALSE;
		}
	}

	return TRUE;
}


BOOLEAN NwSnmpIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{			  
	return FALSE;
}


VOID NwSnmpCallback(NW_DG *udp, CHAR *addr, BYTE *frame, INT length, VOID *param)
{
NwSnmpDg *dg = (NwSnmpDg *) param;	
CHAR *p;
CHAR host[NWSNMP_SIZE_HOST];
CHAR udpport[NWSNMP_SIZE_UDPPORT];
INT port;
INT type;
CHAR *request;
BYTE community[NWSNMP_SIZE_COMMUNITY];
INT commlen;
CHAR address[NWSNMP_SIZE_ADDRESS];
NwSnmpFrame snmpframe;
SNMP_OBJECT objects[NWSNMP_SIZE_OBJECTS];
	
    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWSNMP_PARAM, "nwsnmp: invalid parameter list");
        return;
    }   
    
    strncpy(host, addr, p-addr);
    host[p-addr] = '\0';

	strcpy(udpport, ++p);
	if ((port = atoi(udpport)) <= 0)
	{
        DnpapMessage(DMC_ERROR, NWSNMP_PARAM, "nwsnmp: invalid port value");
        return;
	}

	commlen = NWSNMP_SIZE_COMMUNITY;
	snmpframe.NrObjects = NWSNMP_SIZE_OBJECTS;
	snmpframe.Objects = objects;
	if (Udp2Frame(&type, community, &commlen,
				  frame, length,
				  &snmpframe) == FALSE)	
	{
        DnpapMessage(DMC_ERROR, NWSNMP_UDP, "nwsnmp: can't decode the SNMP frame");
		return;
	}

	switch (type)
	{
	case SNMP_PDU_GET:
		request = "get";
		break;
	case SNMP_PDU_SET:
		request = "set";
		break;
	case SNMP_PDU_NEXT:
		request = "next";
		break;
	case SNMP_PDU_TRAP:
		request = "trap";
		break;
	case SNMP_PDU_RESPONSE:
		request = "response";
		break;
	default:
		return;
	}

    sprintf(address, "snmp:%s:%d:%s:%.*s", host, port, request, commlen, community);

	if (dg->NwInfo.recv != NULL)
		dg->NwInfo.recv((NW_DG *)dg, address, 
						(BYTE *)&snmpframe, 
						sizeof(NwSnmpFrame)+snmpframe.NrObjects*sizeof(SNMP_OBJECT), 
						dg->NwInfo.parm); 

	return;
}


BOOLEAN NwSnmpHost(NW_DG *dg, CHAR *host)
{
NwSnmpDg *snmp = (NwSnmpDg *) dg;

    if (snmp == NULL)
        return FALSE;
    strcpy(host, snmp->SnmpInfo.Host);
    return TRUE;
}


BOOLEAN NwSnmpCommunity(NW_DG *dg, BYTE *community, INT *commlen)
{
NwSnmpDg *snmp = (NwSnmpDg *) dg;	

    if (snmp == NULL)
        return FALSE;
    memcpy(community, snmp->SnmpInfo.Community, *commlen = snmp->SnmpInfo.CommLen);
    return TRUE;
}


BOOLEAN NwSnmpDgAddList(NwSnmpDg *dg)
{
    if (NwSnmpDgList == NULL)
        dg->Next = NULL;
    else
        dg->Next = NwSnmpDgList;
        
    dg->Prev = NULL;
    
    NwSnmpDgList = dg;
        
    return TRUE;
}


BOOLEAN NwSnmpDgRemoveList(NwSnmpDg *dg)
{   
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwSnmpDgList = dg->Next;
        
    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;
        
    return TRUE;
}


NwSnmpDg* NewNwSnmpDg(VOID)
{
NwSnmpDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwSnmpDg))) == NULL)
		return NULL;

	strcpy(dg->SnmpInfo.Host, "");
	dg->SnmpInfo.Address = 0;
	dg->SnmpInfo.Request = 0;
	strcpy(dg->SnmpInfo.Community, "");
	dg->SnmpInfo.CommLen = 0;
	dg->SnmpInfo.Udp = NULL;
	
    dg->Next = NULL;
    dg->Prev = NULL;
	
	return dg;
}


VOID DelNwSnmpDg(NwSnmpDg *dg)
{
    DnpapFree(dg);
}


BOOLEAN Frame2Udp(INT request, BYTE *community, INT commlen, 
				  NwSnmpFrame *snmpframe, 
				  BYTE **udpframe, INT *udpframelen)
{
INT i;
SNMP_PDU pdu;

	pdu.Type = request;

	if (pdu.Type == SNMP_PDU_TRAP)
	{
		for (i = 0; i < snmpframe->Pdu.Trap.IdLen; i++)
			pdu.Trap.Id[i] = snmpframe->Pdu.Trap.Id[i];
		pdu.Trap.IdLen = snmpframe->Pdu.Trap.IdLen;
		pdu.Trap.IpAddress = snmpframe->Pdu.Trap.IpAddress;
		pdu.Trap.General = snmpframe->Pdu.Trap.General;
		pdu.Trap.Specific = snmpframe->Pdu.Trap.Specific;
		pdu.Trap.Time = snmpframe->Pdu.Trap.Time;
	}
	else
	{
		pdu.Request.Id = snmpframe->Pdu.Request.Id;
		pdu.Request.ErrorStatus = snmpframe->Pdu.Request.Status;
		pdu.Request.ErrorIndex = snmpframe->Pdu.Request.Index;
	}

	if (SnmpEnc(udpframe, udpframelen, &pdu, 
				community, commlen, 
				snmpframe->Objects, snmpframe->NrObjects))
		return TRUE;

	return FALSE;
}


BOOLEAN Udp2Frame(INT *request, BYTE *community, INT *commlen, 
				  BYTE *udpframe, INT udpframelen, 
				  NwSnmpFrame *snmpframe)
{
INT i;
SNMP_PDU pdu;
UINT nrobjects;

	if (SnmpDec(udpframe, udpframelen, &pdu, 
				community, *commlen, commlen,
				snmpframe->Objects, snmpframe->NrObjects, &nrobjects))
	{
		*request = pdu.Type;

		if (pdu.Type == SNMP_PDU_TRAP)
		{
			for (i = 0; i < pdu.Trap.IdLen; i++)
				snmpframe->Pdu.Trap.Id[i] = pdu.Trap.Id[i];
			snmpframe->Pdu.Trap.IdLen = pdu.Trap.IdLen;
			snmpframe->Pdu.Trap.IpAddress = pdu.Trap.IpAddress;
			snmpframe->Pdu.Trap.General = pdu.Trap.General;
			snmpframe->Pdu.Trap.Specific = pdu.Trap.Specific;
			snmpframe->Pdu.Trap.Time = pdu.Trap.Time;
		}
		else
		{
			snmpframe->Pdu.Request.Id = pdu.Request.Id;
			snmpframe->Pdu.Request.Status = pdu.Request.ErrorStatus;
			snmpframe->Pdu.Request.Index = pdu.Request.ErrorIndex;
		}

		snmpframe->NrObjects = nrobjects;

		return TRUE;
	}

	return FALSE;
}


BOOLEAN Addr2Snmp(CHAR *addr, CHAR *host, INT *port, 
				  CHAR *request, WORD *type, BYTE *community, INT *commlen)
{
CHAR *p, *p2;
CHAR udpport[NWSNMP_SIZE_UDPPORT];

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
        return FALSE;
    
    strncpy(host, addr, p-addr);
    host[p-addr] = '\0';
    
	if ((p2 = strchr(++p, ':')) == NULL)
	    return FALSE;
    
	strncpy(udpport, p, p2-p);
	udpport[p2-p] = '\0';
	if ((*port = atoi(udpport)) <= 0)
        return FALSE;

	p = p2;
	if ((p2 = strchr(++p, ':')) == NULL)
	    return FALSE;
    
	strncpy(request, p, p2-p);
	request[p2-p] = '\0';

	if (strcmp(request, "get") == 0)
		*type = SNMP_PDU_GET;
	else
	if (strcmp(request, "set") == 0)
		*type = SNMP_PDU_SET;
	else
	if (strcmp(request, "next") == 0)
		*type = SNMP_PDU_NEXT;
	else
	if (strcmp(request, "trap") == 0)
		*type = SNMP_PDU_TRAP;
	else
	if (strcmp(request, "response") == 0)
		*type = SNMP_PDU_RESPONSE;
	else
	    return FALSE;

	*commlen = strlen(++p2);
	memcpy(community, p2, *commlen);
	community[*commlen] = '\0';

	return TRUE;
}
