package fsm;
#########################################################################
## MODULE INFORMATION#
######################
##     FILE     NAME:       fsm.pl
##     SYSTEM   NAME:       fsm
##     ORIGINAL AUTHOR(S):  Jan van Oorschot
##     VERSION  NUMBER:     1.00
##     CREATION DATE:       Tue Jul 26 08:20:30 1994
##
## DESCRIPTION: Library providing Finite State Machine
##              functionality. Due to the limitations of Perl,
##              some of the implementation details are not 
##              very clean.
##              
## Variables maintained in Init fase:            
##                
##   <prefix>SName[state_nr]         = "state-name"
##   <prefix>ISName{"state-name"}    = state_nr;
##   <prefix>SFun[state_nr]          = fun
##   <prefix>STran<state_nr>[signal] = "state_name"
##              
## Variables maintained in Normal fase:
##              
##   <prefix>SName[state_nr]         = "state_name"
##   <prefix>ISName{"state-name"}    = state_nr;
##   <prefix>SFun[state_nr]          = fun
##   <prefix>STran<state_nr>[signal] = state_nr;     # resolved
##              
##
## Functions offered are:
##
##  FSM_Create(<prefix>)
## 
##  FSM_State(<prefix>,"state_name",fun)
##
##  FSM_Transition(<prefix>,"state_from",signal,"state_to")
##
##  FSM_Resolve(<prefix>)
##              
##  FSM_Run(<prefix>,"state_name")            
##              
##  FSM_Dump(<prefix>)
##              
##  FSM_Destroy(<prefix>)
##
## A Finit State Machine normally requires varaibles to preserved.
## In the case of the library, one state machine will often be
## used multiple times, and by different tasks.
## To support this, a task switching mechanisme is implemented.
## The following data structures are maintained by this
## mechanisme:
##
##  <prefix>Vars[VarNr]          = name of variable <VarNr>
##  <prefix>All<VarNr>{$id}      = value in task $id of variable <VarNr>
##
## Functions offered are:
##
##  FSM_TaskInit(<prefix> , @vars)
##
##  FSM_TaskSwitchIn(<prefix>,$id)
##
##  FSM_TaskSwitchOut(<prefix>,$id)
##
#########################################################################
## CHANGE INFORMATION ##
#########################
##      JvO - Jul 26, 1994:  Created
#########################################################################
$Debug = 0;

sub main'FSM_Create
{
    local($prefix) = @_;
    local(*sname) = $prefix . "SName";
    local(*isname) = $prefix . "ISName";
    local(*sfun) = $prefix . "SFun";

    @sname = ();
    %isname = ();
    @sfun = ();
    
    $sname[0] = "exit";
    $isname{"exit"} = 0;

    return 1;
}

sub main'FSM_State
{
    local($prefix,$state_name,$fun) = @_;
    local(*sname) = $prefix . "SName";
    local(*isname) = $prefix . "ISName";
    local(*sfun) = $prefix . "SFun";
    local($nr);
    
    push(@sname,$state_name);
    $nr = $#sname;
    local(*stran) = $prefix . "STran" . $nr;

    if(defined($isname{$state_name}))
    {
	$main'TrickletErrStr = "FSM: double state definition $state_name\n";
	$main'TrickletErrNo = 201;
	return undef;
    }
    $isname{$state_name} = $nr;
    if( ! defined(&$fun) )
    {
	$main'TrickletErrStr = "FSM: function $fun does not exist\n";
	$main'TrickletErrNo = 202;
	return undef;
    }
    $sfun[$nr] = $fun;
    %stran = ();

    return 1;
}

sub main'FSM_Transition
{
    local($prefix,$sfrom,$signal,$sto) = @_;
    local(*sname) = $prefix . "SName";
    local(*isname) = $prefix . "ISName";
    local(*sfun) = $prefix . "SFun";
    local($nr);

    if( ! defined($isname{$sfrom}) )
    {
	$main'TrickletErrStr = "FSM: state $sfrom does not exist, cant define transition\n";
	$main'TrickletErrNo = 203;
	return undef;
    }	
    $nr = $isname{$sfrom};
    local(*stran) = $prefix . "STran" . $nr;

    $stran[$signal] = $sto;

    return 1;
}

sub main'FSM_Resolve
{
    local($prefix) = @_;
    local(*sname) = $prefix . "SName";
    local(*isname) = $prefix . "ISName";
    local(*sfun) = $prefix . "SFun";
    local($nr,$snr);

    for($nr=0;$nr<=$#sname;$nr++)
    {
	local(*stran) = $prefix . "STran" . $nr;
	for($snr=0;$snr<=$#stran;$snr++)
	{
	    if( ! defined($isname{$stran[$snr]}) )
	    {
		$main'TrickletErrStr = "FSM: resolving $stran[$snr] failed\n";
		$main'TrickletErrNo = 204;
		return undef;
	    }
	    $stran[$snr] = $isname{$stran[$snr]};
	}
    }
    return 1;
}

sub main'FSM_Dump
{
    local($prefix) = @_;
    local(*sname) = $prefix . "SName";
    local(*isname) = $prefix . "ISName";
    local(*sfun) = $prefix . "SFun";
    local($nr,$snr);

    for($nr=0;$nr<=$#sname;$nr++)
    {
	printf("%sSName[%d] = %s\n",$prefix,$nr,$sname[$nr]);
	printf("%sISName{%s} = %d\n",
	       $prefix,$sname[$nr],$nr);
	local(*stran) = $prefix . "STran" . $nr;
	for($snr=0;$snr<=$#stran;$snr++)
	{
	    printf("%sSTran<%d(%s)>[%d] = %d(%s)\n",
		   $prefix,
		   $nr,$sname[$nr],
		   $snr,
		   $stran[$snr],$sname[$stran[$snr]]);
	}
	printf("\n");
    }
    return 1;
}

sub main'FSM_Run
{
    local($prefix,$state_name) = @_;
    local(*sname) = $prefix . "SName";
    local(*isname) = $prefix . "ISName";
    local(*sfun) = $prefix . "SFun";
    local($nr,$fun,$signal);

    if($Debug)
    {
	print "FSM: starting in state $state_name\n";
    }				
    if( ! defined($isname{$state_name}) )
    {
	$main'TrickletErrStr = "FSM: unknown starting state $state_name\n";
	$main'TrickletErrNo = 205;
	return undef;
    }

    $nr = $isname{$state_name};
    
    while($nr > 0)
    {
        if($Debug)
        {
	    $st = $sname[$nr]; 
	    print "FSM: running state $st\n";
        }				
	$fun = $sfun[$nr];
	$signal = &$fun();
        if($Debug)
        {
	    print "FSM: state $st returned signal $signal\n";
        }				
	local(*stran) = $prefix . "STran" . $nr;
	if( ! defined $stran[$signal] )
	{
	    $main'TrickletErrStr = "FSM: $fun returns undefined signal $signal\n";
	    $main'TrickletErrNo = 206;
	    return undef;
	}
	$nr = $stran[$signal];
    }
    if($Debug)
    {
	print "FSM: exit\n";
    }				
    return 1;
}
    
sub main'FSM_TaskInit
{
    local($prefix,@vars) =@_;
    local($var);
    local(*avar) = $prefix . "Vars" ;
    @avar = @vars;
    $nr = 0;
    foreach $var (@vars)
    {
	local(*ptr)  = $var;
	local(*aptr) = $prefix . "All" . $nr++;
	%aptr = ();
    }
}

sub main'FSM_TaskSwitchIn
{
    local($prefix,$id) = @_;
    local(*avar) = $prefix . "Vars" ;
    $nr = 0;
    foreach $var (@avar)
    {
	local(*ptr)  = $var;
	local(*aptr) = $prefix . "All" . $nr++;
	$ptr = $aptr{$id};
    }
}

sub main'FSM_TaskSwitchOut
{
    local($prefix,$id) = @_;
    local(*avar) = $prefix . "Vars" ;
    $nr = 0;
    foreach $var (@avar)
    {
	local(*ptr)  = $var;
	local(*aptr) = $prefix . "All" . $nr++;
	$aptr{$id} = $ptr;
    }
}

sub main'FSM_TaskDump
{
    local($prefix) = @_;
    local(*avar) = $prefix . "Vars" ;
    $nr = 0;
    foreach $var (@avar)
    {
	local(*ptr)  = $var;
	local(*aptr) = $prefix . "All" . $nr++;
	print("variable $var = <$ptr>\n");
	foreach $key (keys(%aptr))
	{
	    printf("%sAll%d{%s} = %s\n",
		   $prefix,$nr-1,$key,
		   $aptr{$key});
	}
	print "\n";
    }
}
1;
    
