package discconn;
#
#	This module creates and interrogates a
#       Connection Discovery Collector.
#       Such a collector examines all frames and builds a
#       Source/Destination matrix containing the Ether/IP/UDP
#       and Ether/IP/TCP connection it sees. 
#       
#       This modules implements the following functions:
#
#       discConnCreator($connection,$ownerstring,$interface-number)
#
#               . creates a Connection Discovery Collector
#                   <$connection> should be a Spook connection, and
#                 should point to the correct RMON monitor with
#                 the appropriate community string set.
#                   <$ownerstring> is the identifier for
#                 this collector
#                   <interface-number> is an integer identifying
#                 the network interface in the RMON monitor to be 
#                 used for building the connection matrix.
#               . returns a integer representing the collector number
#                 or 0 if and error occured.
#
#       discConnTerminator($connection,$collector,$ownerstring)
# 
#               . terminates a Connection Discovery COllector
#                   <$connection> should be a correct
#                 spook for this operation.
#               .   <$collector> should be the collector number
#                 for the collector to be destroyed.
#               .   <ownerstring> should be the string
#                 identifying the collector
#
#        
#       discConnGardner($connection,$collector,$id-string,$file-hdnl)
#
#               . collects the date from the Connection Discovery 
#                 Collector. The data is stored in a file in an
#                 ASCI format, each line containing the data
#                 about one connection
#                   <$connection> shoould be a correct connection
#                 for this operation
#                   <$collector> is the collector id 
#                   <file-hndl> file handle of file where
#                   results are written to.
#                   The format of this file can be derived by looking
#                   at the source code of this perl script :-).
#                   But in short, every line contains a row
#                   from one of the tables in the Discovery Collector.
#                   The first and the last line contain starting
#                   and stop time and status info of the collector. 
#####################################################################
require "spook.pl";	   
require "rmon/rmon.pl";
#
# define the functions
#
#
# function iniFun will be called during initialisation
# of the collector. This is the place to set all kind
# of collector specific values. In the case of etherStats, 
# this is only the ethernet-interface number etherStatsDataSource
#
$iniIface = "ifIndex[1]";
sub discConnCreatorCB
{
	local($connNr , $collName , $collNr) = @_;
	local($SNMPRequest , $res);
	
	$SNMPRequest = $collName . "DataSource\[$collNr\]=$iniIface";
	$res = &main'snmp_set($connNr,$SNMPRequest);
	if($res ne $iniIface)
	{
		return 0;
	}
	return 1;
}	   

# create the etherStats collector with the correct inFun
sub main'discConnCreator
{
    local($conn,$owner,$iface) = @_;

    $iniIface = $iface;
    
    $collNr = &main'create_new_collector(
				    $conn,
				    "discConnControl",
				    $owner,
				    discconn'discConnCreatorCB) || return 0;
    return( $collNr) ;
}

# destroy the discConn collector
sub main'discConnTerminator
{
    local($conn,$coll, $owner) = @_;

    return &main'remove_collector($conn,"discConnControl",$coll,$owner);
}

#
# receive a row of the table, for the use with snmp_tbl
#
sub discConnGardnerCB
{
	local($i,$j,@repl) = @_;
	if($j == 1)
	{
	    print $localFile "\n$localId $localTable ";
	}
	print $localFile "$repl[2] ";
}

#
# request the discConn table
#
# input variables:
#      
#      1: $conn        spook connection number
#      2: $host        string with name of host running Beholder
#      3: $comm        community to use 
#      4: $coll        collector number of discConn collector
#      5: $id          identification string
#      6: $file        file handle to write to
#
sub main'discConnGardner
{
    local( $conn, $coll, $id, $file) = @_;
    local( @res );


    $localId = $id;
    $localFile = $file;

    @res = &main'snmp_get($conn,
         (
            "discConnControlIndex[$coll]",
            "discConnControlDataSource[$coll]",
            "discConnControlTotalConnections[$coll]",
            "discConnControlEthernetConnections[$coll]",
            "discConnControlIPConnections[$coll]",
            "discConnControlUDPConnections[$coll]",
            "discConnControlTCPConnections[$coll]",
            "discConnControlLastDeleteTime[$coll]",
            "discConnControlOwner[$coll]",
            "discConnControlStatus[$coll]"
         ));
    $time = time();
    print $file "\n$id discConnControlStart $time $res[1] $res[2] $res[3] $res[4] $res[5] $res[6] $res[7] $res[8] $res[9] $res[10]";  

    $localTable = "discEthernetConnSD";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discEthernetConnSDSource[$coll]",
            "discEthernetConnSDDestination[$coll]",
            "discEthernetConnSDIndex[$coll]",
            "discEthernetConnSDPkts[$coll]",
            "discEthernetConnSDOctets[$coll]",
            "discEthernetConnSDLRUTime[$coll]"
         ));

    $localTable = "discIPConnSD";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discIPConnSDEthernetSource[$coll]",
            "discIPConnSDSource[$coll]",
            "discIPConnSDEthernetDestination[$coll]",
            "discIPConnSDDestination[$coll]",
            "discIPConnSDIndex[$coll]",
            "discIPConnSDPkts[$coll]",
            "discIPConnSDOctets[$coll]",
            "discIPConnSDLRUTime[$coll]"
         ));

    $localTable = "discUDPConnSD";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discUDPConnSDEthernetSource[$coll]",
            "discUDPConnSDIPSource[$coll]",
            "discUDPConnSDSource[$coll]",
            "discUDPConnSDEthernetDestination[$coll]",
            "discUDPConnSDIPDestination[$coll]",
            "discUDPConnSDDestination[$coll]",
            "discUDPConnSDIndex[$coll]",
            "discUDPConnSDPkts[$coll]",
            "discUDPConnSDOctets[$coll]",
            "discUDPConnSDLRUTime[$coll]"
         ));
			      
    $localTable = "discTCPConnSD";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discTCPConnSDEthernetSource[$coll]",
            "discTCPConnSDIPSource[$coll]",
            "discTCPConnSDSource[$coll]",
            "discTCPConnSDEthernetDestination[$coll]",
            "discTCPConnSDIPDestination[$coll]",
            "discTCPConnSDDestination[$coll]",
            "discTCPConnSDIndex[$coll]",
            "discTCPConnSDPkts[$coll]",
            "discTCPConnSDOctets[$coll]",
            "discTCPConnSDLRUTime[$coll]"
         ));

    $localTable = "discEthernetConnDS";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discEthernetConnDSSource[$coll]",
            "discEthernetConnDSDestination[$coll]",
            "discEthernetConnDSIndex[$coll]",
            "discEthernetConnDSPkts[$coll]",
            "discEthernetConnDSOctets[$coll]",
            "discEthernetConnDSLRUTime[$coll]"
         ));

    $localTable = "discIPConnDS";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discIPConnDSEthernetSource[$coll]",
            "discIPConnDSSource[$coll]",
            "discIPConnDSEthernetDestination[$coll]",
            "discIPConnDSDestination[$coll]",
            "discIPConnDSIndex[$coll]",
            "discIPConnDSPkts[$coll]",
            "discIPConnDSOctets[$coll]",
            "discIPConnDSLRUTime[$coll]"
         ));

    $localTable = "discUDPConnDS";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discUDPConnDSEthernetSource[$coll]",
            "discUDPConnDSIPSource[$coll]",
            "discUDPConnDSSource[$coll]",
            "discUDPConnDSEthernetDestination[$coll]",
            "discUDPConnDSIPDestination[$coll]",
            "discUDPConnDSDestination[$coll]",
            "discUDPConnDSIndex[$coll]",
            "discUDPConnDSPkts[$coll]",
            "discUDPConnDSOctets[$coll]",
            "discUDPConnDSLRUTime[$coll]"
         ));

    $localTable = "discTCPConnDS";
    &main'snmp_tbl($conn,discconn'discConnGardnerCB,
         (
            "discTCPConnDSEthernetSource[$coll]",
            "discTCPConnDSIPSource[$coll]",
            "discTCPConnDSSource[$coll]",
            "discTCPConnDSEthernetDestination[$coll]",
            "discTCPConnDSIPDestination[$coll]",
            "discTCPConnDSDestination[$coll]",
            "discTCPConnDSIndex[$coll]",
            "discTCPConnDSPkts[$coll]",
            "discTCPConnDSOctets[$coll]",
            "discTCPConnDSLRUTime[$coll]"
         ));

    @res = &main'snmp_get($conn,
         (
            "discConnControlIndex[$coll]",
            "discConnControlDataSource[$coll]",
            "discConnControlTotalConnections[$coll]",
            "discConnControlEthernetConnections[$coll]",
            "discConnControlIPConnections[$coll]",
            "discConnControlUDPConnections[$coll]",
            "discConnControlTCPConnections[$coll]",
            "discConnControlLastDeleteTime[$coll]",
            "discConnControlOwner[$coll]",
            "discConnControlStatus[$coll]"
         ));
    $time = time();
    print $file "\n$id discConnControlEnd $time $res[1] $res[2] $res[3] $res[4] $res[5] $res[6] $res[7] $res[8] $res[9] $res[10]";  

    return 1;
}
1;





