package discnode;
#
#	This module creates and interrogates a
#       Node Discovery Collector.
#       Such a collector examines all frames and builds a
#       node array containing the Ether/IP/UDP
#       and Ether/IP/TCP information of all nodes it has Seen
#       
#       This modules implements the following functions:
#
#       discNodeCreator($connection,$ownerstring,$interface-number)
#
#               . creates a Node Discovery Collector
#                   <$connection> should be a Spook connection, and
#                 should point to the correct RMON monitor with
#                 the appropriate community string set.
#                   <$ownerstring> is the identifier for
#                 this collector
#                   <interface-number> is an integer identifying
#                 the network interface in the RMON monitor to be 
#                 used for building the node array.
#               . returns a integer representing the collector number
#                 or 0 if and error occured.
#
#       discNodeTerminator($connection,$collector,$ownerstring)
# 
#               . terminates a Node Discovery COllector
#                   <$connection> should be a correct
#                 spook for this operation.
#               .   <$collector> should be the collector number
#                 for the collector to be destroyed.
#               .   <ownerstring> should be the string
#                 identifying the collector
#
#        
#       discNodeGardner($connection,$collector,$id-string,$file-hdnl)
#
#               . collects the date from the Node Discovery 
#                 Collector. The data is stored in a file in an
#                 ASCI format, each line containing the data
#                 about one connection
#                   <$connection> shoould be a correct connection
#                 for this operation
#                   <$collector> is the collector id 
#                   <file-hndl> file handle of file where
#                   results are written to.
#                   The format of this file can be derived by looking
#                   at the source code of this perl script :-).
#                   But in short, every line contains a row
#                   from one of the tables in the Discovery Collector.
#                   The first and the last line contain starting
#                   and stop time and status info of the collector. 
#####################################################################
#
require "spook.pl";	   
require "rmon/rmon.pl";
#
# define the functions
#
#
# function iniFun will be called during initialisation
# of the collector. This is the place to set all kind
# of collector specific values. In the case of etherStats, 
# this is only the ethernet-interface number etherStatsDataSource
#
$iniIface = "ifIndex[1]";
sub discNodeCreatorCB
{
	local($connNr , $collName , $collNr) = @_;
	local($SNMPRequest , $res);
	
	$SNMPRequest = $collName . "DataSource\[$collNr\]=$iniIface";
	$res = &main'snmp_set($connNr,$SNMPRequest);
	if($res ne $iniIface)
	{
		return 0;
	}
	return 1;
}	   

# create the etherStats collector with the correct inFun
sub main'discNodeCreator
{
    local($conn,$owner,$iface) = @_;

    $iniIface = $iface;
    
    $collNr = &main'create_new_collector(
				    $conn,
				    "discNodeControl",
				    $owner,
				    discnode'discNodeCreatorCB) || return 0;
    return( $collNr) ;
}

# destroy the discNode collector
sub main'discNodeTerminator
{
    local($conn,$coll, $owner) = @_;

    return &main'remove_collector($conn,"discNodeControl",$coll,$owner);
}

#
# receive a row of the table, for the use with snmp_tbl
#
sub discNodeGardnerCB
{
	local($i,$j,@repl) = @_;
	if($j == 1)
	{
	    print $localFile "\n$localId $localTable ";
	}
	print $localFile "$repl[2] ";
}

#
# request the discNode table
#
# input variables:
#      
#      1: $conn        spook connection number
#      2: $host        string with name of host running Beholder
#      3: $comm        community to use 
#      4: $coll        collector number of discNode collector
#      5: $id          identification string
#      6: $file        file handle to write to
#
sub main'discNodeGardner
{
    local( $conn, $coll, $id, $file) = @_;
    local( @res );

    $localId = $id;
    $localFile = $file;

    $localTable = "discNodeControl";
    @res = &main'snmp_get($conn,
         (
            "discNodeControlIndex[$coll]",
            "discNodeControlDataSource[$coll]",
            "discNodeControlTotalAccessPoints[$coll]",
            "discNodeControlEthernetAccessPoints[$coll]",
            "discNodeControlIPAccessPoints[$coll]",
            "discNodeControlUDPAccessPoints[$coll]",
            "discNodeControlTCPAccessPoints[$coll]",
            "discNodeControlLastDeleteTime[$coll]",
            "discNodeControlOwner[$coll]",
            "discNodeControlStatus[$coll]"
         ));
    $time = time();
    print $file "\n$id discNodeControlStart $time  $res[1] $res[2] $res[3] $res[4] $res[5] $res[6] $res[7] $res[8] $res[9] $res[10]";  

    $localTable = "discEthernetNode";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
            "discEthernetNodeAddress[$coll]",
            "discEthernetNodeCreationOrder[$coll]",
            "discEthernetNodeIndex[$coll]",
            "discEthernetNodeInPkts[$coll]",
            "discEthernetNodeInOctets[$coll]",
            "discEthernetNodeOutPkts[$coll]",
            "discEthernetNodeOutOctets[$coll]",
            "discEthernetNodeLRUTime[$coll]"
         ));

    $localTable = "discIPNode";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
	    "discIPNodeEthernetAddress[$coll]",
            "discIPNodeAddress[$coll]",
            "discIPNodeCreationOrder[$coll]",
            "discIPNodeIndex[$coll]",
            "discIPNodeInPkts[$coll]",
            "discIPNodeInOctets[$coll]",
            "discIPNodeOutPkts[$coll]",
            "discIPNodeOutOctets[$coll]",
            "discIPNodeLRUTime[$coll]"
         ));

    $localTable = "discUDPNode";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
	    "discUDPNodeEthernetAddress[$coll]",
            "discUDPNodeIPAddress[$coll]",
            "discUDPNodeAddress[$coll]",
            "discUDPNodeCreationOrder[$coll]",
            "discUDPNodeIndex[$coll]",
            "discUDPNodeInPkts[$coll]",
            "discUDPNodeInOctets[$coll]",
            "discUDPNodeOutPkts[$coll]",
            "discUDPNodeOutOctets[$coll]",
            "discUDPNodeLRUTime[$coll]"
         ));

    $localTable = "discTCPNode";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
	    "discTCPNodeEthernetAddress[$coll]",
            "discTCPNodeIPAddress[$coll]",
            "discTCPNodeAddress[$coll]",
            "discTCPNodeCreationOrder[$coll]",
            "discTCPNodeIndex[$coll]",
            "discTCPNodeInPkts[$coll]",
            "discTCPNodeInOctets[$coll]",
            "discTCPNodeOutPkts[$coll]",
            "discTCPNodeOutOctets[$coll]",
            "discTCPNodeLRUTime[$coll]"
         ));

    $localTable = "discEthernetNodeTime";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
	    "discEthernetNodeTimeAddress[$coll]",
            "discEthernetNodeTimeCreationOrder[$coll]",
            "discEthernetNodeTimeIndex[$coll]",
            "discEthernetNodeTimeInPkts[$coll]",
            "discEthernetNodeTimeInOctets[$coll]",
            "discEthernetNodeTimeOutPkts[$coll]",
            "discEthernetNodeTimeOutOctets[$coll]",
            "discEthernetNodeTimeLRUTime[$coll]"
         ));

    $localTable = "discIPNodeTime";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
	    "discIPNodeTimeEthernetAddress[$coll]",
            "discIPNodeTimeAddress[$coll]",
            "discIPNodeTimeCreationOrder[$coll]",
            "discIPNodeTimeIndex[$coll]",
            "discIPNodeTimeInPkts[$coll]",
            "discIPNodeTimeInOctets[$coll]",
            "discIPNodeTimeOutPkts[$coll]",
            "discIPNodeTimeOutOctets[$coll]",
            "discIPNodeTimeLRUTime[$coll]"
         ));

    $localTable = "discUDPNodeTime";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
	    "discUDPNodeTimeEthernetAddress[$coll]",
            "discUDPNodeTimeIPAddress[$coll]",
            "discUDPNodeTimeAddress[$coll]",
            "discUDPNodeTimeCreationOrder[$coll]",
            "discUDPNodeTimeIndex[$coll]",
            "discUDPNodeTimeInPkts[$coll]",
            "discUDPNodeTimeInOctets[$coll]",
            "discUDPNodeTimeOutPkts[$coll]",
            "discUDPNodeTimeOutOctets[$coll]",
            "discUDPNodeTimeLRUTime[$coll]"
         ));

    $localTable = "discTCPNodeTime";
    &main'snmp_tbl($conn,discnode'discNodeGardnerCB,
         (
	    "discTCPNodeTimeEthernetAddress[$coll]",
            "discTCPNodeTimeIPAddress[$coll]",
            "discTCPNodeTimeAddress[$coll]",
            "discTCPNodeTimeCreationOrder[$coll]",
            "discTCPNodeTimeIndex[$coll]",
            "discTCPNodeTimeInPkts[$coll]",
            "discTCPNodeTimeInOctets[$coll]",
            "discTCPNodeTimeOutPkts[$coll]",
            "discTCPNodeTimeOutOctets[$coll]",
            "discTCPNodeTimeLRUTime[$coll]"
         ));

    $localTable = "discNodeControl";
    @res = &main'snmp_get($conn,
         (
            "discNodeControlIndex[$coll]",
            "discNodeControlDataSource[$coll]",
            "discNodeControlTotalAccessPoints[$coll]",
            "discNodeControlEthernetAccessPoints[$coll]",
            "discNodeControlIPAccessPoints[$coll]",
            "discNodeControlUDPAccessPoints[$coll]",
            "discNodeControlTCPAccessPoints[$coll]",
            "discNodeControlLastDeleteTime[$coll]",
            "discNodeControlOwner[$coll]",
            "discNodeControlStatus[$coll]"
         ));
    $time = time();
    print $file "\n$id discNodeControlEnd $time  $res[1] $res[2] $res[3] $res[4] $res[5] $res[6] $res[7] $res[8] $res[9] $res[10]";  
    return 1;
}
1;





