package history;
#
#	This module creates and interrogates a
#       History Collector.
#       Such a collector examines all frames and builds a
#       node array containing the Ethernet
#       information of all ethernet historys it has seen
#       
#       This modules implements the following functions:
#
#       historyCreator($connection,$ownerstring,$interface-number)
#
#               . creates a history collector
#                   <$connection> should be a Spook connection, and
#                 should point to the correct RMON monitor with
#                 the appropriate community string set.
#                   <$ownerstring> is the identifier for
#                 this collector
#                   <interface-number> is an integer identifying
#                 the network interface in the RMON monitor to be 
#                 used for building the node array.
#               . returns a integer representing the collector number
#                 or 0 if and error occured.
#
#       historyTerminator($connection,$collector,$ownerstring)
# 
#               . terminates a history Collector
#                   <$connection> should be a correct
#                 spook for this operation.
#               .   <$collector> should be the collector number
#                 for the collector to be destroyed.
#               .   <ownerstring> should be the string
#                 identifying the collector
#
#        
#       historyGardner($connection,$collector,$id-string,$file-hdnl)
#
#               . collects the date from the history
#                 Collector. The data is stored in a file in an
#                 ASCI format, each line containing the data
#                 about one connection
#                   <$connection> shoould be a correct connection
#                 for this operation
#                   <$collector> is the collector id 
#                   <file-hndl> file handle of file where
#                   results are written to.
#                   The format of this file can be derived by looking
#                   at the source code of this perl script :-).
#                   But in short, every line contains a row
#                   from one of the tables in the Discovery Collector.
#                   The first and the last line contain starting
#                   and stop time and status info of the collector. 
#####################################################################
require "spook.pl";	   
require "rmon/rmon.pl";
#
# define the functions
#
#
# function iniFun will be called during initialisation
# of the collector. This is the place to set all kind
# of collector specific values. In the case of etherStats, 
# this is only the ethernet-interface number etherStatsDataSource
#
$iniIface = "ifIndex[1]";
$iniInt = 10;
$iniBuck = 4;

sub historyCreatorCB
{
	local($connNr , $collName , $collNr) = @_;
	local($varName , $res , $iface , $int);
	
	$res = &main'snmp_set($connNr,$collName . "DataSource\[$collNr\]=$iniIface");
	if($res ne $iniIface)
	{
		return 0;
	}

	$res = &main'snmp_set($connNr,$collName . "Interval\[$collNr\]=$iniInt");
	if($res != $iniInt)
	{
		return 0;
	}
 	$res = &main'snmp_set($connNr,$collName . "BucketsRequested\[$collNr\]=$iniBuck");
	if($res != $iniBuck)
	{
		return 0;
	}
	return 1;
}	   

# create the etherStats collector with the correct inFun
sub main'historyCreator
{
    local($conn,$owner,$iface,$int,$buck) = @_;

    $iniIface = $iface;
    $iniInt = $int;
    $iniBuck = $buck;
    
    $collNr = &main'create_new_collector(
				    $conn,
				    "historyControl",
				    $owner,
				    history'historyCreatorCB) || return 0;
    return( $collNr) ;
}

# destroy the history collector
sub main'historyTerminator
{
    local($conn,$coll, $owner) = @_;

    return &main'remove_collector($conn,"historyControl",$coll,$owner);
}

#
# receive a row of the table, for the use with snmp_tbl
#
sub historyGardnerCB
{
	local($i,$j,@repl) = @_;
	if($j == 1)
	{
	    print $localFile "\n$localId $localTable ";
	}
	print $localFile "$repl[2] ";
}

#
# request the history table
#
# input variables:
#      
#      1: $conn        spook connection number
#      2: $coll        collector number of host collector
#      3: $id          identification string
#      6: $file        file handle to write to
#
sub main'historyGardner
{
    local( $conn, $coll, $id, $file) = @_;
    local( @res );

    $localId = $id;
    $localFile = $file;

    $localTable = "historyControl";
    @res = &main'snmp_get($conn,
         (
            "sysUpTime[0]",
            "historyControlIndex[$coll]",
            "historyControlDataSource[$coll]",
            "historyControlOwner[$coll]",
            "historyControlStatus[$coll]",
            "historyControlInterval[$coll]",
            "historyControlBucketsRequested[$coll]"
         ));
    $time = time();
    print $file "\n$id historyControlStart $time $res[0] $res[1] $res[2] $res[3] $res[4] $res[5] $res[6]";  

    $localTable = "etherHistory";
    &main'snmp_tbl($conn,history'historyGardnerCB,
         (
	  "etherHistorySampleIndex[$coll]",
	  "etherHistoryIntervalStart[$coll]",
	  "etherHistoryDropEvents[$coll]",
	  "etherHistoryOctets[$coll]",
	  "etherHistoryUtilization[$coll]",
	  "etherHistoryPkts[$coll]",
	  "etherHistoryBroadcastPkts[$coll]",
	  "etherHistoryMulticastPkts[$coll]"
         ));
    $localTable = "historyControl";
    @res = &main'snmp_get($conn,
         (
            "sysUpTime[0]",
            "historyControlIndex[$coll]",
            "historyControlDataSource[$coll]",
            "historyControlOwner[$coll]",
            "historyControlStatus[$coll]",
            "historyControlInterval[$coll]",
            "historyControlBucketsRequested[$coll]"
         ));
    $time = time();
    print $file "\n$id historyControlEnd $time $res[0] $res[1] $res[2] $res[3] $res[4] $res[5] $res[6]";  
    return 1;
}
#
# Display a file containing the result of the gardner using a
# gnuplot process
#
# 1:   $file   file name of file containing gardner output
# 2:   $plot   file-handle of gnuplot process
#
sub main'historyPainter
{
    local($file,$plot) = @_;
    local($line,$load);
    local($lRealTime,$lMonTime,$lMinBuck,$lMinTime,$lMaxBuck,$lMaxTime);

    open(FILE,"<$file") || return 0;
    open(GFILE,">/tmp/history.tmp") || return 0;
    print GFILE "# gnu history plot file \n";
    $lMinBuck = 64000;
    $lMaxBuck = 0;
    while($line = <FILE>)
    {
	@res = split(' ',$line);
	if($res[1] eq "historyControlStart")
	{
	    $lRealTime = $res[2];
	    $lMonTime = $res[3];
	}
	if($res[1] eq "etherHistory")
	{
	    if( $lMinBuck > $res[2])
	    {
		$lMinBuck = $res[2];
		$lMinTime = $res[3];
	    }
	    if( $lMaxBuck < $res[2])
	    {
		$lMaxBuck = $res[2];
		$lMaxTime = $res[3];
	    }
	    $load = $res[6]/10000;  # load goes from 0 to 1
	    print GFILE "$res[2] $load\n";
	}
    }
    close(GFILE);
    close(FILE);

    if($lMinBuck > $lMaxBuck)
    {
	return 0;		# no measurement
    }
    # derive x axis and x ticks from the collected time info
    $nrTic = 2;
    $lTimeOffset = $lRealTime - $lMonTime/100;
    $lRMinTime = $lMinTime/1000 + $lTimeOffset;
    $lRMaxTime = $lMaxTime/1000 + $lTimeOffset;
    $timeStep = ($lRMaxTime - $lRMinTime)/($nrTic+1);
    $buckStep = ($lMaxBuck - $lMinBuck)/($nrTic+1);

    print $plot "set xrange [$lMinBuck:$lMaxBuck]\n";

    print $plot "set xtics (";
    ($sec,$min,$hour,@res) = localtime($lRMinTime);
    printf($plot "\"%02d:%02d.%02d\" %f ,",$hour,$min,$sec,$lMinBuck);
    $curBuck = $lMinBuck;
    for(
	$curTime = $lRMinTime + $timeStep; 
	$curTime <$lRMaxTime; 
	$curTime = $curTime + $timeStep
       )
    {
	$curBuck = $curBuck + $buckStep;
	($sec,$min,$hour,@res) = localtime($curTime);
	printf($plot "\"%02d:%02d.%02d\" %f ,",$hour,$min,$sec,$curBuck);
    }
    ($sec,$min,$hour,@res) = localtime($lRMaxTime);
    printf($plot "\"%02d:%02d.%02d  \" %f )\n",$hour,$min,$sec,$lMaxBuck);

    print $plot "set yrange [0:0.5]\n";
    print $plot "plot \'/tmp/history.tmp\' title 'network load' with lines\n";
    return 1;
}
1;





