package host;
#
#	This module creates and interrogates a
#       Host Collector.
#       Such a collector examines all frames and builds a
#       node array containing the Ethernet
#       information of all ethernet hosts it has seen
#       
#       This modules implements the following functions:
#
#       hostCreator($connection,$ownerstring,$interface-number)
#
#               . creates a host collector
#                   <$connection> should be a Spook connection, and
#                 should point to the correct RMON monitor with
#                 the appropriate community string set.
#                   <$ownerstring> is the identifier for
#                 this collector
#                   <interface-number> is an integer identifying
#                 the network interface in the RMON monitor to be 
#                 used for building the node array.
#               . returns a integer representing the collector number
#                 or 0 if and error occured.
#
#       hostTerminator($connection,$collector,$ownerstring)
# 
#               . terminates a host Collector
#                   <$connection> should be a correct
#                 spook for this operation.
#               .   <$collector> should be the collector number
#                 for the collector to be destroyed.
#               .   <ownerstring> should be the string
#                 identifying the collector
#
#        
#       hostGardner($connection,$collector,$id-string,$file-hdnl)
#
#               . collects the data from the host
#                 Collector. The data is stored in a file in an
#                 ASCI format, each line containing the data
#                 about one connection
#                   <$connection> shoould be a correct connection
#                 for this operation
#                   <$collector> is the collector id 
#                   <file-hndl> file handle of file where
#                   results are written to.
#                   The format of this file can be derived by looking
#                   at the source code of this perl script :-).
#                   But in short, every line contains a row
#                   from one of the tables in the Discovery Collector.
#                   The first and the last line contain starting
#                   and stop time and status info of the collector. 
#####################################################################
require "spook.pl";	   
require "rmon/rmon.pl";
#
# define the functions
#
#
# function iniFun will be called during initialisation
# of the collector. This is the place to set all kind
# of collector specific values. In the case of etherStats, 
# this is only the ethernet-interface number etherStatsDataSource
#
$iniIface = "ifIndex[1]";
sub hostCreatorCB
{
	local($connNr , $collName , $collNr) = @_;
	local($SNMPRequest , $res);
	
	$SNMPRequest = $collName . "DataSource\[$collNr\]=$iniIface";
	$res = &main'snmp_set($connNr,$SNMPRequest);
	if($res ne $iniIface)
	{
		return 0;
	}
	return 1;
}	   

# create the etherStats collector with the correct inFun
sub main'hostCreator
{
    local($conn,$owner,$iface) = @_;

    $iniIface = $iface;
    
    $collNr = &main'create_new_collector(
				    $conn,
				    "hostControl",
				    $owner,
				    host'hostCreatorCB) || return 0;
    return( $collNr) ;
}

# destroy the host collector
sub main'hostTerminator
{
    local($conn,$coll, $owner) = @_;

    return &main'remove_collector($conn,"hostControl",$coll,$owner);
}

#
# receive a row of the table, for the use with snmp_tbl
#
sub hostGardnerCB
{
	local($i,$j,@repl) = @_;
	if($j == 1)
	{
	    print $localFile "\n$localId $localTable ";
	}
	print $localFile "$repl[2] ";
}

#
# request the host table
#
# input variables:
#      
#      1: $conn        spook connection number
#      4: $coll        collector number of host collector
#      5: $id          identification string
#      6: $file        file handle to write to
#
sub main'hostGardner
{
    local( $conn, $coll, $id, $file) = @_;
    local( @res );

    $localId = $id;
    $localFile = $file;

    $localTable = "hostControl";
    @res = &main'snmp_get($conn,
         (
            "hostControlIndex[$coll]",
            "hostControlDataSource[$coll]",
            "hostControlTableSize[$coll]",
            "hostControlLastDeleteTime[$coll]",
            "hostControlOwner[$coll]",
            "hostControlStatus[$coll]"
         ));
    $time = time();
    print $file "\n$id hostControlStart $time $res[1] $res[2] $res[3] $res[4] $res[5] $res[6]";  

    $localTable = "host";
    &main'snmp_tbl($conn,host'hostGardnerCB,
         (
            "hostAddress[$coll]",
            "hostCreationOrder[$coll]",
            "hostIndex[$coll]",
            "hostInPkts[$coll]",
            "hostOutPkts[$coll]",
            "hostInOctets[$coll]",
            "hostOutOctets[$coll]",
            "hostOutErrors[$coll]",
            "hostOutBroadcastPkts[$coll]",
            "hostOutMulticastPkts[$coll]"
         ));

    $localTable = "hostTime";
    &main'snmp_tbl($conn,host'hostGardnerCB,
         (
            "hostTimeAddress[$coll]",
            "hostTimeCreationOrder[$coll]",
            "hostTimeIndex[$coll]",
            "hostTimeInPkts[$coll]",
            "hostTimeOutPkts[$coll]",
            "hostTimeInOctets[$coll]",
            "hostTimeOutOctets[$coll]",
            "hostTimeOutErrors[$coll]",
            "hostTimeOutBroadcastPkts[$coll]",
            "hostTimeOutMulticastPkts[$coll]"
         ));

    $localTable = "hostControl";
    @res = &main'snmp_get($conn,
         (
            "hostControlIndex[$coll]",
            "hostControlDataSource[$coll]",
            "hostControlTableSize[$coll]",
            "hostControlLastDeleteTime[$coll]",
            "hostControlOwner[$coll]",
            "hostControlStatus[$coll]"
         ));
    $time = time();
    print $file "\n$id hostControlEnd $time $res[1] $res[2] $res[3] $res[4] $res[5] $res[6]";  
    return 1;
}
1;





