/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       nwxmp.c
**     SYSTEM   NAME:       nwxmp
**     ORIGINAL AUTHOR(S):  Jan van Oorschot
**     VERSION  NUMBER:     1.00
**     CREATION DATE:       Tue Dec  7 17:26:42 1993
**
** DESCRIPTION: Example application using the NW, Block and DNPAP
**              libraries.
**		This application performs generalised roundtrip 
**		measurements.
**		It is called as follows:
**
**   nwround <host> <l1min> <l1max> <l1step> <l2min> <l2max> l2step> <nr> <id>
**
**			<host>	host to send to  ex. udp:dutepp0:2000
**			<l1min>	min frame length (UDP) to echoer
**                      <l1max> max frame length (UDP) to echoer
**                      <l1step>step of frame length (UDP) to echoer
**			<l2min>	min frame length (UDP) from echoer
**                      <l2max> max frame length (UDP) from echoer
**                      <l2step>step offrame length (UDP) from echoer 
**			<nr>	number of echo frames
**			<id>	measurement id
**
**		After the measurements, the results will be written
**		to stdout in the following format, each line represents
**		one roundtrip measurement
**
**		<type> <id> <cnt> <l1> <l2> <t1> <t2> <t3> <t4>
*************************************************************************
** CHANGE INFORMATION **
*************************
**      JvO - Dec 7, 1993:  Created
**      RJB - Feb 10, 1994:
**      change length of L1 and L2
*************************************************************************/
/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <message.h>
#include <config.h>
#include <block.h>
#include <timer.h>
#include <dtime.h>
#include "../../include/nw.h"


#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "nwxmp.h"

#define SIZE_HOST		256
#define SIZE_VAR		256
#define SIZE_VALUE		256
#define SIZE_TIME		256



VOID CB_Timer(TIMER_DESCR *timer, ULONG now, VOID *parm);
VOID CB_UDP_Recv(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param);

static NW_DG *UDP_Conn;

static BYTE  fr[MAX_LEN];
static char *cpHost;
static int iL1,iL2, iL1step,iL2step;
static int iL1min,iL1max,iL1steps;
static int iL2min,iL2max,iL2steps;
static int iL1tel, iL2tel;
static int iNrFrames,iID;
static int iReady;
static int iSndCnt;
static int iRecCnt;
static FRAME *results;

int main(int argc, char **argv)
{
    TIMER_DESCR *handle1;
    int i;

    /* read and check the arguments */
    if(argc != 10)
    {
        printf("%d\n",argc);
	fprintf(stderr,"call as: nwround <host> <l1min> <l1max> <l1step> <l2min> <l2max> l2step> <nr> <id>\n");
	exit(1);
    }

    cpHost = argv[1];
    iL1min = atoi(argv[2]);
    iL1max = atoi(argv[3]);
    iL1steps = atoi(argv[4]);
    iL2min = atoi(argv[5]);
    iL2max = atoi(argv[6]);
    iL2steps = atoi(argv[7]);
    iNrFrames = atoi(argv[8]);
    iID = atoi(argv[9]);

    if( 
       (iL1max > MAX_LEN) || (iL1min < sizeof(FRAME)) ||
       (iL2max > MAX_LEN) || (iL2min < sizeof(FRAME)) ||
       (iNrFrames > MAX_NR))
    {
	fprintf(stderr," error in parameters\n");
	exit(2);
    }

    /* allocate the results array */
    
    if((results = (FRAME *) malloc(sizeof(FRAME)*(iL1steps+1)*(iL2steps+1)*iNrFrames)) == NULL)
    {
	fprintf(stderr,"could not allocate results array\n");
	exit(1);
    }

    /* Initialise all subsystems */

	if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
		DnpapExit(1);
	}

	MessageInit("stdio:out", MSGFILE, MSGSTDIO, NULL);

	if (NwInit() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MibInit failed");
		DnpapExit(1);
	}

	if (BlockInit() != 0)
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");

    if (!TimerInit())
	exit(1);

    /* open a UDP connection with the target host's echo port */
    if((UDP_Conn = NwDgConnect(cpHost,CB_UDP_Recv,NULL)) == NULL)
    {
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgConnect failed<%s>",
		     argv[1]);
	DnpapExit(1);
    }
    /* initiate a time which will fire once */
    if ((handle1=TimerRegister(
			       CB_Timer,
			       NULL, /* optional argument */
			       1000L, /* duration timer in ms */
			       1, /* number of times to fire timer */
			       TIMER_TYPE_RECOVER))==0)
        exit(1);
    /* Start the event handling*/
    iReady = 0;
    while (! iReady)
	if (Block() != 0)
	    DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");

    /* print results to stdout */
    for(i=0;i<iRecCnt;i++)
    {
	printf("%lu %lu %lu %lu %lu %lu %lu %lu %lu\n",
	       results[i].type,
	       results[i].id,
	       results[i].cnt,
	       results[i].l1,
	       results[i].l2,
	       results[i].t1,
	       results[i].t2,
	       results[i].t3,
	       results[i].t4);
    }
    return 0;
}


VOID CB_Timer(TIMER_DESCR *timer, ULONG now, VOID *parm)
{
    FRAME *fr1;
    /* send the first datagram to the destination */
    iSndCnt = 0;
    iRecCnt = 0;
    iL1step = (iL1max - iL1min)/(iL1steps-1);
    iL2step = (iL2max - iL2min)/(iL2steps-1);
    iL1tel = 1;
    iL1tel = 1;
    iL1 = iL1min;
    iL2 = iL2min;

    fr1 = (FRAME *) fr;
    fr1->type = htonl(TYPE_ECHO);
    fr1->id = htonl(iID);
    fr1->cnt = htonl(iSndCnt);
    fr1->l1 = htonl(iL1);
    fr1->l2 = htonl(iL2);
    fr1->t1 = htonl(TimeUsec());
    if(!NwDgSend(UDP_Conn,(BYTE *) fr1,iL1))
    {	
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgCSend failed");
	DnpapExit(1);
    }       
}



VOID CB_UDP_Recv(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param)
{
    int i=0;
    long waittime;
    FRAME *fr1;
    /* first handle the new frame */
    fr1 = (FRAME *) frame;
    fr1->t4 = htonl(TimeUsec());

    results[iRecCnt].type = ntohl(fr1->type);
    results[iRecCnt].id = ntohl(fr1->id);
    results[iRecCnt].cnt = ntohl(fr1->cnt);
    results[iRecCnt].l1 = ntohl(fr1->l1);
    results[iRecCnt].l2 = ntohl(fr1->l2);
    results[iRecCnt].t1 = ntohl(fr1->t1);
    results[iRecCnt].t2 = ntohl(fr1->t2);
    results[iRecCnt].t3 = ntohl(fr1->t3);
    results[iRecCnt].t4 = ntohl(fr1->t4);
    iRecCnt++;
    /* send the next frame */
    
    if( (iL2tel == iL2steps) && (iL1tel == iL1steps)) 
      {  if(! (iSndCnt < iNrFrames))
	   {
	     iReady = 1;
	     return;
           }
         else
	   {
             iSndCnt++;
             iL1 = iL1min;
             iL1tel = 1;
             iL2 = iL2min;
             iL2tel = 1;
           }
       }
       else
	 {
           if (iL2tel == iL2steps)
	     { 
               iL2 = iL2min;
               iL2tel = 1;
               iL1 += iL1step;
               iL1tel++;
             }
           else
	     {
               iL2 += iL2step;
               iL2tel++;
	       i+=10;
               i-=10;

	     }
	 }
    
    fr1 = (FRAME *) fr;
    fr1->type = htonl(TYPE_ECHO);
    fr1->id = htonl(iID);
    fr1->cnt = htonl(iSndCnt);
    fr1->l1 = htonl(iL1);
    fr1->l2 = htonl(iL2);
    fr1->t1 = htonl(TimeUsec());
    if(!NwDgSend(UDP_Conn,(BYTE *) fr1,iL1))
    {	
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgCSend failed");
	DnpapExit(1);
    }       
}










